/* ##############################################################################################
**
** Copyright 2012-2016 CNRS, INPT
** Copyright 2013-2015 UPS
**  
** This file is part of qr_mumps.
**  
** qr_mumps is free software: you can redistribute it and/or modify
** it under the terms of the GNU Lesser General Public License as 
** published by the Free Software Foundation, either version 3 of 
** the License, or (at your option) any later version.
**  
** qr_mumps is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU Lesser General Public License for more details.
**  
** You can find a copy of the GNU Lesser General Public License
** in the qr_mumps/doc directory.
**
** ##############################################################################################*/


/*##############################################################################################*/
/** @file qrm_mumps.h
 * Header file for the C interface
 *
 */
/*##############################################################################################*/

#if defined(zprec) || defined(cprec)
#include <complex.h>
#endif

#include <string.h>
#include "qrm_common_c.h"

struct dqrm_spmat_type_c{
  int              *irn, *jcn;
  double  *val;
  int               m, n, nz, sym;
  void             *h; 
};

struct dqrm_spfct_type_c{
  int            m;
  int            n;
  int            nz;
  int            sym;
  int           *cperm_in;
  int            icntl[20];
  float          rcntl[10];
  long long int  gstats[10];
  void          *h; 
};


int  dqrm_spmat_init_c(struct dqrm_spmat_type_c *qrm_spmat_c);
int  dqrm_spmat_destroy_c(struct dqrm_spmat_type_c *qrm_spmat_c);
int  dqrm_spfct_init_c(struct dqrm_spfct_type_c *qrm_spfct_c, struct dqrm_spmat_type_c *qrm_spmat_c);
int  dqrm_spfct_destroy_c(struct dqrm_spfct_type_c *qrm_spfct_c);
void dqrm_readmat_c(char *matfile, struct dqrm_spmat_type_c *qrm_spmat_c);
int  dqrm_analyse_c(struct dqrm_spmat_type_c *qrm_spmat_c,
                    struct dqrm_spfct_type_c *qrm_spfct_c,
                    const char transp);
int  dqrm_factorize_c(struct dqrm_spmat_type_c *qrm_spmat_c,
                      struct dqrm_spfct_type_c *qrm_spfct_c,
                      const char transp);
int  dqrm_solve_c(struct dqrm_spfct_type_c *qrm_spfct_c, const char transp,
                  double *b, double *x, const int nrhs);
int  dqrm_apply_c(struct dqrm_spfct_type_c *qrm_spfct_c, const char transp,
                  double *b, const int nrhs);
void dqrm_spmat_mv_c(struct dqrm_spmat_type_c *qrm_spmat_c, const char transp,
                     const double alpha, double *x, 
                     const double beta, double *y, 
                     const int nrhs);
int  dqrm_spmat_nrm_c(struct dqrm_spmat_type_c *qrm_spmat_c, const char ntype, 
                   double *nrm);
int  dqrm_vecnrm_c(const double *x, const int n, const int nrhs, 
                   const char ntype, double *nrm);
int  dqrm_spbackslash_c(struct dqrm_spmat_type_c *qrm_spmat_c, double *b, 
                        double *x, const int nrhs, const char transp);
int  dqrm_spfct_backslash_c(struct dqrm_spfct_type_c *qrm_spfct_c, double *b, 
                          double *x, const int nrhs, const char transp);
int  dqrm_spposv_c(struct dqrm_spmat_type_c *qrm_spmat_c, double *b, 
                          double *x, const int nrhs);
int  dqrm_least_squares_c(struct dqrm_spmat_type_c *qrm_spmat_c, double *b, 
                          double *x, const int nrhs, const char transp);
int  dqrm_min_norm_c(struct dqrm_spmat_type_c *qrm_spmat_c, double *b, 
                          double *x, const int nrhs, const char transp);
int  dqrm_residual_norm_c(struct dqrm_spmat_type_c *qrm_spmat_c, double *b, 
                          double *x, const int nrhs, double *nrm, const char transp);
int  dqrm_residual_orth_c(struct dqrm_spmat_type_c *qrm_spmat_c, double *r, 
                          const int nrhs, double *nrm, const char transp);


int  dqrm_spfct_set_i4_c (struct dqrm_spfct_type_c *qrm_spfct_c, const char *string, int val);
int  dqrm_spfct_set_r4_c (struct dqrm_spfct_type_c *qrm_spfct_c, const char *string, float val);
int  dqrm_spfct_get_i4_c (struct dqrm_spfct_type_c *qrm_spfct_c, const char *string, int *val);
int  dqrm_spfct_get_r4_c (struct dqrm_spfct_type_c *qrm_spfct_c, const char *string, float *val);
int  dqrm_spfct_get_i8_c(struct dqrm_spfct_type_c *qrm_spfct_c, const char *string, long long *val);
