!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_print_tree.F90
!! This file contains various routine for printing trees
!!
!!
!! ##############################################################################################


subroutine qrm_print_nsteps_tree(file, adata, small)

  use qrm_adata_mod
  use qrm_mem_mod
  implicit none

  type(qrm_adata_type), target :: adata
  character                    :: file*(*)
  integer, optional            :: small

  integer                      :: i, p, n, f
  integer                      :: ismall
  integer, pointer             :: nfrows(:)

  if(present(small)) then
     ismall = small
  else
     ismall = 0
  end if

  if(.not. qrm_allocated(adata%nfrows)) then
     nfrows => adata%rc
  else
     nfrows => adata%nfrows
  end if
  
  open(4, file=file, action='write')
  
  write(4,'("graph G {")')
  write(4,'("node [color=black,")')
  write(4,'("fillcolor=white,")')
  write(4,'("shape=circle,")')
  write(4,'("style=filled")')
  write(4,'("];")')

  do i=1, adata%nnodes

     f = adata%torder(i)
     
     if(adata%small(f).eq.0) then
        write(4,'("node",i9.9,"[fillcolor=white, label="" node:",i9,"\n m:",i9,"\n n:",'// &
             'i9,"\n np:",i9,"""];")')f,f,nfrows(f),adata%rc(f),                           &
             adata%cp_ptr(f+1)-adata%cp_ptr(f)
     else if((adata%small(f).gt.0) .and. (ismall.ge.1)) then
        write(4,'("node",i9.9,"[fillcolor=gray , label="" node:",i9,"\n m:",i9,"\n n:",'// &
             'i9,"\n np:",i9,"""];")')f,f,nfrows(f),adata%rc(f),                           &
             adata%cp_ptr(f+1)-adata%cp_ptr(f)
     else if((adata%small(f).lt.0) .and. (ismall.ge.2)) then
        write(4,'("node",i9.9,"[fillcolor=gray , label="" node:",i9,"\n m:",i9,"\n n:",'// &
             'i9,"\n np:",i9,"""];")')f,f,nfrows(f),adata%rc(f),                           &
             adata%cp_ptr(f+1)-adata%cp_ptr(f)
     end if
     
  end do



  do i=1, adata%nnodes

     f = adata%torder(i)
     if((adata%small(f).gt.0) .and. ismall.lt.1) cycle
     if((adata%small(f).lt.0) .and. ismall.lt.2) cycle
     p = adata%parent(f)

     if(p.eq.0) cycle
     
     write(4,'("node",i9.9," -- node",i9.9)')p,f

  end do

  write(4,'("}")')

  close(4)


  return

end subroutine qrm_print_nsteps_tree




