!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################

! This routine is a variant of the LAPACK _TPQRT routine which allows
! for handling matrices that have a staircase structure

subroutine cqrm_tpqrt( m, n, l, nb, stair, a, lda, b, ldb, t, ldt, work, info )
  use qrm_parameters_mod
  implicit none
  
  integer   :: info, lda, ldb, ldt, n, m, l, nb, ofsa, ofsb
  integer   :: stair(*)
  complex(r32) :: a( lda, * ), b( ldb, * ), t( ldt, * ), work( * )

  integer   ::  i, j, ib, im, in, il, js

  ! if((l.ne.0) .and. (l.ne.m)) then
     ! call xerbla( 'ctpqrt', -9999 )
     ! return
  ! end if

  
  i = 1
  ! move to the first nonzero panel
  if(stair(1) .lt. 0) then
     js = 1
  else
     do js=1, n, nb
        if(stair(min(js+nb,n)) .gt. 0) exit
        i = i+nb
     end do
  end if
  
  if(js.gt.n) return
  
  do j=js, n, nb
     ib = min( n-j+1, nb )
     if(stair(1) .lt. 0) then
        ! in this case we can have any l
        im = min( m-l+i+ib-1, m )
        if( i.ge.l ) then
           il = 0
        else
           il = im-m+l-i+1
        end if
     else
        ! In this case the block is traversed by stair and tuhs we
        ! only hande the case l==0
        im = min(max(stair(j+ib-1),0),m)
        il = 0
     end if

     if(im.gt.0) then
        call ctpqrt2( im, ib, il, a(i,j), lda, b( 1, j ), ldb, &
             & t(1, j ), ldt, info )
        
        if( j+ib.le.n ) then
           call ctprfb( qrm_left, qrm_conj_transp, &
                qrm_forward, qrm_colwise,          &
                im, n-j-ib+1, ib, il,              &
                b( 1, j ), ldb,                    &
                t( 1, j ), ldt,                    &
                a( i, j+ib ), lda,                 &
                b( 1, j+ib ), ldb,                 &
                work, ib )
        end if
     end if

     i = i+ib
  end do

  do j=1, n
     t(mod(j-1,nb)+2:ldt,j) = qrm_czero
  end do
  
  return

end subroutine cqrm_tpqrt

