!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################

subroutine cqrm_trmm(side, uplo, transa, diag, m, n, k, alpha, a, lda, b, ldb)

  use qrm_parameters_mod
  implicit none
  
  complex(r32)   :: alpha
  integer        :: lda, ldb, m, n, k
  character      :: diag, side, transa, uplo
  complex(r32)   :: a(lda,*),b(ldb,*)

  integer        :: rt, ct, rs, cs

  if(side.eq.qrm_left) then
     if (uplo .eq.qrm_upper) then

        ct = 1
        rs = 1
        if(k.ge.m) then
           cs = m+1
           rt = 1
        else
           rt = m-k+1
           cs = 1
        end if

        if (((k.gt.m) .and. ((transa.eq.qrm_transp) .or. &
             (transa.eq.qrm_conj_transp))) .or.           &
             ((m.gt.k) .and. (transa.eq.qrm_no_transp))) then
           call cgemm(transa, qrm_no_transp, &
                abs(m-k), n, min(k,m),       &
                qrm_cone,                    &
                a(rs,cs), lda,               &
                b(rt,1), ldb,                &
                alpha,                       &
                b(cs,1), ldb)
        end if

        call ctrmm(side, uplo, transa, diag, &
             min(m,k), n,                    &
             alpha,                          &
             a(rt,ct), lda,                  &
             b(rt,1), ldb)
        
        if ( ((k.gt.m) .and. (transa.eq.qrm_no_transp)) .or. &
             ((m.gt.k) .and. ((transa.eq.qrm_transp) .or. &
             (transa.eq.qrm_no_transp))) ) then
           call cgemm(transa, qrm_no_transp, &
                min(m,k), n, abs(k-m),       &
                alpha,                       &  
                a(rs,cs), lda,               &
                b(cs,1), ldb,                &
                qrm_cone,                    &
                b(rt,1), ldb)
        end if
     else
        write(*,'("TRMM variant with UPLO=L not yet implemented")')
     end if
  else
     write(*,'("TRMM variant with side=R not yet implemented")')
  end if

  return
end subroutine cqrm_trmm
