!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################

subroutine dqrm_trsm(side, uplo, transa, diag, m, n, k, alpha, a, lda, b, ldb)
  use qrm_parameters_mod
  implicit none
  
  real(r64)   :: alpha
  integer        :: lda, ldb, m, n, k
  character      :: diag, side, transa, uplo
  real(r64)   :: a(lda,*),b(ldb,*)

  integer        :: rt, ct, rs, cs
  real(r64)   :: ialpha
  
  if(side.eq.qrm_left) then
     if (uplo .eq.qrm_upper) then

        rt = m - min(m,k) + 1
        ct = 1
        rs = 1
        if(k.ge.m) then
           cs = m+1
        else
           cs = 1
        end if
        
        ialpha = alpha
        if (((k.gt.m) .and. (transa.eq.qrm_no_transp)) .or.  &
            ((m.gt.k) .and. ((transa.eq.qrm_transp).or.      &
            (transa.eq.qrm_transp)))) then
           call dgemm(transa, qrm_no_transp, &
                min(m,k), n, abs(k-m),       &
                qrm_dmone,                   &
                a(rs,cs), lda,               &
                b(cs,1), ldb,                &
                ialpha,                      &
                b(rt,1), ldb)
           ialpha = qrm_done
        end if
        call dtrsm(side, uplo, transa, diag, &
             min(m,k), n,                    &
             ialpha,                         &
             a(rt,ct), lda,                  &
             b(rt,1), ldb)
        
        if (((k.gt.m) .and. ((transa.eq.qrm_transp) .or. &
             (transa.eq.qrm_transp))) .or.          &
             ((m.gt.k) .and. (transa.eq.qrm_no_transp))) then
           call dgemm(transa, qrm_no_transp, &
                abs(k-m), n, min(m,k),       &
                qrm_dmone,                   &  
                a(rs,cs), lda,               &
                b(rt,1), ldb,                &
                alpha,                       &
                b(cs,1), ldb)
        end if
     else
        write(*,'("TRSM variant with UPLO=L not yet implemented")')
     end if
  else
     write(*,'("TRSM variant with side=R not yet implemented")')
  end if

  return
end subroutine dqrm_trsm
