!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################

! This routine is a variant of the LAPACK _TPMQRT routine which allows
! for handling matrices that have a staircase structure

subroutine zqrm_tpmqrt( side, trans, m, n, k, l, nb, stair, &
     & v, ldv, t, ldt, &
     & a, lda, b, ldb, work, info )

  use qrm_parameters_mod
  implicit none

  character    :: side, trans
  integer      :: info, k, ldv, lda, ldb, m, n, l, nb, ldt
  integer      :: stair(*)
  complex(r64) :: v( ldv, * ), a( lda, * ), b( ldb, * )
  complex(r64) :: t( ldt, * ), work( * )

  integer      ::  i, j, ib, im, in, il, ik, is, je, js
  logical      ::  lsame

  ! if((l.ne.0) .and. (l.ne.m)) then
     ! call xerbla( 'ztpmqrt', -9999 )
     ! return
  ! end if

  if(.not. lsame(side,qrm_left)) then
     call xerbla( 'ztpmqrt', -9998 )
     return
  end if

  is = 1
  if(stair(1) .lt. 0) then
     js = 1
  else
     do js=1, k, nb
        if(stair(min(js+nb,k)) .gt. 0) exit
        is = is+nb
     end do
  end if
  
  if(js.gt.k) return

  if(lsame(trans,qrm_conj_transp)) then
     i = is
     do j=js, k, nb
        ib = min( k-j+1, nb )
        
        if(stair(1) .lt. 0) then
           ! in this case we can have any l
           im = min( m-l+i+ib-1, m )
           if( i.ge.l ) then
              il = 0
           else
              il = im-m+l-i+1
           end if
        else
           ! In this case the block is traversed by stair and tuhs we
           ! only hande the case l==0
           im = min(max(stair(j+ib-1),0),m)
           il = 0
        end if
        
        if(im.gt.0) then
           call ztprfb( qrm_left, qrm_conj_transp, &
                qrm_forward, qrm_colwise,          &
                im, n, ib, il,                     &
                & v( 1, j ), ldv, t( 1, j ), ldt,  &
                & a( i, 1 ), lda, b, ldb, work, ib )
        end if
        
        i = i+ib
     end do
  else if(lsame(trans,qrm_no_transp)) then

     je = js + ((k-js)/nb)*nb
     i  = is + ((k-js)/nb)*nb
     do j=je, js, -nb
        ib = min( k-j+1, nb )
        if(stair(1) .lt. 0) then
           ! in this case we can have any l
           im = min( m-l+i+ib-1, m )
           if( i.ge.l ) then
              il = 0
           else
              il = im-m+l-i+1
           end if
        else
           ! In this case the block is traversed by stair and tuhs we
           ! only hande the case l==0
           im = min(max(stair(j+ib-1),0),m)
           il = 0
        end if
        
        if(im.gt.0) then
           call ztprfb( qrm_left, qrm_no_transp,  &
                qrm_forward, qrm_colwise,         &
                im, n, ib, il,                    &
                & v( 1, j ), ldv, t( 1, j ), ldt, &
                & a( i, 1 ), lda, b, ldb, work, ib )
        end if
        
        i = i-nb
     end do


  end if

  return

end subroutine zqrm_tpmqrt



subroutine zqrm_tpmqrt_old( side, trans, m, n, k, l, nb, stair, &
     & ofsa, ofsb, v, ldv, t, ldt, &
     & a, lda, b, ldb, work, info )

  use qrm_parameters_mod
  implicit none

  character    :: side, trans
  integer      :: info, k, ldv, lda, ldb, m, n, l, nb, ldt, ofsa, ofsb
  integer      :: stair(*)
  complex(r64) :: v( ldv, * ), a( lda, * ), b( ldb, * )
  complex(r64) :: t( ldt, * ), work( * )

  integer      ::  i, j, ib, im, in, il, ik, is, js, ie, je
  logical      ::  lsame

  ! if((l.ne.0) .and. (l.ne.m)) then
     ! call xerbla( 'ztpmqrt', -9999 )
     ! return
  ! end if

  if(.not. lsame(side,qrm_left)) then
     call xerbla( 'ztpmqrt', -9998 )
     return
  end if

  is = 1
  do js=1, k, nb
     if(stair(min(js+nb,k)) .gt. ofsb) exit
     if(stair(min(js+nb,k)) .gt. ofsa) is = is+nb
  end do
  if(js.gt.k) return

  if(lsame(trans,qrm_conj_transp)) then
     i = is
     do j=js, k, nb
        ib = min( k-j+1, nb )
        if(l.eq.0) then
           ! todo: check
           im = min(stair(j+ib-1)-ofsb,m)
           il = 0
        else if (l.eq.m) then
           im = min(j+ib-js,m)
           if(j-js.gt.l) then
              il = 0
           else
              il = im - j + js
           end if
        end if
        call ztprfb( qrm_left, qrm_conj_transp, &
             qrm_forward, qrm_colwise,          &
             im, n, ib, il,                     &
             & v( 1, j ), ldv, t( 1, j ), ldt,  &
             & a( i, 1 ), lda, b, ldb, work, ib )

        i = i+ib
     end do
  else if(lsame(trans,qrm_no_transp)) then

     je = js + ((k-js)/nb)*nb
     i  = is + ((k-js)/nb)*nb
     do j=je, js, -nb
        ib = min( k-j+1, nb )
        if(l.eq.0) then
           ! todo: check
           im = min(stair(j+ib-1)-ofsb,m)
           il = 0
        else if (l.eq.m) then
           im = min(j+ib-js,m)
           if(j-js.gt.l) then
              il = 0
           else
              il = im - j + js
           end if
        end if
        call ztprfb( qrm_left, qrm_no_transp,  &
             qrm_forward, qrm_colwise,         &
             im, n, ib, il,                    &
             & v( 1, j ), ldv, t( 1, j ), ldt, &
             & a( i, 1 ), lda, b, ldb, work, ib )

        i = i-nb
     end do


  end if

  return

end subroutine zqrm_tpmqrt_old
