!! ##############################################################################################
!!
!! Copyright 2012-2018 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!!
!! This file contains synchronous and asynchronous routines for
!! counting the number of zeros along the diagonal of a triangular matrix
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine cqrm_dsmat_trdcn(a, d, eps, n, info)

  use cqrm_dsmat_mod, alias=>cqrm_dsmat_trdcn
  use qrm_dscr_mod

  implicit none

  type(cqrm_dsmat_type)       :: a
  integer                     :: d
  integer, optional           :: n, info
  real(r32)                   :: eps

  type(qrm_dscr_type)         :: qrm_dscr

  ! error handling
  integer                     :: err
  character(len=*), parameter :: name='qrm_dsmat_trdcn'

  err = 0

  call qrm_dscr_init(qrm_dscr)
  
  call cqrm_dsmat_trdcn_async(qrm_dscr, a, d, eps, n)

  ! wait for all the tasks in the dscriptor
  call qrm_barrier(qrm_dscr, err)
  
  call qrm_dscr_destroy(qrm_dscr)

9998 continue
  if(present(info)) info = err
  return

end subroutine cqrm_dsmat_trdcn


subroutine cqrm_dsmat_trdcn_async(qrm_dscr, a, d, eps, n)

  use cqrm_dsmat_mod, alias=>cqrm_dsmat_trdcn_async
  use cqrm_dsmat_task_mod
  use qrm_dscr_mod

  implicit none

  type(cqrm_dsmat_type)       :: a
  integer                     :: d
  integer, optional           :: n
  real(r32)                   :: eps
  type(qrm_dscr_type)         :: qrm_dscr

  integer                     :: in, i, nn, nb

  ! error handling
  integer                     :: info
  character(len=*), parameter :: name='qrm_dsmat_trdcn_async'

  if(qrm_dscr%info.ne.0) return
  info = 0
  
  __QRM_IFPRESENT3(in, n, a%n)

  if(in.le.0) return
  
  nb = (in-1)/a%mb+1

  nb = min(nb, min(size(a%blocks,1), size(a%blocks,2)) )

  
  do i=1, nb
     nn = min(a%mb, in-(i-1)*a%mb)
     call cqrm_block_trdcn_task(qrm_dscr, &
          a%blocks(i,i), nn, d, eps)
  end do

  
9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return

end subroutine cqrm_dsmat_trdcn_async
