!! ##############################################################################################
!!
!! Copyright 2012-2018 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!!
!! This file contains synchronous and asynchronous routines for
!! computing updates corresponding to a tpqr factorization
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine dqrm_dsmat_tpmqr(v, a, b, t, ib, ts, work, prio, info)

  use dqrm_dsmat_mod, alias=>dqrm_dsmat_tpmqr
  use qrm_dscr_mod
  implicit none

  type(dqrm_dsmat_type)       :: v, a, b, t
  type(dqrm_ws_type)          :: work
  integer                     :: ib
  character                   :: ts
  integer, optional           :: prio
  integer, optional           :: info

  type(qrm_dscr_type)         :: qrm_dscr

  ! error handling
  integer                     :: err
  character(len=*), parameter :: name='qrm_dsmat_tpmqr'

  err = 0
  call qrm_dscr_init(qrm_dscr)

  call dqrm_dsmat_tpmqr_async(qrm_dscr, v, a, b, t, ib, ts, work, prio)

  ! wait for all the tasks in the dscriptor
  call qrm_barrier(qrm_dscr, err)

  call qrm_dscr_destroy(qrm_dscr)

9999 continue
  if(present(info)) info = err
  return
end subroutine dqrm_dsmat_tpmqr

subroutine dqrm_dsmat_tpmqr_async(qrm_dscr, v, a, b, t, ib, ts, work, prio)

  use dqrm_dsmat_mod, alias=>dqrm_dsmat_tpmqr_async
  use qrm_dscr_mod
  implicit none

  type(qrm_dscr_type)         :: qrm_dscr
  type(dqrm_dsmat_type)       :: v, a, b, t
  type(dqrm_ws_type)          :: work
  integer                     :: ib
  character                   :: ts
  integer, optional           :: prio

  integer                     :: i, j, k, m, nbr, nbc, nbt, iprio

  ! error handling
  integer                     :: info
  character(len=*), parameter :: name='qrm_dsmat_tpmqr_async'

  if(qrm_dscr%info.ne.0) return
  info = 0

  if(min(v%m,v%n).eq.0) return

  __QRM_IFPRESENT3(iprio, prio, 0)
  
  nbr = size(v%blocks,1)
  nbc = size(a%blocks,2)
  nbt = size(v%blocks,2)
  
  
  do k=1, nbt
     if(ts.eq.qrm_square) then
        m = nbr
     else
        m = min(k-1,nbr)
     end if
     do i = 1, m
        do j=1, nbc
           call dqrm_hitpmqrt(qrm_dscr,       &
                v%blocks(i,k),                &
                t%blocks(i,k),                &
                a%blocks(k,j),                &
                b%blocks(i,j),                &
                a%nb, ib, qrm_square,         &
                work, iprio)
        end do
     end do
     if((ts.eq.qrm_triang) .and. (m.lt.nbr)) then
        i = k
        do j=1, nbc
           call dqrm_hitpmqrt(qrm_dscr,       &
                v%blocks(i,k),                &
                t%blocks(i,k),                &
                a%blocks(k,j),                &
                b%blocks(i,j),                &
                a%nb, ib, qrm_triang,         &
                work, iprio)
        end do
     end if

  end do

9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return
end subroutine dqrm_dsmat_tpmqr_async
