!! ##############################################################################################
!!
!! Copyright 2012-2018 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!!
!! This file contains synchronous and asynchronous routines for
!! computing a triangular system solution
!!
!! ##############################################################################################

#include "qrm_common.h"


subroutine zqrm_dsmat_potrf(uplo, a, m, k, prio, info)

  use zqrm_dsmat_mod, alias => zqrm_dsmat_potrf
  use qrm_dscr_mod
  use qrm_parameters_mod
  implicit none

  type(zqrm_dsmat_type)       :: a
  character                   :: uplo
  integer, optional           :: m, k, prio, info

  type(qrm_dscr_type)         :: qrm_dscr

  ! error handling
  integer                     :: err
  character(len=*), parameter :: name='qrm_dsmat_potrf'

  err = 0

  call qrm_dscr_init(qrm_dscr)

  call zqrm_dsmat_potrf_async(qrm_dscr, uplo, a, m, k, prio)

  ! wait for all the tasks in the dscriptor
  call qrm_barrier(qrm_dscr, info)

  call qrm_dscr_destroy(qrm_dscr)

9999 continue
  return

end subroutine zqrm_dsmat_potrf

subroutine zqrm_dsmat_potrf_async(qrm_dscr, uplo, a, m, k, prio)

  use zqrm_dsmat_mod, alias => zqrm_dsmat_potrf_async
  use qrm_parameters_mod
  use qrm_dscr_mod
  use zqrm_dsmat_task_mod
  implicit none

  type(qrm_dscr_type)         :: qrm_dscr
  type(zqrm_dsmat_type)       :: a
  character                   :: uplo
  integer, optional           :: prio, m, k

  integer                     :: i, j, l, iprio
  integer                     :: im, in, ik, mm, nn, kk, mb, nb, kb

  ! error handling
  integer                     :: info
  character(len=*), parameter :: name='qrm_dsmat_potrf_async'

  if(qrm_dscr%info.ne.0) return
  if(min(a%m,a%n).le.0) return
  info = 0

  __QRM_IFPRESENT3(iprio, prio, 0)
  __QRM_IFPRESENT3(im, m, a%m)
  __QRM_IFPRESENT3(ik, k, a%m)

  ik = min(im,ik)
  if(ik.le.0) return
  
  mb  = (im-1)/a%mb+1
  kb  = (ik-1)/a%mb+1
  
  if(present(prio)) then
     iprio = prio
  else
     iprio = 0
  end if

  if(uplo.eq.qrm_upper) then
     do l=1, kb
        if(l.eq.kb) then
           kk = ik-(l-1)*a%mb
           mm = min(a%mb, im-(l-1)*a%mb)
        else
           kk = a%mb
           mm = a%mb
        end if
              
        call zqrm_potrf_task(qrm_dscr,             &
             uplo,                                 & 
             mm, kk,                               &
             a%blocks(l,l),                        & 
             iprio )
        do j = l+1, mb
           if(j.eq.mb) then
              nn = im-(j-1)*a%mb
           else
              nn = a%mb
           end if
           call zqrm_trsm_task(qrm_dscr,           & 
                qrm_left, uplo,                    &
                qrm_conj_transp, qrm_no_unit,      &
                kk, nn, mm,                        &
                qrm_zone, a%blocks(l,l),           &
                a%blocks(l,j), iprio)

           do i = l+1, j-1
              call zqrm_gemm_task(qrm_dscr,        &
                   qrm_conj_transp, qrm_no_transp, &
                   a%mb, nn, kk,                   &
                   qrm_zmone,                      &
                   a%blocks(l,i), a%blocks(l,j),   &
                   qrm_zone, a%blocks(i,j), iprio)
           end do
           call zqrm_herk_task(qrm_dscr,           &
                uplo, qrm_conj_transp,             &
                nn, kk,                            &
                qrm_dmone, a%blocks(l,j),          &
                qrm_done,  a%blocks(j,j), iprio)
        end do
     end do
  else
  end if

9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return

end subroutine zqrm_dsmat_potrf_async

