!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!!
!! This file contains the module which holds the data types used during the factorization.
!!
!! ##############################################################################################


#include "qrm_common.h"

!> @brief This module contains the definition of a dense matrix
!!
module cqrm_dsmat_task_mod

  use iso_c_binding

  interface qrm_gemm_task
     subroutine cqrm_gemm_task(qrm_dscr, transa, transb, m, n, k, alpha, a, b, beta, c, prio)
       use qrm_dscr_mod
       use cqrm_dsmat_mod
       type(qrm_dscr_type), target :: qrm_dscr
       type(cqrm_block_type)       :: a, b, c
       integer                     :: m, n, k
       integer, target             :: prio
       complex(r32), target           :: alpha, beta
       character, target           :: transa, transb
       type(c_ptr)                 :: sizeof
     end subroutine cqrm_gemm_task
  end interface qrm_gemm_task

  interface qrm_higeqrt_task
     subroutine cqrm_higeqrt_task(qrm_dscr, a, t, k, nb, ib, work, prio)
       use qrm_dscr_mod
       use cqrm_dsmat_mod
       type(qrm_dscr_type), target :: qrm_dscr
       type(cqrm_block_type)       :: a, t
       integer                     :: k, nb
       type(cqrm_ws_type)          :: work
       integer, target             :: prio, ib
     end subroutine cqrm_higeqrt_task
  end interface qrm_higeqrt_task

  interface qrm_higmeqrt_task
     subroutine cqrm_higemqrt_task(qrm_dscr, trans, a, t, c, k, j, nb, ib, work, prio)
       use qrm_dscr_mod
       use cqrm_dsmat_mod
       type(qrm_dscr_type), target :: qrm_dscr
       character, target           :: trans
       type(cqrm_block_type)       :: a, t, c
       integer                     :: k, j, nb
       type(cqrm_ws_type)          :: work
       integer, target             :: prio, ib
     end subroutine cqrm_higemqrt_task
  end interface qrm_higmeqrt_task

  interface qrm_hitpqrt_task
     subroutine cqrm_hitpqrt_task(qrm_dscr, a1, a2, t, k, nb, ib, ts, work, prio)
       use qrm_dscr_mod
       use cqrm_dsmat_mod
       type(qrm_dscr_type), target :: qrm_dscr
       type(cqrm_block_type)       :: a1, a2, t
       integer                     :: k, nb
       character                   :: ts
       type(cqrm_ws_type)          :: work
       integer, target             :: prio, ib
     end subroutine cqrm_hitpqrt_task
  end interface qrm_hitpqrt_task

  interface qrm_hitpmqrt_task
     subroutine cqrm_hitpmqrt_task(qrm_dscr, trans, v, t, c1, c2, k, j, nb, ib, ts, work, prio)
       use qrm_dscr_mod
       use cqrm_dsmat_mod
       type(qrm_dscr_type), target :: qrm_dscr
       character, target           :: trans
       type(cqrm_block_type)       :: v, c1, c2, t
       integer                     :: k, j, nb
       character                   :: ts
       type(cqrm_ws_type)          :: work
       integer, target             :: prio, ib
     end subroutine cqrm_hitpmqrt_task
  end interface qrm_hitpmqrt_task

  interface qrm_trsm_task
     subroutine cqrm_trsm_task(qrm_dscr, side, uplo, transa, diag, m, n, k, alpha, a, b, prio)
       use qrm_parameters_mod
       use qrm_dscr_mod
       use cqrm_dsmat_mod
       type(qrm_dscr_type), target :: qrm_dscr
       character, target           :: side, uplo, transa, diag
       integer                     :: m, n, k
       complex(r32)                :: alpha
       type(cqrm_block_type)       :: a, b
       integer, target             :: prio
     end subroutine cqrm_trsm_task
  end interface qrm_trsm_task

  interface qrm_trmm_task
     subroutine cqrm_trmm_task(qrm_dscr, side, uplo, transa, diag, m, n, k, alpha, a, b, prio)
       use qrm_dscr_mod
       use cqrm_dsmat_mod
       type(qrm_dscr_type), target :: qrm_dscr
       character, target           :: side, uplo, transa, diag
       integer                     :: m, n, k
       complex(r32), target           :: alpha
       type(cqrm_block_type)       :: a, b
       integer, target             :: prio
     end subroutine cqrm_trmm_task
  end interface qrm_trmm_task

  interface qrm_block_copy_task
     subroutine cqrm_block_copy_task(qrm_dscr, transp, a, b, ia, ja, ib, jb, m, n, l)
       use qrm_dscr_mod
       use cqrm_dsmat_mod
       type(qrm_dscr_type), target    :: qrm_dscr
       character                      :: transp
       type(cqrm_block_type), target  :: a
       type(cqrm_block_type), target  :: b
       integer                        :: ia, ja, ib, jb
       integer                        :: m, n, l
     end subroutine cqrm_block_copy_task
  end interface qrm_block_copy_task

  interface qrm_block_axpy_task
     subroutine cqrm_block_axpy_task(qrm_dscr, alpha, a, b, ia, ja, ib, jb, m, n, l)
       use qrm_dscr_mod
       use cqrm_dsmat_mod
       type(qrm_dscr_type), target    :: qrm_dscr
       complex(r32)                      :: alpha
       type(cqrm_block_type), target  :: a
       type(cqrm_block_type), target  :: b
       integer                        :: ia, ja, ib, jb
       integer                        :: m, n, l
     end subroutine cqrm_block_axpy_task
  end interface qrm_block_axpy_task

  interface qrm_block_extadd_task
     subroutine cqrm_block_extadd_task(qrm_dscr, a, b, br, bc, i, j, m, n, l, inout, op, rowmap, colmap &
#if defined(have_starpu)
          ,hdl &
#endif
          )
       use qrm_dscr_mod
       use cqrm_dsmat_mod
       type(qrm_dscr_type), target    :: qrm_dscr
       type(cqrm_dsmat_type), target  :: a
       type(cqrm_dsmat_type), target  :: b
       integer                        :: br, bc
       integer                        :: i, j, m, n, l
       character                      :: inout, op
       integer                        :: rowmap(:,:)
       integer, optional              :: colmap(:,:)
#if defined(have_starpu)
       type(c_ptr), optional          :: hdl
#endif
     end subroutine cqrm_block_extadd_task
  end interface qrm_block_extadd_task

  interface qrm_block_nrm_task
     subroutine cqrm_block_nrm_task(qrm_dscr, a, m, n, ssq)
       use cqrm_dsmat_mod
       use qrm_dscr_mod
       type(qrm_dscr_type), target   :: qrm_dscr
       type(cqrm_block_type), target :: a
       integer, target               :: m, n
       type(cqrm_ws_type)            :: ssq
     end subroutine cqrm_block_nrm_task
  end interface qrm_block_nrm_task

  interface qrm_block_trdcn_task
     subroutine cqrm_block_trdcn_task(qrm_dscr, a, n, d, eps)
       use cqrm_dsmat_mod
       use qrm_dscr_mod
       type(qrm_dscr_type), target   :: qrm_dscr
       type(cqrm_block_type), target :: a
       integer, target               :: n, d
       real(r32)                     :: eps
     end subroutine cqrm_block_trdcn_task
  end interface qrm_block_trdcn_task

  interface qrm_potrf_task
     subroutine cqrm_potrf_task(qrm_dscr, uplo, m, k, a, prio)
       use qrm_dscr_mod
       use cqrm_dsmat_mod
       type(qrm_dscr_type), target :: qrm_dscr
       character, target           :: uplo
       integer, target             :: m, k
       type(cqrm_block_type)       :: a
       integer, target             :: prio
     end subroutine cqrm_potrf_task
  end interface qrm_potrf_task

  interface qrm_herk_task
     subroutine cqrm_herk_task(qrm_dscr, uplo, transa, n, k, alpha, a, beta, c, prio)
       use qrm_dscr_mod
       use cqrm_dsmat_mod
       type(qrm_dscr_type)         :: qrm_dscr
       character                   :: uplo, transa
       integer                     :: n, k
       real(r32)                   :: alpha, beta
       type(cqrm_block_type)       :: a, c
       integer                     :: prio
     end subroutine cqrm_herk_task
  end interface qrm_herk_task


end module cqrm_dsmat_task_mod
