!! ##############################################################################################
!!
!! Copyright 2012-2018 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine cqrm_gemm_task(qrm_dscr, transa, transb, m, n, k, alpha, a, b, beta, c, prio)

#if defined(have_starpu)
    use fstarpu_mod
    use cqrm_starpu_mod
    use qrm_starpu_common_mod
#endif
    use qrm_parameters_mod
    use qrm_string_mod
    use qrm_dscr_mod
    use cqrm_dsmat_mod
#if defined (perfprof)
    use qrm_perfprof_mod
#endif
    implicit none

    type(qrm_dscr_type), target :: qrm_dscr
    character, target           :: transa, transb
    integer, target             :: m, n, k
    complex(r32), target        :: alpha, beta
    type(cqrm_block_type)       :: a, b, c
    integer, target             :: prio

    integer, target             :: gemmprio
    type(c_ptr)                 :: sizeof
    integer                     :: ma, na, lda, mb, nb, ldb, mc, nc, ldc
    integer(kind=8)             :: ts, te
#if defined(have_starpu)
    type(c_ptr), target           :: dscrpt
#endif

  if(qrm_dscr%info.ne.0) return
  
#if defined (have_starpu)
    if(qrm_dscr%seq.eq.0) then

#if defined(dprec)
       sizeof = FSTARPU_SZ_REAL8
#elif defined(sprec)
       sizeof = FSTARPU_SZ_REAL4
#elif defined(zprec)
       sizeof = FSTARPU_SZ_COMPLEX8
#elif defined(cprec)
       sizeof = FSTARPU_SZ_COMPLEX4
#endif
       
       if(qrm_dscr%ngpu>0) then
          gemmprio = gemm_hp_prio_
       else
          gemmprio = prio
       end if

       dscrpt = c_loc(qrm_dscr)

       call fstarpu_insert_task((/ cqrm_gemm_cl,                &
            FSTARPU_VALUE, c_loc(dscrpt), FSTARPU_SZ_C_PTR,     &
            FSTARPU_VALUE, c_loc(transa), FSTARPU_SZ_CHARACTER, &
            FSTARPU_VALUE, c_loc(transb), FSTARPU_SZ_CHARACTER, &
            FSTARPU_VALUE, c_loc(m), FSTARPU_SZ_C_INT,          &
            FSTARPU_VALUE, c_loc(n), FSTARPU_SZ_C_INT,          &
            FSTARPU_VALUE, c_loc(k), FSTARPU_SZ_C_INT,          &
            FSTARPU_VALUE, c_loc(alpha),  sizeof,               &
            FSTARPU_VALUE, c_loc(beta),   sizeof,               &
            FSTARPU_R, a%hdl,                                   &
            FSTARPU_R, b%hdl,                                   &
            FSTARPU_RW, c%hdl,                                  &
            FSTARPU_PRIORITY, c_loc(gemmprio),                  &
            FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),             &
            C_NULL_PTR /))
    else
#endif
       lda  = size(a%c,1)
       ldb  = size(b%c,1)
       ldc  = size(c%c,1)
       
       __QRM_PROF_START(ts)
       call cgemm(transa, transb, &
            m, n, k,              &
            alpha,                &
            a%c(1,1), lda,        &
            b%c(1,1), ldb,        &
            beta,                 &
            c%c(1,1), ldc)
       __QRM_PROF_STOP(qrm_prof_gemm_task_,ts,te)
#if defined (have_starpu)
    end if
#endif
    
    return
    
  end subroutine cqrm_gemm_task


#if defined (have_starpu)
recursive subroutine cqrm_starpu_gemm_cpu_func(buffers, cl_arg) bind(C)

  use qrm_parameters_mod
  use fstarpu_mod
  use qrm_string_mod
  use qrm_dscr_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none
  
  type(c_ptr), value           :: cl_arg
  type(c_ptr), value           :: buffers
                                
  integer                      :: ma, na, lda, mb, nb, ldb, mc, nc, ldc
  integer, target              :: m, n, k
  complex(r32), pointer        :: a(:,:), b(:,:), c(:,:)
  complex(r32), target         :: alpha, beta
  character, target            :: transa, transb
  type(c_ptr), target          :: dscr_c
  type(qrm_dscr_type), pointer :: qrm_dscr
  integer(kind=8)              :: ts, te
  
  call fstarpu_unpack_arg(cl_arg,(/ c_loc(dscr_c), &
       c_loc(transa), c_loc(transb),               &
       c_loc(m), c_loc(n), c_loc(k),               &
       c_loc(alpha), c_loc(beta) /))

  ! until StarPU has a proper error handling method we have to check
  ! manually
  call c_f_pointer(dscr_c, qrm_dscr)
  if(qrm_dscr%info.ne.0) return
  
  ma   = fstarpu_matrix_get_nx(buffers, 0)
  na   = fstarpu_matrix_get_ny(buffers, 0)
  lda  = fstarpu_matrix_get_ld(buffers, 0)
  mb   = fstarpu_matrix_get_nx(buffers, 1)
  nb   = fstarpu_matrix_get_ny(buffers, 1)
  ldb  = fstarpu_matrix_get_ld(buffers, 1)
  mc   = fstarpu_matrix_get_nx(buffers, 2)
  nc   = fstarpu_matrix_get_ny(buffers, 2)
  ldc  = fstarpu_matrix_get_ld(buffers, 2)
  
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 0), a, shape=(/lda,na/))
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 1), b, shape=(/ldb,nb/))
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 2), c, shape=(/ldc,nc/))
  
  __QRM_PROF_START(ts)
  call cgemm(transa, transb,  &
              m , n , k,      &
              alpha,          &
              a(1,1), lda,    &
              b(1,1), ldb,    &
              beta,           &
              c(1,1), ldc)
  __QRM_PROF_STOP(qrm_prof_gemm_task_,ts,te)
  
  return
end subroutine cqrm_starpu_gemm_cpu_func


#if defined (have_cuda)
recursive subroutine cqrm_starpu_gemm_cuda_func(buffers, cl_arg) bind(C)

  use qrm_parameters_mod
  use fstarpu_mod
  use qrm_string_mod
  use qrm_dscr_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none
  
  type(c_ptr), value           :: cl_arg
  type(c_ptr), value           :: buffers
                                
  integer                      :: ma, na, lda, mb, nb, ldb, mc, nc, ldc
  integer, target              :: m, n, k
  complex(r32), target         :: alpha, beta
  character, target            :: transa, transb
  type(c_ptr), target          :: dscr_c
  type(qrm_dscr_type), pointer :: qrm_dscr
  integer(kind=8)              :: ts, te
  type(c_ptr)                  :: handlep, ap, bp, cp
  integer                      :: id
  
  interface
     subroutine cqrm_cuda_gemm(handlep, &
          transa, transb,               &
          m, n, k,                      &
          alpha,                        &
          a, lda,                       &
          b, ldb,                       &
          beta,                         &
          c, ldc) bind(C)
       use qrm_parameters_mod
       use iso_c_binding
       type(c_ptr), value       :: handlep, a, b, c
       character(c_char), value :: transa, transb
       integer, value           :: m, n, k, lda, ldb, ldc
       complex(r32)             :: alpha, beta
     end subroutine cqrm_cuda_gemm
  end interface
  
  call fstarpu_unpack_arg(cl_arg,(/ c_loc(dscr_c), &
       c_loc(transa), c_loc(transb),               &
       c_loc(m), c_loc(n), c_loc(k),               &
       c_loc(alpha), c_loc(beta) /))

  ! until StarPU has a proper error handling method we have to check
  ! manually
  call c_f_pointer(dscr_c, qrm_dscr)
  if(qrm_dscr%info.ne.0) return
  
  ma   = fstarpu_matrix_get_nx(buffers, 0)
  na   = fstarpu_matrix_get_ny(buffers, 0)
  lda  = fstarpu_matrix_get_ld(buffers, 0)
  mb   = fstarpu_matrix_get_nx(buffers, 1)
  nb   = fstarpu_matrix_get_ny(buffers, 1)
  ldb  = fstarpu_matrix_get_ld(buffers, 1)
  mc   = fstarpu_matrix_get_nx(buffers, 2)
  nc   = fstarpu_matrix_get_ny(buffers, 2)
  ldc  = fstarpu_matrix_get_ld(buffers, 2)
  
  ap   = fstarpu_matrix_get_ptr(buffers, 0)
  bp   = fstarpu_matrix_get_ptr(buffers, 1)
  cp   = fstarpu_matrix_get_ptr(buffers, 2)

  id      = fstarpu_worker_get_id();
  handlep = qrm_dscr%cuhandles(id+1)
  
  __QRM_PROF_START(ts)
  call cqrm_cuda_gemm(handlep, &
       transa, transb, &
       m, n, k, &
       alpha, &
       ap, lda, &
       bp, ldb, &
       beta, &
       cp, ldc)
  __QRM_PROF_STOP(qrm_prof_gemm_task_,ts,te)

  return
  
end subroutine cqrm_starpu_gemm_cuda_func


#endif
#endif
