!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine cqrm_hitpqrt_task(qrm_dscr, a1, a2, t, k, nb, ib, ts, work, prio)

#if defined(have_starpu)
  use fstarpu_mod
  use cqrm_starpu_mod
  use qrm_starpu_common_mod
#endif
  use cqrm_dsmat_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(qrm_dscr_type), target   :: qrm_dscr
  type(cqrm_block_type), target :: a1, a2, t
  integer                       :: k, nb
  character                     :: ts
  type(cqrm_ws_type)            :: work
  integer, target               :: prio, ib

#if defined(have_starpu)
  type(c_ptr), target           :: dscrpt, strpt
  type(c_ptr)                   :: hdla1, hdla2, hdlt
#endif
  integer, target               :: i, j, m, l, e, tpprio, na1, lda1, lda2, nt, ldt, info
  integer(kind=8)               :: tb, te

  if(qrm_dscr%info.ne.0) return

  if(a1%partitioned) then
     i     = (k-1)*nb+1
     if(ts.eq.qrm_square) then
        m = size(a2%c,1)
        l = 0
     else if(ts.eq.qrm_triang) then
        e = min(size(a2%c,1), size(a2%c,2))
        if(i.gt.e) then
           m = e
           l = 0
        else
           l = max(min(nb, e-i+1),0)
           m = i+l-1
        end if
     end if
  else
     i     = 1
     if(ts.eq.qrm_square) then
        m = size(a2%c,1)
        l = 0
     else if(ts.eq.qrm_triang) then
        m = min(size(a2%c,1), i+size(a2%c,2)-1)
        l = m
     end if
  end if


#if defined (have_starpu)

  if(qrm_dscr%seq.eq.0) then

     ! submit a tpqrt task
     if(a1%partitioned) then
        hdla1 = a1%shdls(k)
        hdla2 = a2%shdls(k)
        hdlt  = t%shdls(k)
     else
        hdla1 = a1%hdl
        hdla2 = a2%hdl
        hdlt  = t%hdl
     end if

     if(qrm_dscr%ngpu>0) then
        tpprio = tpqrt_hp_prio_
     else
        tpprio = prio + tpqrt_prio_
     end if

     dscrpt = c_loc(qrm_dscr)
     if(qrm_allocated(a2%stair)) then
        strpt = c_loc(a2%stair(i))
     else
        strpt = c_null_ptr
     end if

     call fstarpu_insert_task((/ cqrm_tpqrt_cl,           &
          FSTARPU_VALUE, c_loc(dscrpt), FSTARPU_SZ_C_PTR, &
          FSTARPU_VALUE, c_loc(i),      FSTARPU_SZ_C_INT, &
          FSTARPU_VALUE, c_loc(m),      FSTARPU_SZ_C_INT, &
          FSTARPU_VALUE, c_loc(l),      FSTARPU_SZ_C_INT, &
          FSTARPU_VALUE, c_loc(ib),     FSTARPU_SZ_C_INT, &
          FSTARPU_VALUE, c_loc(strpt),  FSTARPU_SZ_C_PTR, &
          FSTARPU_RW, hdla1,                              &
          FSTARPU_RW, hdla2,                              &
          FSTARPU_RW, hdlt,                               &
          FSTARPU_SCRATCH, work%hdl,                      &
          FSTARPU_PRIORITY, c_loc(tpprio),                &
          FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),         &
          C_NULL_PTR /))
     ! call fstarpu_task_wait_for_all()
  else
#endif

     ! make a simple call to lapack tpqrt
     na1  = size(a1%c,2)
     lda1 = size(a1%c,1)
     lda2 = size(a2%c,1)
     nt   = size(t%c,2)
     ldt  = size(t%c,1)

     __QRM_PROF_START(tb)
     if(qrm_allocated(a2%stair)) then
        call cqrm_tpqrt(m, na1, l, ib,  &
             a2%stair(i),               &
             a1%c(i,i), lda1,           &
             a2%c(1,i), lda2,           &
             t%c(1,i), ldt,             &
             work%c(1,1), info)
     else
        call cqrm_tpqrt(m, na1, l, ib,  &
             -1,                        &
             a1%c(i,i), lda1,           &
             a2%c(1,i), lda2,           &
             t%c(1,i), ldt,             &
             work%c(1,1), info)
     end if
     __QRM_PROF_STOP(qrm_prof_tpqrt_task_,tb,te)

#if defined (have_starpu)
  end if
#endif

  return
end subroutine cqrm_hitpqrt_task

#if defined (have_starpu)
recursive subroutine cqrm_starpu_tpqrt_cpu_func(buffers, cl_arg) bind(C)
  use qrm_dscr_mod
  use fstarpu_mod
  use qrm_parameters_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(c_ptr), value           :: cl_arg
  type(c_ptr), value           :: buffers

  type(c_ptr), target          :: dscr_c, stair_c
  integer(c_int), target       :: i, m, l, ib, j, ii, jj
  integer                      :: ma1, na1, lda1, ma2, na2, lda2, nt, ldt, nw, ldw, info
  complex(r32), pointer        :: a1(:,:), a2(:,:), t(:,:), w(:,:)
  integer, pointer             :: stair(:)
  type(qrm_dscr_type), pointer :: qrm_dscr
  integer(kind=8)              :: tb, te


  call fstarpu_unpack_arg(cl_arg,(/ c_loc(dscr_c), c_loc(i), c_loc(m), c_loc(l), c_loc(ib), &
       c_loc(stair_c) /))

  ! until StarPU has a proper error handling method we have to check
  ! manually
  call c_f_pointer(dscr_c, qrm_dscr)
  if(qrm_dscr%info.ne.0) return

  ma1   = fstarpu_matrix_get_nx(buffers, 0)
  na1   = fstarpu_matrix_get_ny(buffers, 0)
  lda1  = fstarpu_matrix_get_ld(buffers, 0)

  ma2   = fstarpu_matrix_get_nx(buffers, 1)
  na2   = fstarpu_matrix_get_ny(buffers, 1)
  lda2  = fstarpu_matrix_get_ld(buffers, 1)

  nt    = fstarpu_matrix_get_ny(buffers, 2)
  ldt   = fstarpu_matrix_get_ld(buffers, 2)

  nw    = fstarpu_matrix_get_ny(buffers, 3)
  ldw   = fstarpu_matrix_get_ld(buffers, 3)

  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 0), a1, shape=(/lda1,na1/))
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 1), a2, shape=(/lda2,na2/))
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 2), t,  shape=(/ldt,nt/))
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 3), w,  shape=(/ldw,nw/))

  __QRM_PROF_START(tb)
  if(c_associated(stair_c)) then
     call c_f_pointer(stair_c, stair, shape=(/na2/))
     call cqrm_tpqrt(m, na1, l, ib, &
          stair(1),                 &
          a1(i,1), lda1,            &
          a2(1,1), lda2,            &
          t(1,1), ldt,              &
          w(1,1), info)
  else
     call cqrm_tpqrt(m, na1, l, ib, &
          -1,                       &
          a1(i,1), lda1,            &
          a2(1,1), lda2,            &
          t(1,1), ldt,              &
          w(1,1), info)
  end if
  __QRM_PROF_STOP(qrm_prof_tpqrt_task_,tb,te)

  return
end subroutine cqrm_starpu_tpqrt_cpu_func

#if defined (have_cuda)
recursive subroutine cqrm_starpu_tpqrt_cuda_func(buffers, cl_arg) bind(C)
  use qrm_dscr_mod
  use fstarpu_mod
  use qrm_parameters_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(c_ptr), value           :: cl_arg
  type(c_ptr), value           :: buffers

  type(c_ptr), target          :: dscr_c, stair_c, a1p, a2p, tp, wp, handlep
  integer(c_int), target       :: i, m, l, ib, lw
  integer                      :: ma1, na1, lda1, ma2, na2, lda2, nt, ldt, nw, ldw, info, id
  type(qrm_dscr_type), pointer :: qrm_dscr
  integer(kind=8)              :: tb, te

  interface
     subroutine cqrm_cuda_tpqrt(handlep, &
          m, n, l, nb, i,                &    
          a1, lda1,                      &    
          a2, lda2,                      &    
          t , ldt,                       &    
          wd, lwd) bind(C)
       use iso_c_binding
       type(c_ptr), value       :: handlep, a1, a2, t, wd
       integer, value           :: m, n, l, nb, i, lda1, lda2, ldt, lwd
     end subroutine cqrm_cuda_tpqrt
  end interface
  

  call fstarpu_unpack_arg(cl_arg,(/ c_loc(dscr_c), c_loc(i), c_loc(m), c_loc(l), c_loc(ib), &
       c_loc(stair_c) /))

  ! until StarPU has a proper error handling method we have to check
  ! manually
  call c_f_pointer(dscr_c, qrm_dscr)
  if(qrm_dscr%info.ne.0) return

  ma1   = fstarpu_matrix_get_nx(buffers, 0)
  na1   = fstarpu_matrix_get_ny(buffers, 0)
  lda1  = fstarpu_matrix_get_ld(buffers, 0)

  ma2   = fstarpu_matrix_get_nx(buffers, 1)
  na2   = fstarpu_matrix_get_ny(buffers, 1)
  lda2  = fstarpu_matrix_get_ld(buffers, 1)

  nt    = fstarpu_matrix_get_ny(buffers, 2)
  ldt   = fstarpu_matrix_get_ld(buffers, 2)

  nw    = fstarpu_matrix_get_ny(buffers, 3)
  ldw   = fstarpu_matrix_get_ld(buffers, 3)
  lw    = nw*ldw
  
  a1p = fstarpu_matrix_get_ptr(buffers, 0)
  a2p = fstarpu_matrix_get_ptr(buffers, 1)
  tp  = fstarpu_matrix_get_ptr(buffers, 2)
  wp  = fstarpu_matrix_get_ptr(buffers, 3)

  id      = fstarpu_worker_get_id();
  handlep = qrm_dscr%cuhandles(id+1)

  __QRM_PROF_START(tb)
  call cqrm_cuda_tpqrt(handlep, &
       m, na2, l, ib, i,        &
       a1p, lda1,               &
       a2p, lda2,               &
       tp, ldt,                 &
       wp, lw)
  __QRM_PROF_STOP(qrm_prof_tpqrt_task_,tb,te)


  return
end subroutine cqrm_starpu_tpqrt_cuda_func
#endif

#endif
