!! ##############################################################################################
!!
!! Copyright 2012-2018 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################
!! 
!! This file contains a routine for copying a block
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine dqrm_block_copy_task(qrm_dscr, transp, a, b, ia, ja, ib, jb, m, n, l)
#if defined(have_starpu)
  use fstarpu_mod
  use dqrm_starpu_mod
#endif
  use qrm_dscr_mod
  use dqrm_dsmat_mod
  use qrm_string_mod
  use iso_c_binding
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none


  type(qrm_dscr_type), target    :: qrm_dscr
  character, target              :: transp
  type(dqrm_block_type), target  :: a                    ! the source block
  type(dqrm_block_type), target  :: b                    ! the destination block
  integer, target                :: ia, ja, ib, jb       ! ia and ja are the staring row and col
                                                         ! in the A matrix matrix. ib and jb
                                                         ! the same for B
  integer, target                :: m, n, l              ! m and n are the
                                                         ! number of rows and cols to
                                                         ! be copied. l is the number of
                                                         ! rows in the upper trapezoidal
                                                         ! part of the block (same meaning
                                                         ! as in tpqrt
#if defined (have_starpu)
  integer, target                :: prio
  type(c_ptr), target            :: dscrpt
#endif
  integer(kind=8)                :: ts, te

  if(qrm_dscr%info.ne.0) return
  
#if defined (have_starpu)
  if(qrm_dscr%seq.eq.0) then
     
     dscrpt = c_loc(qrm_dscr)

     prio = 0

     call fstarpu_insert_task((/ dqrm_block_copy_cl,                &
          FSTARPU_VALUE, c_loc(dscrpt),       FSTARPU_SZ_C_PTR,     &
          FSTARPU_VALUE, c_loc(transp),       FSTARPU_SZ_CHARACTER, &
          FSTARPU_VALUE, c_loc(ia),           FSTARPU_SZ_INTEGER,   &
          FSTARPU_VALUE, c_loc(ja),           FSTARPU_SZ_INTEGER,   &
          FSTARPU_VALUE, c_loc(ib),           FSTARPU_SZ_INTEGER,   &
          FSTARPU_VALUE, c_loc(jb),           FSTARPU_SZ_INTEGER,   &
          FSTARPU_VALUE, c_loc(m),            FSTARPU_SZ_INTEGER,   &
          FSTARPU_VALUE, c_loc(n),            FSTARPU_SZ_INTEGER,   &
          FSTARPU_VALUE, c_loc(l),            FSTARPU_SZ_INTEGER,   &
          FSTARPU_R, a%hdl,                                         &
          FSTARPU_RW, b%hdl,                                        &
          FSTARPU_PRIORITY, c_loc(prio),                            &
          FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),                   &
          C_NULL_PTR /))
          call fstarpu_task_wait_for_all()

  else
#endif
     __QRM_PROF_START(ts)
     call dqrm_lacpy(transp,                   &
                     a%c, size(a%c,1), ia, ja, &
                     b%c, size(b%c,1), ib, jb, &
                     m, n, l)
     __QRM_PROF_STOP(qrm_prof_lacpy_task_,ts,te)
#if defined (have_starpu)
  end if
#endif

  return
end subroutine dqrm_block_copy_task


#if defined (have_starpu)
recursive subroutine dqrm_starpu_block_copy_cpu_func(buffers, cl_arg) bind(C)
  use fstarpu_mod
  use dqrm_dsmat_mod
  use qrm_dscr_mod
#if defined (perfprof)
    use qrm_perfprof_mod
#endif
  implicit none

  type(c_ptr), value             :: cl_arg
  type(c_ptr), value             :: buffers

  integer, target                :: ia, ja, ib, jb, m, n, l
  type(c_ptr), target            :: dscr_c
  type(qrm_dscr_type), pointer   :: qrm_dscr
  character, target              :: transp
  integer                        :: ma, na, lda, mb, nb, ldb
  real(r64), pointer          :: a(:,:), b(:,:)
  integer(kind=8)                :: ts, te
  
  call fstarpu_unpack_arg(cl_arg,(/ c_loc(dscr_c), c_loc(transp), &
       c_loc(ia), c_loc(ja), c_loc(ib), c_loc(jb),                &
       c_loc(m), c_loc(n), c_loc(l) /))

  ! until StarPU has a proper error handling method we have to check
  ! manually
  call c_f_pointer(dscr_c, qrm_dscr)

  if(qrm_dscr%info.ne.0) return

  ma   = fstarpu_matrix_get_nx(buffers, 0)
  na   = fstarpu_matrix_get_ny(buffers, 0)
  lda  = fstarpu_matrix_get_ld(buffers, 0)
  mb   = fstarpu_matrix_get_nx(buffers, 1)
  nb   = fstarpu_matrix_get_ny(buffers, 1)
  ldb  = fstarpu_matrix_get_ld(buffers, 1)
  
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 0), a, shape=(/lda,na/))
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 1), b, shape=(/ldb,nb/))
  
  __QRM_PROF_START(ts)
  call dqrm_lacpy(transp,              &
                  a(1,1), lda, ia, ja, &
                  b(1,1), ldb, ib, jb, &
                  m, n, l)
  __QRM_PROF_STOP(qrm_prof_lacpy_task_,ts,te)
  return

end subroutine dqrm_starpu_block_copy_cpu_func
#endif
