!! ##############################################################################################
!!
!! Copyright 2012-2018 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!

subroutine dqrm_block_trdcn_task(qrm_dscr, a, n, d, eps)

#if defined(have_starpu)
  use fstarpu_mod
  use dqrm_starpu_mod
#endif
  use dqrm_dsmat_mod
  use qrm_dscr_mod
  use qrm_parameters_mod
  implicit none

  type(qrm_dscr_type), target   :: qrm_dscr
  type(dqrm_block_type), target :: a
  integer, target               :: n, d
  real(r64), target             :: eps

  integer                       :: ds, i
#if defined(have_starpu)
  type(c_ptr), target           :: dscrpt, dptr
  type(c_ptr)                   :: sizeof
#endif

  if(qrm_dscr%info.ne.0) return
  
#if defined (have_starpu)
    if(qrm_dscr%seq.eq.0) then

#if defined(dprec) || defined(zprec)
       sizeof = FSTARPU_SZ_REAL8
#elif defined(sprec) || defined(cprec)
       sizeof = FSTARPU_SZ_REAL4
#endif
       dscrpt = c_loc(qrm_dscr)
       dptr   = c_loc(d)

       call fstarpu_insert_task((/ dqrm_block_trdcn_cl,         &
            FSTARPU_VALUE, c_loc(dscrpt), FSTARPU_SZ_C_PTR,     &
            FSTARPU_VALUE, c_loc(n), FSTARPU_SZ_C_INT,          &
            FSTARPU_VALUE, c_loc(dptr), FSTARPU_SZ_C_PTR,       &
            FSTARPU_VALUE, c_loc(eps),  sizeof,                 &
            FSTARPU_R, a%hdl,                                   &
            FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),             &
            C_NULL_PTR /))
    else
#endif

       ds = 0
       do i=1, n
          if(abs(a%c(i,i)) .lt. abs(eps)) then
             ds = ds+1
          end if
       end do

       if(ds.gt.0) then
          call qrm_atomic_add(d, ds)
          if(eps.lt.qrm_dzero) then
             call qrm_error_set(qrm_dscr%info, 32)
             call qrm_error_print(32, 'dqrm_starpu_block_trdcn')
          end if
       end if
       
#if defined (have_starpu)
    end if
#endif
       
  return
end subroutine dqrm_block_trdcn_task



#if defined (have_starpu)
recursive subroutine dqrm_starpu_block_trdcn_cpu_func(buffers, cl_arg) bind(C)

  use qrm_parameters_mod
  use fstarpu_mod
  use qrm_string_mod
  use qrm_dscr_mod
  implicit none
  
  type(c_ptr), value           :: cl_arg
  type(c_ptr), value           :: buffers
                                
  integer                      :: ma, na, lda, info
  integer, target              :: n
  real(r64), pointer        :: a(:,:)
  real(r64), target            :: eps
  type(c_ptr), target          :: dscr_c, dptr
  type(qrm_dscr_type), pointer :: qrm_dscr
  integer                      :: ds, i
  integer, pointer             :: d
  
  call fstarpu_unpack_arg(cl_arg,(/ c_loc(dscr_c), c_loc(n), c_loc(dptr), &
       c_loc(eps) /))
  
  ! until StarPU has a proper error handling method we have to check
  ! manually
  call c_f_pointer(dscr_c, qrm_dscr)
  if(qrm_dscr%info.ne.0) return

  ma   = fstarpu_matrix_get_nx(buffers, 0)
  na   = fstarpu_matrix_get_ny(buffers, 0)
  lda  = fstarpu_matrix_get_ld(buffers, 0)
  
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 0), a, shape=(/lda,na/))

  ds = 0
  do i=1, n
     if(abs(a(i,i)) .lt. abs(eps)) then
        ds = ds+1
     end if
  end do

  ! call c_f_pointer(dptr, d)
  if(ds.gt.0) then
     call qrm_atomic_add(dptr, ds)
     if(eps.lt.qrm_dzero) then
        call qrm_error_set(qrm_dscr%info, 32)
        call qrm_error_print(32, 'dqrm_starpu_block_trdcn')
     end if
  end if
  
  return
end subroutine dqrm_starpu_block_trdcn_cpu_func
#endif
