!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine dqrm_hitpmqrt_task(qrm_dscr, trans, v, t, c1, c2, k, j, nb, ib, ts, work, prio)

#if defined(have_starpu)
  use fstarpu_mod
  use dqrm_starpu_mod
  use qrm_starpu_common_mod
#endif
  use dqrm_dsmat_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(qrm_dscr_type), target   :: qrm_dscr
  character, target             :: trans
  type(dqrm_block_type), target :: v, c1, c2, t
  integer                       :: k, j, nb
  character                     :: ts
  type(dqrm_ws_type)            :: work
  integer, target               :: prio, ib

#if defined(have_starpu)
  type(c_ptr)                   :: hdlv, hdlc1, hdlc2, hdlt
  type(c_ptr), target           :: dscrpt, strpt
#endif
  integer, target               :: i, m, l, e, tpmprio
  integer                       :: jj, ldc1, ldc2, ldt, ldv, nc1, nv, info
  integer(kind=8)               :: tb, te

  if(qrm_dscr%info.ne.0) return

  if(v%partitioned) then
     i     = (k-1)*nb+1
     if(ts.eq.qrm_square) then
        m = size(v%c,1)
        l = 0
     else if(ts.eq.qrm_triang) then
        e = min(size(v%c,1), size(v%c,2))
        if(i.gt.e) then
           m = e
           l = 0
        else
           l = max(min(nb, e-i+1),0)
           m = i+l-1
        end if
     end if
  else
     i     = 1
     if(ts.eq.qrm_square) then
        m = size(v%c,1)
        l = 0
     else if(ts.eq.qrm_triang) then
        m = min(size(v%c,1), i+size(v%c,2)-1)
        l = m
     end if
  end if

#if defined (have_starpu)

  if(qrm_dscr%seq.eq.0) then

     ! submit a geqrt task
     if(v%partitioned) then
        hdlv  = v%shdls(k)
        hdlt  = t%shdls(k)
     else
        hdlv  = v%hdl
        hdlt  = t%hdl
     end if

     if(c1%partitioned) then
        hdlc1 = c1%shdls(j)
        hdlc2 = c2%shdls(j)
        if(qrm_dscr%ngpu>0) then  ! FIXME: ngpu is not the right thing to check
           tpmprio = tpmqrt_i_hp_prio_
        else
           tpmprio = prio + tpmqrt_i_prio_
        end if
     else
        hdlc1 = c1%hdl
        hdlc2 = c2%hdl
        if(qrm_dscr%ngpu>0) then
           tpmprio = tpmqrt_o_hp_prio_
        else
           tpmprio = prio + tpmqrt_o_prio_
        end if
     end if

     dscrpt = c_loc(qrm_dscr)
     if(qrm_allocated(v%stair)) then
        strpt = c_loc(v%stair(i))
     else
        strpt = c_null_ptr
     end if

     call fstarpu_insert_task((/ dqrm_tpmqrt_cl,              &
          FSTARPU_VALUE, c_loc(dscrpt), FSTARPU_SZ_C_PTR,     &
          FSTARPU_VALUE, c_loc(i),      FSTARPU_SZ_C_INT,     &
          FSTARPU_VALUE, c_loc(m),      FSTARPU_SZ_C_INT,     &
          FSTARPU_VALUE, c_loc(l),      FSTARPU_SZ_C_INT,     &
          FSTARPU_VALUE, c_loc(ib),     FSTARPU_SZ_C_INT,     &
          FSTARPU_VALUE, c_loc(trans),  FSTARPU_SZ_CHARACTER, &
          FSTARPU_VALUE, c_loc(strpt),  FSTARPU_SZ_C_PTR,     &
          FSTARPU_R, hdlv,                                    &
          FSTARPU_RW, hdlc1,                                  &
          FSTARPU_RW, hdlc2,                                  &
          FSTARPU_R, hdlt,                                    &
          FSTARPU_SCRATCH, work%hdl,                          &
          FSTARPU_PRIORITY, c_loc(tpmprio),                   &
          FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx),             &
          C_NULL_PTR /))
     ! call fstarpu_task_wait_for_all()
  else
#endif

     ! make a simple call to lapack tpmqrt
     nc1  = size(c1%c,2)
     ldc1 = size(c1%c,1)
     ldc2 = size(c2%c,1)
     ldv  = size(v%c,1)
     ldt  = size(t%c,1)
     nv   = size(v%c,2)
     jj   = (j-1)*nb+1

     __QRM_PROF_START(tb)
     if(qrm_allocated(v%stair)) then
        call dqrm_tpmqrt(qrm_left, trans, &
             m, nc1, nv, l, ib,           &
             v%stair(1),                  &
             v%c(1,i), ldv,               &
             t%c(1,i), ldt,               &
             c1%c(i,jj), ldc1,            &
             c2%c(1,jj), ldc2,            &
             work%c(1,1), info)
     else
        call dqrm_tpmqrt(qrm_left, trans, &
             m, nc1, nv, l, ib,           &
             -1,                          &
             v%c(1,i), ldv,               &
             t%c(1,i), ldt,               &
             c1%c(i,jj), ldc1,            &
             c2%c(1,jj), ldc2,            &
             work%c(1,1), info)
     end if
     __QRM_PROF_STOP(qrm_prof_tpmqrt_task_,tb,te)

#if defined (have_starpu)
  end if
#endif

  return
end subroutine dqrm_hitpmqrt_task



#if defined (have_starpu)
recursive subroutine dqrm_starpu_tpmqrt_cpu_func(buffers, cl_arg) bind(C)
  use qrm_parameters_mod
  use qrm_dscr_mod
  use fstarpu_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(c_ptr), value           :: cl_arg
  type(c_ptr), value           :: buffers

  integer(c_int), target       :: i, m, l, ib
  character, target            :: trans
  type(c_ptr), target          :: dscr_c, stair_c
  integer                      :: mv, nv, ldv, mc1, nc1, ldc1, mc2, nc2, ldc2, nt, ldt, nw, ldw
  integer                      :: info, ii, jj
  real(r64), pointer        :: v(:,:), c1(:,:), c2(:,:), t(:,:), w(:,:)
  integer, pointer             :: stair(:)
  type(qrm_dscr_type), pointer :: qrm_dscr
  integer(kind=8)              :: tb, te


  call fstarpu_unpack_arg(cl_arg,(/ c_loc(dscr_c), c_loc(i), c_loc(m), c_loc(l), c_loc(ib), &
       c_loc(trans), c_loc(stair_c) /))

  ! until StarPU has a proper error handling method we have to check
  ! manually
  call c_f_pointer(dscr_c, qrm_dscr)
  if(qrm_dscr%info.ne.0) return

  mv    = fstarpu_matrix_get_nx(buffers, 0)
  nv    = fstarpu_matrix_get_ny(buffers, 0)
  ldv   = fstarpu_matrix_get_ld(buffers, 0)
  mc1   = fstarpu_matrix_get_nx(buffers, 1)
  nc1   = fstarpu_matrix_get_ny(buffers, 1)
  ldc1  = fstarpu_matrix_get_ld(buffers, 1)
  mc2   = fstarpu_matrix_get_nx(buffers, 2)
  nc2   = fstarpu_matrix_get_ny(buffers, 2)
  ldc2  = fstarpu_matrix_get_ld(buffers, 2)
  nt    = fstarpu_matrix_get_ny(buffers, 3)
  ldt   = fstarpu_matrix_get_ld(buffers, 3)
  nw    = fstarpu_matrix_get_ny(buffers, 4)
  ldw   = fstarpu_matrix_get_ld(buffers, 4)

  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 0), v,  shape=(/ldv,nv/))
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 1), c1, shape=(/ldc1,nc1/))
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 2), c2, shape=(/ldc2,nc2/))
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 3), t,  shape=(/ldt,nt/))
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 4), w,  shape=(/ldw,nw/))

  __QRM_PROF_START(tb)
  if(c_associated(stair_c)) then
     call c_f_pointer(stair_c, stair, shape=(/nv/))
     call dqrm_tpmqrt(qrm_left, trans, &
          m, nc1, nv, l, ib,           &
          stair(1),                    &
          v(1,1), ldv,                 &
          t(1,1), ldt,                 &
          c1(i,1), ldc1,               &
          c2(1,1), ldc2,               &
          w(1,1), info)
  else
     call dqrm_tpmqrt(qrm_left, trans, &
          m, nc1, nv, l, ib,           &
          -1,                          &
          v(1,1), ldv,                 &
          t(1,1), ldt,                 &
          c1(i,1), ldc1,               &
          c2(1,1), ldc2,               &
          w(1,1), info)
  end if
  __QRM_PROF_STOP(qrm_prof_tpmqrt_task_,tb,te)

  ! TODO: add check for return code in info

  return
end subroutine dqrm_starpu_tpmqrt_cpu_func

#if defined (have_cuda)
recursive subroutine dqrm_starpu_tpmqrt_cuda_func(buffers, cl_arg) bind(C)
  use qrm_parameters_mod
  use qrm_dscr_mod
  use fstarpu_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none

  type(c_ptr), value           :: cl_arg
  type(c_ptr), value           :: buffers

  integer(c_int), target       :: i, m, l, ib
  character, target            :: trans
  type(c_ptr), target          :: dscr_c, stair_c, vp, tp, c1p, c2p, wp, handlep
  integer                      :: mv, nv, ldv, mc1, nc1, ldc1, mc2, nc2, ldc2, nt, ldt, nw, ldw
  integer                      :: info, id, lw
  type(qrm_dscr_type), pointer :: qrm_dscr
  integer(kind=8)              :: tb, te

  interface
     subroutine dqrm_cuda_tpmqrt(handlep, &
          side, trans,                    &
          m, n, k, l, nb,                 &
          stair, i,                       &
          v, ldv,                         &
          t, ldt,                         &
          a, lda,                         &
          b, ldb,                         &
          w, ldw) bind(C)
       use iso_c_binding
       type(c_ptr), value       :: handlep, v, t, a, b, w, stair
       character(c_char), value :: side, trans
       integer, value           :: m, n, k, l, nb, i, ldt, lda, ldb, ldw, ldv
     end subroutine dqrm_cuda_tpmqrt
  end interface

  call fstarpu_unpack_arg(cl_arg,(/ c_loc(dscr_c), c_loc(i), c_loc(m), c_loc(l), c_loc(ib), &
       c_loc(trans), c_loc(stair_c) /))

  ! until StarPU has a proper error handling method we have to check
  ! manually
  call c_f_pointer(dscr_c, qrm_dscr)
  if(qrm_dscr%info.ne.0) return

  mv    = fstarpu_matrix_get_nx(buffers, 0)
  nv    = fstarpu_matrix_get_ny(buffers, 0)
  ldv   = fstarpu_matrix_get_ld(buffers, 0)
  mc1   = fstarpu_matrix_get_nx(buffers, 1)
  nc1   = fstarpu_matrix_get_ny(buffers, 1)
  ldc1  = fstarpu_matrix_get_ld(buffers, 1)
  mc2   = fstarpu_matrix_get_nx(buffers, 2)
  nc2   = fstarpu_matrix_get_ny(buffers, 2)
  ldc2  = fstarpu_matrix_get_ld(buffers, 2)
  nt    = fstarpu_matrix_get_ny(buffers, 3)
  ldt   = fstarpu_matrix_get_ld(buffers, 3)
  nw    = fstarpu_matrix_get_ny(buffers, 4)
  ldw   = fstarpu_matrix_get_ld(buffers, 4)
  lw    = nw*ldw;
  
  vp    = fstarpu_matrix_get_ptr(buffers, 0)
  c1p   = fstarpu_matrix_get_ptr(buffers, 1)
  c2p   = fstarpu_matrix_get_ptr(buffers, 2)
  tp    = fstarpu_matrix_get_ptr(buffers, 3)
  wp    = fstarpu_matrix_get_ptr(buffers, 4)

  id      = fstarpu_worker_get_id();
  handlep = qrm_dscr%cuhandles(id+1)

  ! TODO: add check for return code in info
  __QRM_PROF_START(tb)
  call dqrm_cuda_tpmqrt(handlep, &
       qrm_left, trans,          &
       m, nc1, nv, l, ib,        &
       stair_c, i,               &
       vp, ldv,                  &
       tp, ldt,                  &
       c1p, ldc1,                &
       c2p, ldc2,                &
       wp, lw)
  __QRM_PROF_STOP(qrm_prof_tpmqrt_task_,tb,te)
  
  return
end subroutine dqrm_starpu_tpmqrt_cuda_func
#endif



#endif
