!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


subroutine cqrm_assemble_front(qrm_dscr, qrm_spfct, front, child)
  ! this subroutine performs the assembly of child into front
  !
  use cqrm_spfct_mod
  use cqrm_fdata_mod
  use cqrm_dsmat_mod
  use qrm_dscr_mod
  use qrm_common_mod
  use qrm_error_mod
  implicit none

  
  type(qrm_dscr_type)            :: qrm_dscr
  type(cqrm_spfct_type)          :: qrm_spfct
  type(cqrm_front_type)          :: front, child

  if(qrm_spfct%sym.gt.0) then
     if(min(child%m-child%npiv,  child%n-child%npiv).le.0) return
  else
     if(min(child%ne-child%npiv, child%n-child%npiv).le.0) return
  end if
  
#if defined(have_starpu)
  if(child%small.ne.0) then
     if(qrm_spfct%sym.gt.0) then
        call cqrm_dsmat_extadd_async(qrm_dscr, child%f, front%f,                 &
             i=child%npiv+1, j=child%npiv+1,                                     &
             m=child%m-child%npiv, n=child%n-child%npiv, l=child%m-child%npiv,   &
             inout='o', op='a',                                                  &
             rowmap = child%rowmap, colmap=child%colmap,                         &
             hdl=child%sym_handle)
     else
        call cqrm_dsmat_extadd_async(qrm_dscr, child%f, front%f,                 &
             i=child%npiv+1, j=child%npiv+1,                                     &
             m=child%ne-child%npiv, n=child%n-child%npiv, l=child%ne-child%npiv, &
             inout='o', op='c',                                                  &
             rowmap = child%rowmap, colmap=child%colmap,                         &
             hdl=child%sym_handle)
     end if
  else
#endif
     if(qrm_spfct%sym.gt.0) then
        call cqrm_dsmat_extadd_async(qrm_dscr, child%f, front%f,                 &
             i=child%npiv+1, j=child%npiv+1,                                     & 
             m=child%m-child%npiv, n=child%n-child%npiv, l=child%m-child%npiv,   &
             inout='o', op='a',                                                  &
             rowmap = child%rowmap, colmap=child%colmap)
     else
        call cqrm_dsmat_extadd_async(qrm_dscr, child%f, front%f,                 &
             i=child%npiv+1, j=child%npiv+1,                                     &
             m=child%ne-child%npiv, n=child%n-child%npiv, l=child%ne-child%npiv, &
             inout='o', op='c',                                                  &
             rowmap = child%rowmap, colmap=child%colmap)
     end if
#if defined(have_starpu)
  end if
#endif

  return

end subroutine cqrm_assemble_front
