!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


#include "qrm_common.h"

subroutine dqrm_activate_front(qrm_spfct, front, work, info)
  use dqrm_spfct_mod
  use dqrm_dsmat_mod
  use dqrm_fdata_mod
  use qrm_mem_mod
  use dqrm_utils_mod
  use qrm_common_mod
#if defined (perfprof)
  use qrm_perfprof_mod
#endif

  implicit none

  type(dqrm_spfct_type), target  :: qrm_spfct
  type(dqrm_front_type)          :: front
  integer, optional              :: work(:)
  integer, optional              :: info

  type(qrm_adata_type), pointer  :: adata
  type(dqrm_fdata_type), pointer :: fdata
  type(dqrm_front_type), pointer :: cfront
  integer                        :: i, j, k
  integer                        :: m, n, mb, nb, ib, ne, npiv, pinth
  integer                        :: fnum, father, nc, nr, mm, nn, bh, tmpmem
  integer(kind=8)                :: ts, te, ts1, te1, ts2, te2, ts3, te3
  logical                        :: dopin
  
  ! error management
  integer                        :: err
  character(len=*), parameter    :: name='qrm_activate_front'

  __QRM_PROF_START(ts)
  
  err = 0
  ! tmpmem = qrm_tot_mem
  
  adata       => qrm_spfct%adata
  fdata       => qrm_spfct%fdata
  fnum        =  front%num

  front%f%m   =  front%m   ! FIXME: check if both are needed
  front%f%n   =  front%n

  if(min(front%m,front%m).le.0) goto 9999
  
  father     =  adata%parent(fnum)
  
  m          = front%m
  n          = front%n
  nb         = front%nb
  mb         = front%mb
  ib         = front%ib
  ne         = front%ne
  nc         = (n-1)/mb + 1
  nr         = (m-1)/mb + 1
  npiv       = front%npiv

  bh         = front%bh
  
  if(err.eq.0) call qrm_alloc(front%cols,  n,   err)
  if(err.eq.0) call qrm_alloc(front%rows,  m,   err)
  if(qrm_spfct%sym.eq.0) then 
     if(err.eq.0) call qrm_alloc(front%stair, n+1, err)
  end if
  __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)

  ! initialize the column indices and build the colmap
  front%cols(1:front%n) = adata%fcol(adata%fcol_ptr(fnum): &
       & adata%fcol_ptr(fnum+1)-1)

  ! call qrm_prnt_array(front%cols, 'frontcols')
  __QRM_PROF_START(ts1)

  call dqrm_build_colmapping(adata, fdata, front, work, err)
  __QRM_INFO_CHECK(err, name,'qrm_build_colmapping',9999)

  __QRM_PROF_STOP(qrm_prof_activate1_,ts1,te1)
  __QRM_PROF_START(ts2)

  call qrm_alloc(front%arowmap, front%anrows, err)
  __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)

  call dqrm_build_rowmapping(adata, fdata, front, qrm_spfct%sym, err)
  __QRM_INFO_CHECK(err, name,'qrm_build_rowmapping',9999)

  __QRM_PROF_STOP(qrm_prof_activate2_,ts2,te2)
  __QRM_PROF_START(ts3)

  call qrm_get(qrm_spfct, 'qrm_pinth', pinth)
  dopin = (min(m,n).ge.pinth).and.(front%small.eq.0)

  if(qrm_spfct%sym.eq.0) then
     call dqrm_dsmat_init(front%f, m, n, mb, nb, info=err, pin=dopin, stair=front%stair, seq=front%small.ne.0)
     __QRM_INFO_CHECK(err, name,'qrm_dsmat_init',9999)
     call dqrm_dsmat_init_t(front%f, front%t, ib, bh, info=err, pin=dopin, seq=front%small.ne.0)
     __QRM_INFO_CHECK(err, name,'qrm_dsmat_init_t',9999)
  else
     call dqrm_dsmat_init(front%f, m, n, mb, mb, info=err, pin=dopin, seq=front%small.ne.0, uplo='u')
     __QRM_INFO_CHECK(err, name,'qrm_dsmat_init',9999)
  end if

  __QRM_PROF_STOP(qrm_prof_activate3_,ts3,te3)


  ! call  qrm_print_frontstruct(front)
9999 continue 

  __QRM_PROF_STOP(qrm_prof_activate_,ts,te)

  ! should do cleanup
  if(present(info)) info = err
  return


contains


  subroutine qrm_print_frontstruct(front)
    use qrm_common_mod
    implicit none
    type(dqrm_front_type) :: front

    write(*,'("========== ",i4," ==========")')front%num
    write(*,'("m :",i4)')front%m
    write(*,'("n :",i4)')front%n
    write(*,'("ne:",i4)')front%ne
    write(*,'("np:",i4)')front%np
    call qrm_prnt_array(front%rows  ,'rows  ')
    call qrm_prnt_array(front%cols  ,'cols  ')
    ! call qrm_prnt_array(front%colmap(:,1),'colmap')
    ! call qrm_prnt_array(front%rowmap(:,1),'rowmap')
    call qrm_prnt_array(front%aiptr ,'aiptr ')
    call qrm_prnt_array(front%ajcn  ,'ajcn  ')
    call qrm_prnt_array(front%aval  ,'aval  ')
    
    return
  end subroutine qrm_print_frontstruct

  subroutine dqrm_build_colmapping(adata, fdata, front, work, info)

    implicit none
    type(qrm_adata_type), target   :: adata
    type(dqrm_fdata_type), target  :: fdata
    type(dqrm_front_type)          :: front
    integer, target, optional      :: work(:)
    integer, optional              :: info

    integer                        :: nch, f, c, i, j, k, cn, cnpiv, ce
    integer, pointer               :: ccols(:), gcolmap(:)
    integer, allocatable           :: ptrs(:)
    type(dqrm_front_type), pointer :: cfront
    logical                        :: goodmem
    integer(kind=8)                :: ts, te, ts1, te1, ts2, te2, ts3, te3
    ! error management
    integer                        :: err
    character(len=*), parameter    :: name='qrm_activate_front'

    err = 0
    
    if(present(work)) then
       if(size(work).lt.qrm_spfct%n) then
          goodmem = .true.
       else
          goodmem = .false.
          gcolmap => work
       end if
    else
       goodmem = .true.
    end if

    f = front%num
    nch = adata%childptr(f+1)-adata%childptr(f)

    if (.not.goodmem) then
    
       do j=1, front%n
          k = front%cols(j)
          gcolmap(k)=j
       end do
    
       do i=1, nch
          c = adata%child(adata%childptr(f)+i-1)
          cfront => fdata%front_list(c)
          call qrm_alloc(cfront%colmap, cfront%n, 3, err)
          __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)
          cfront%colmap(1:cfront%npiv,:) = -1
          ccols => adata%fcol(adata%fcol_ptr(c):adata%fcol_ptr(c+1)-1)
          do j=cfront%npiv+1, cfront%n
             ! this is the column mapping on the child. cfront%colmap(k)=j
             ! means that the k-th column of cfront will be assembled into the
             ! j-th column of front
             cfront%colmap(j,1) = gcolmap(ccols(j))
             cfront%colmap(j,2) = (cfront%colmap(j,1)-1)/front%mb+1
             cfront%colmap(j,3) = mod(cfront%colmap(j,1)-1,front%mb)+1
             
          end do
       end do
       
       ! do i=1, front%anrows
          ! do j= front%aiptr(i), front%aiptr(i+1)-1
             ! front%ajcn(j) = gcolmap(front%ajcn(j))
          ! end do
       ! end do
       
    else

       __QRM_PROF_START(ts1)
       allocate(ptrs(max(nch,front%anrows)))
       
       do i=1, nch
          c = adata%child(adata%childptr(f)+i-1)
          cfront => fdata%front_list(c)
          call qrm_alloc(cfront%colmap, cfront%n, 3, err)
          __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)
          cfront%colmap(1:cfront%npiv,:) = -1  ! this cols do not have to be assembled
          ptrs(i) = cfront%npiv+1
       end do
       
       __QRM_PROF_STOP(qrm_prof_colmap1_,ts1,te1)
       __QRM_PROF_START(ts2)
       ! compute the mapping for the child nodes.  cfront%colmap(i)=j
       ! means that column i+cfront%npiv of the child has to be assembled
       ! into column j of the father.
       ! The instructions below rely on the fact that both front%cols and
       ! cfront%cols are sorted with respect to the global column
       ! permutation
       do j=1, front%n
          do i=1, nch
             c = adata%child(adata%childptr(f)+i-1)
             cfront => fdata%front_list(c)
             if(ptrs(i).gt.cfront%n) cycle
             
             ccols => adata%fcol(adata%fcol_ptr(c):adata%fcol_ptr(c+1)-1)
             
             if(ccols(ptrs(i)).eq.front%cols(j)) then
                cfront%colmap(ptrs(i),1) = j
                cfront%colmap(ptrs(i),2) = (j-1)/front%mb+1
                cfront%colmap(ptrs(i),3) = mod(j-1,front%mb)+1
                ptrs(i) = ptrs(i)+1
             end if
             
          end do
       end do
       
       __QRM_PROF_STOP(qrm_prof_colmap2_,ts2,te2)
       __QRM_PROF_START(ts3)
       ! As for the coefficients of the original sparse matrix, we simply
       ! replace the corresponding global column indices with the column
       ! indices of the front. These instructions also rely on the
       ! assumption that front%ajcn is sorted wrt the global column
       ! permutation (which is doen in the qrm_factorization_init routine
       ! ptrs(1:front%anrows) = front%aiptr(1:front%anrows)
       ! do j=1, front%n
       !    do i=1, front%anrows
       !       if(ptrs(i).ge.front%aiptr(i+1)) cycle
       !       ! this while loop is necessary to cope with the presence of
       !       ! duplicates
       !       do while(front%ajcn(ptrs(i)).eq.front%cols(j))
       !          front%ajcn(ptrs(i)) = j
       !          ptrs(i) = ptrs(i)+1
       !          if(ptrs(i).ge.front%aiptr(i+1)) exit
       !       end do
       !    end do
       ! end do

       deallocate(ptrs)
       __QRM_PROF_STOP(qrm_prof_colmap3_,ts3,te3)

    end if

9999 continue
    if(present(info)) info = err
    return

  end subroutine dqrm_build_colmapping




  subroutine dqrm_build_rowmapping(adata, fdata, front, sym, info)
    implicit none

    type(dqrm_front_type)          :: front
    type(qrm_adata_type)           :: adata
    type(dqrm_fdata_type), target  :: fdata
    integer                        :: sym
    integer, optional              :: info
    
    integer, allocatable           :: first(:)
    integer                        :: i, ne, npiv, f, c, p, row, roff

    ! error management
    integer                        :: err
    character(len=*), parameter    :: name='qrm_build_rowmapping'

    err = 0

    if(sym.gt.0) then
       do p = adata%childptr(fnum), adata%childptr(fnum+1)-1
          c = adata%child(p)
          cfront => fdata%front_list(c)
          call qrm_alloc(cfront%rowmap, cfront%m, 3, err)
          __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)
          cfront%rowmap  = cfront%colmap
       end do
       front%arowmap  = (/ ( i, i=1,front%anrows ) /)
       front%rows   = front%cols
       return
    end if
    
    if (front%num .eq. 1) then
       roff = 0
    else
       roff = adata%stair(front%num-1)
    end if

    if(front%ne .le. front%ib) then
       front%stair = front%m
       ! build rowmap for the rows from the original matrix
       row = 0
       do i=1, front%anrows
          row = row+1
          front%arowmap(i) = row
          front%rows(row) = adata%rperm(roff+i)
       end do
       
       ! build rowmap for the rows from the children
       do p = adata%childptr(fnum), adata%childptr(fnum+1)-1
          c = adata%child(p)
          cfront => fdata%front_list(c)

          ! WARNING: the qrm_dsmat_copy assumes that the rowmap has as
          ! many entries as the number of rows in the matrix. However
          ! rows beyond ne should not be accessed during assemlies and
          ! thus this should be safe
          call qrm_alloc(cfront%rowmap, cfront%ne, 3, err)
          __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)
          ! allocate(cfront%rowmap(cfront%ne))
               
          ! ne is the number of Householder vectors computed on the
          ! child c. npiv is the number of fully assembled pivots in c
          ne   = cfront%ne
          npiv = cfront%npiv
          
          ! this is the row mapping on the child. cfront%rowmap(k)=i
          ! means that the k-th row of cfront will be assembled into the
          ! i-th row of front
          cfront%rowmap(1:npiv,:) = -1  ! do not assemble these rows
          do i=npiv+1, ne
             row = row+1
             cfront%rowmap(i,1) = row
             cfront%rowmap(i,2) = (row-1)/front%mb+1
             cfront%rowmap(i,3) = mod(row-1,front%mb)+1
          end do
       end do
    else
       front%stair = 0
       allocate(first(front%anrows))
       
       first = front%n+1
       ! count in the rows from the original matrix
       do i=1, front%anrows
          ! This instruction below relies on the assumption that in each
          ! row, the coefficients are sorted in increasing order wrt the
          ! global permutation
          first(i) = front%ajcn(front%aiptr(i))
          front%stair(first(i)+1) = front%stair(first(i)+1)+1
       end do

       ! count in rows coming from CBs
       do p = adata%childptr(fnum), adata%childptr(fnum+1)-1
          c = adata%child(p)
          cfront => fdata%front_list(c)
          
          ! ne is the number of Householder vectors computed on the
          ! child c. npiv is the number of fully assembled pivots in c
          ne = cfront%ne
          npiv = cfront%npiv
          
          ! count in all the rows on the CB of c
          do i=npiv+1, ne
             f = cfront%colmap(i,1)
             front%stair(f+1) = front%stair(f+1)+1
          end do
       end do
       
       ! finalize stair
       do i=2, front%n+1
          front%stair(i) = front%stair(i)+front%stair(i-1)
       end do


       ! build rowmap for the rows from the original matrix
       do i=1, front%anrows
          f = first(i) ! f is the front-local column index of the first
          ! coefficient in this row 
          front%stair(f) = front%stair(f)+1
          row = front%stair(f)
          front%arowmap(i) = row
          front%rows(row) = adata%rperm(roff+i)
       end do
       
       ! build rowmap for the rows from the children
       do p = adata%childptr(fnum), adata%childptr(fnum+1)-1
          c = adata%child(p)
          cfront => fdata%front_list(c)
          ! WARNING: the qrm_dsmat_copy assumes that the rowmap has as
          ! many entries as the number of rows in the matrix. However
          ! rows beyond ne should not be accessed during assemlies and
          ! thus this should be safe
          call qrm_alloc(cfront%rowmap, cfront%ne, 3, err)
          __QRM_INFO_CHECK(err, name,'qrm_alloc',9999)
          
          ! ne is the number of Householder vectors computed on the
          ! child c. npiv is the number of fully assembled pivots in c
          ne   = cfront%ne
          npiv = cfront%npiv
          
          ! this is the row mapping on the child. cfront%rowmap(k)=i
          ! means that the k-th row of cfront will be assembled into the
          ! i-th row of front
          cfront%rowmap(1:-npiv,:) = -1
          do i=npiv+1, ne
             f = cfront%colmap(i,1)
             front%stair(f) = front%stair(f)+1
             row = front%stair(f)
             cfront%rowmap(i,1) = row
             cfront%rowmap(i,2) = (row-1)/front%mb+1
             cfront%rowmap(i,3) = mod(row-1,front%mb)+1
          end do
       end do
    end if
    
    ! prevent stair from being above the diagonal. CBs become a nightmare otherwise

    ! FIXME: check if at least we can skip (above) diagonal blocks
    ! that are not even touched by the staricase. In a first moment we
    ! can keep those symbolically but just skip the related
    ! operations. This will reduce the amount of flops and the number
    ! of asm operations but won't change the parent's size. Ideally we
    ! want to also take them into account when computing the parent's
    ! structure but that may be too complex.
    do j=1, front%n
       front%stair(j) = max(front%stair(j),min(j,front%m))
    end do

9999 continue
    if(present(info)) info = err
    if(allocated(first)) deallocate(first)
    return
  end subroutine dqrm_build_rowmapping

  
end subroutine dqrm_activate_front
