!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_init_block.F90
!! This file contains a routine which performs the initialization of a block in a front
!!
!!
!! ##############################################################################################


#include "qrm_common.h"


subroutine dqrm_init_block(qrm_spfct, fnum, br, bc, info)
  use qrm_parameters_mod
  use qrm_mem_mod
  use dqrm_spfct_mod
  use dqrm_fdata_mod
  use dqrm_utils_mod
  use qrm_common_mod
  use qrm_sort_mod
  use dqrm_factorization_mod, protect => dqrm_init_block
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none
  
  type(dqrm_spfct_type), target  :: qrm_spfct
  integer                        :: fnum, br, bc
  integer, optional              :: info
  
  integer                        :: i, j, row, col, c, brn, bcn

  type(dqrm_front_type), pointer :: front
  integer(kind=8)                :: ts, te

  ! error management
  integer                         :: err
  character(len=*), parameter     :: name='qrm_init_block'

  __QRM_PROF_START(ts)

  err = 0

  front => qrm_spfct%fdata%front_list(fnum)

  if( (front%n .le. 0) .or. (front%m .le. 0)) then
     goto 9999
  end if

  front%f%blocks(br,bc)%c = qrm_dzero
  
  ! at this point we're ready to assemble the rows from the original matrix
  rows: do i=1, front%anrows

     row = front%arowmap(i)
     brn = (row-1)/front%mb+1

     if(brn.lt.br) cycle rows
     if(brn.gt.br) exit  rows
     
     row = mod(row-1,front%mb)+1

     !sweep this row and assemble it
     cols: do j=front%aiptr(i), front%aiptr(i+1)-1
        col = front%ajcn(j)
        bcn = (col-1)/front%mb+1
        
        if(bcn.lt.bc) cycle cols
        if(bcn.gt.bc) cycle cols
        col = mod(col-1,front%mb)+1
        front%f%blocks(brn,bcn)%c(row,col) = front%f%blocks(brn,bcn)%c(row,col)+front%aval(j)
     end do cols
  end do rows

  
9999 continue

  __QRM_PROF_STOP(qrm_prof_init_block_task_,ts,te)

  if(present(info)) info = err
  return

end subroutine dqrm_init_block
