!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_init_front.F90
!! This file contains a routine which performs the initialization of a front
!!
!!
!! ##############################################################################################


#include "qrm_common.h"


subroutine zqrm_init_front(qrm_spfct, front, info)
  use qrm_mem_mod
  use zqrm_spfct_mod
  use zqrm_fdata_mod
  use zqrm_utils_mod
  use qrm_common_mod
  use qrm_sort_mod
  use zqrm_factorization_mod, protect => zqrm_init_front
#if defined (perfprof)
  use qrm_perfprof_mod
#endif
  implicit none
  
  type(zqrm_spfct_type), target  :: qrm_spfct
  type(zqrm_front_type)          :: front
  integer, optional              :: info
  
  integer                        :: i, j, row, col, c
  integer                        :: m, n, npiv, p, mn, bcn, brn, nc, nr, cbcn, mm
  integer                        :: mb, nb, ib

  integer                        :: fnum
  type(zqrm_front_type), pointer :: cfront
  type(qrm_adata_type), pointer  :: adata
  type(zqrm_fdata_type), pointer :: fdata
  integer(kind=8)                :: peak
  integer(kind=8)                :: ts, te

  ! error management
  integer                         :: err
  character(len=*), parameter     :: name='qrm_init_front'

  __QRM_PROF_START(ts)
  err = 0
  
  cfront => null()

  ! call system_clock(ts)
  
  ! to make things easier
  adata       => qrm_spfct%adata
  fdata       => qrm_spfct%fdata
  ! fnum        =  front%num
  ! front%m     =  adata%nfrows(fnum)
  ! front%n     =  adata%rc(fnum)

  if( (front%n .le. 0) .or. (front%m .le. 0)) then
     ! nothing to do here. Mark the front as done
     front%np = 0
     goto 9999
  end if


  ! This commented code below can be used in place of all the init_block tasks
  
  ! colsf: do j=1, front%nc
  !    rowsf: do i=1, front%nr
  !       if(qrm_allocated(front%f%blocks(i,j)%c)) front%f%blocks(i,j)%c = qrm_zzero
  !    end do rowsf
  ! end do colsf
  
  ! colst: do j=1, size(front%t%blocks,2)
  !    rowst: do i=1, size(front%t%blocks,1)
  !       if(qrm_allocated(front%t%blocks(i,j)%c)) front%t%blocks(i,j)%c = qrm_zzero
  !    end do rowst
  ! end do colst

  ! at this point we're ready to assemble the rows from the original matrix
  ! do i=1, front%anrows
     
  !    row = front%arowmap(i)
  !    brn = (row-1)/mb+1
  !    row = mod(row-1,mb)+1
     
  !    !sweep this row and assemble it
  !    do p=front%aiptr(i), front%aiptr(i+1)-1
  !       col = front%ajcn(p)
  !       bcn = (col-1)/mb+1
  !       col = mod(col-1,mb)+1
  !       front%f%blocks(brn,bcn)%c(row,col) = front%f%blocks(brn,bcn)%c(row,col)+front%aval(p)
  !    end do
  ! end do
  

  do p = adata%childptr(front%num), adata%childptr(front%num+1)-1
     c = adata%child(p)
     cfront => fdata%front_list(c)
     
     ! mn is the number of elimination steps computed on the
     ! child c. npiv is the number of fully assembled pivots in c
     mn   = min(cfront%n,cfront%m)
     npiv = cfront%npiv
     if(npiv.eq.mn) cycle
     
     ! fill-up front%rows using the rowmap computed in the activation
     do i=npiv+1, mn
        row = cfront%rowmap(i,1)
        front%rows(row) = cfront%rows(i)
     end do
  end do
  
9999 continue

  __QRM_PROF_STOP(qrm_prof_init_front_task_,ts,te)

  if(present(info)) info = err
  return

end subroutine zqrm_init_front
