!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!!
!! This file contains the cqrm_factorization_mod with  the generic interfaces
!! for all the factorization related routines.
!!
!! ##############################################################################################


!> @brief This module contains all the generic interfaces for the typed routines
!! in the factorization phase
module cqrm_factorization_mod

  !> @brief Generic interface for the @link     ::cqrm_factorization_init
  !> @endlink routine
  interface qrm_factorization_init
     subroutine cqrm_factorization_init(qrm_dscr, qrm_mat, qrm_spfct, transp)
       use cqrm_spmat_mod
       use cqrm_spfct_mod
       type(qrm_dscr_type), target    :: qrm_dscr
       type(cqrm_spmat_type)                    :: qrm_mat
       type(cqrm_spfct_type)                    :: qrm_spfct
       character, intent(in)                    :: transp
     end subroutine cqrm_factorization_init
  end interface

  interface qrm_factorization_init_core
     subroutine cqrm_factorization_init_core(qrm_mat, qrm_spfct, transp, info)
       use cqrm_spmat_mod
       use cqrm_spfct_mod
       type(cqrm_spmat_type)                    :: qrm_mat
       type(cqrm_spfct_type)                    :: qrm_spfct
       character, intent(in)                    :: transp
       integer, optional                        :: info
     end subroutine cqrm_factorization_init_core
  end interface

  !> @brief Generic interface for the @link     ::cqrm_factorize
  !> @endlink routine
  interface qrm_factorize
     subroutine cqrm_factorize(qrm_mat, qrm_spfct, transp, info)
       use cqrm_spmat_mod
       use cqrm_spfct_mod
       type(cqrm_spmat_type)                    :: qrm_mat
       type(cqrm_spfct_type)                    :: qrm_spfct
       character, optional, intent(in)          :: transp
       integer, optional                        :: info
     end subroutine cqrm_factorize
  end interface

  !> @brief Generic interface for the @link     ::cqrm_factorize_async
  !> @endlink routine
  interface qrm_factorize_async
     subroutine cqrm_factorize_async(qrm_dscr, qrm_mat, qrm_spfct, transp)
       use cqrm_spmat_mod
       use cqrm_spfct_mod
       use qrm_dscr_mod
       type(qrm_dscr_type)                      :: qrm_dscr
       type(cqrm_spmat_type)                    :: qrm_mat
       type(cqrm_spfct_type)                    :: qrm_spfct
       character, optional, intent(in)          :: transp
     end subroutine cqrm_factorize_async
  end interface

  !> @brief Generic interface for the @link     ::cqrm_factorization_core
  !> @endlink routine
  interface qrm_factorization_core
     subroutine cqrm_factorization_core(qrm_dscr, qrm_spfct)
       use cqrm_spfct_mod
       use qrm_dscr_mod
       type(qrm_dscr_type)                      :: qrm_dscr
       type(cqrm_spfct_type)                    :: qrm_spfct
     end subroutine cqrm_factorization_core
  end interface

  !> @brief Generic interface for the @link     ::cqrm_init_front
  !> @endlink routine
  interface qrm_init_front
     subroutine cqrm_init_front(qrm_spfct, front, info)
       use cqrm_fdata_mod
       use cqrm_spfct_mod
       type(cqrm_spfct_type), target            :: qrm_spfct
       type(cqrm_front_type)                    :: front
       integer, optional                        :: info
     end subroutine cqrm_init_front
  end interface qrm_init_front

  !> @brief Generic interface for the @link     ::cqrm_init_block
  !> @endlink routine
  interface qrm_init_block
     subroutine cqrm_init_block(qrm_spfct, fnum, br, bc, info)
       use cqrm_fdata_mod
       use cqrm_spfct_mod
       type(cqrm_spfct_type), target            :: qrm_spfct
       integer                                  :: fnum, br, bc
       integer, optional                        :: info
     end subroutine cqrm_init_block
  end interface qrm_init_block

  !> @brief Generic interface for the @link     ::cqrm_init_block
  !> @endlink routine
  interface qrm_clean_block
     subroutine cqrm_clean_block(qrm_spfct, fnum, br, bc, info)
       use cqrm_fdata_mod
       use cqrm_spfct_mod
       type(cqrm_spfct_type), target            :: qrm_spfct
       integer                                  :: fnum, br, bc
       integer, optional                        :: info
     end subroutine cqrm_clean_block
  end interface qrm_clean_block

  !> @brief Generic interface for the @link     ::cqrm_factorize_front
  !> @endlink routine
  interface qrm_activate_front
     subroutine cqrm_activate_front(qrm_spfct, front, work, info)
       use cqrm_fdata_mod
       use cqrm_spfct_mod
       type(cqrm_spfct_type), target            :: qrm_spfct
       type(cqrm_front_type)                    :: front
       integer, optional, target                :: work(:)
       integer, optional                        :: info
     end subroutine cqrm_activate_front
  end interface qrm_activate_front


  !> @brief Generic interface for the @link     ::cqrm_do_subtree
  !> @endlink routine
  interface qrm_do_subtree
     subroutine cqrm_do_subtree(qrm_spfct, front, flops, info)
       use cqrm_spfct_mod
       use cqrm_fdata_mod
       use qrm_parameters_mod
       type(cqrm_spfct_type), target            :: qrm_spfct
       type(cqrm_front_type)                    :: front
       real(r32)                                :: flops
       integer, optional                        :: info
     end subroutine cqrm_do_subtree
  end interface

  !> @brief Generic interface for the @link     ::cqrm_clean_front
  !> @endlink routine
  interface qrm_clean_front
     subroutine cqrm_clean_front(qrm_spfct, front, info)
       use cqrm_spfct_mod
       use cqrm_fdata_mod
       type(cqrm_spfct_type), target            :: qrm_spfct
       type(cqrm_front_type)                    :: front
       integer, optional                        :: info
     end subroutine cqrm_clean_front
  end interface


  !> @brief Generic interface for the @link     :: cqrm_assemble_front
  !> @endlink routine
  interface qrm_assemble_front
     subroutine cqrm_assemble_front(qrm_dscr, qrm_spfct, front, child)
       use cqrm_spfct_mod
       use cqrm_fdata_mod
       use qrm_dscr_mod
       type(qrm_dscr_type)            :: qrm_dscr
       type(cqrm_spfct_type)          :: qrm_spfct
       type(cqrm_front_type)          :: front, child
     end subroutine cqrm_assemble_front
  end interface qrm_assemble_front
  
end module cqrm_factorization_mod
