!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!!
!! This file contains a module with generic interfaces for all the solve typed routines
!!
!! ##############################################################################################


!> @brief This module contains all the interfaces for the typed routines
!! in the solve phase
module cqrm_solve_mod

  interface qrm_apply
     procedure                              :: cqrm_spfct_unmqr2d, cqrm_spfct_unmqr1d
  end interface qrm_apply

  !> Generic interface for the cqrm_spfct_unmqr2d and cqrm_spfct_unmqr1d routines
  !!
  interface qrm_spfct_unmqr
     procedure                              :: cqrm_spfct_unmqr2d, cqrm_spfct_unmqr1d
  end interface qrm_spfct_unmqr

  interface qrm_apply_async
     procedure                              :: cqrm_spfct_unmqr_async
  end interface qrm_apply_async
  
  interface qrm_spfct_unmqr_async
     subroutine cqrm_spfct_unmqr_async(qrm_dscr, qrm_spfct, transp, qrm_sdata_b)
       use cqrm_spfct_mod
       use cqrm_sdata_mod
       use qrm_dscr_mod
       type(qrm_dscr_type)                  :: qrm_dscr
       type(cqrm_spfct_type), intent(in)    :: qrm_spfct
       type(cqrm_sdata_type)                :: qrm_sdata_b
       character(len=*), intent(in)         :: transp
     end subroutine cqrm_spfct_unmqr_async
  end interface qrm_spfct_unmqr_async



  ! !> Generic interface for the cqrm_spfct_unmqr2d and cqrm_spfct_unmqr1d routines
  ! !!
  interface cqrm_spfct_unmqr
     subroutine cqrm_spfct_unmqr2d(qrm_spfct, transp, b, info)
       use cqrm_spfct_mod
       type(cqrm_spfct_type), intent(in)    :: qrm_spfct
       complex(r32), intent(inout)          :: b(:,:)
       character(len=*), intent(in)         :: transp
       integer, optional                    :: info
     end subroutine cqrm_spfct_unmqr2d
     subroutine cqrm_spfct_unmqr1d(qrm_spfct, transp, b, info)
       use cqrm_spfct_mod
       type(cqrm_spfct_type), intent(in)    :: qrm_spfct
       complex(r32), intent(inout)          :: b(:)
       character(len=*), intent(in)         :: transp
       integer, optional                    :: info
     end subroutine cqrm_spfct_unmqr1d
  end interface cqrm_spfct_unmqr


  !> Generic interface for the cqrm_spfct_trsm and cqrm_spfct_trsm1d  routines
  interface qrm_solve
     procedure                              :: cqrm_spfct_trsm1d, cqrm_spfct_trsm2d
  end interface qrm_solve

  interface qrm_solve_async
     procedure                              :: cqrm_spfct_trsm_async
  end interface qrm_solve_async

  !> Generic interface for the cqrm_spfct_trsm and cqrm_spfct_trsm1d  routines
  interface qrm_spfct_trsm
     procedure                              :: cqrm_spfct_trsm1d, cqrm_spfct_trsm2d
  end interface qrm_spfct_trsm

  
  interface qrm_spfct_trsm_async
     subroutine cqrm_spfct_trsm_async(qrm_dscr, qrm_spfct, transp, qrm_sdata_b, qrm_sdata_x)
       use cqrm_spfct_mod
       use cqrm_sdata_mod
       use qrm_dscr_mod
       type(qrm_dscr_type)                  :: qrm_dscr
       type(cqrm_spfct_type), intent(in)    :: qrm_spfct
       type(cqrm_sdata_type)                :: qrm_sdata_b, qrm_sdata_x
       character(len=*), intent(in)         :: transp
     end subroutine cqrm_spfct_trsm_async
  end interface qrm_spfct_trsm_async


  !> Generic interface for the cqrm_spfct_trsm2d and cqrm_spfct_trsm1d routines
  interface cqrm_spfct_trsm
     subroutine cqrm_spfct_trsm2d(qrm_spfct, transp, b, x, info)
       use cqrm_spfct_mod
       type(cqrm_spfct_type), target        :: qrm_spfct
       complex(r32), intent(inout)             :: b(:,:)
       complex(r32), intent(out)               :: x(:,:)
       character(len=*)                     :: transp
       integer, optional                    :: info
     end subroutine cqrm_spfct_trsm2d
     subroutine cqrm_spfct_trsm1d(qrm_spfct, transp, b, x, info)
       use cqrm_spfct_mod
       type(cqrm_spfct_type), target        :: qrm_spfct
       complex(r32), intent(inout)             :: b(:)
       complex(r32), intent(out)               :: x(:)
       character(len=*)                     :: transp
       integer, optional                    :: info
     end subroutine cqrm_spfct_trsm1d
  end interface cqrm_spfct_trsm

  interface qrm_node_unmqr_task
     subroutine cqrm_node_unmqr_task(qrm_dscr, transp, qrm_spfct, inode, qrm_sdata_b)
       use cqrm_spfct_mod
       use cqrm_sdata_mod
       use qrm_dscr_mod
       type(qrm_dscr_type)                  :: qrm_dscr
       type(cqrm_spfct_type), target        :: qrm_spfct
       type(cqrm_sdata_type)                :: qrm_sdata_b
       character(len=*), intent(in)         :: transp
       integer                              :: inode
     end subroutine cqrm_node_unmqr_task
  end interface qrm_node_unmqr_task
  
  interface qrm_node_trsm_task
     subroutine cqrm_node_trsm_task(qrm_dscr, transp, qrm_spfct, inode, qrm_sdata_b, qrm_sdata_x)
       use cqrm_spfct_mod
       use cqrm_sdata_mod
       use qrm_dscr_mod
       type(qrm_dscr_type)                  :: qrm_dscr
       type(cqrm_spfct_type), target        :: qrm_spfct
       type(cqrm_sdata_type)                :: qrm_sdata_b, qrm_sdata_x
       character(len=*), intent(in)         :: transp
       integer                              :: inode
     end subroutine cqrm_node_trsm_task
  end interface qrm_node_trsm_task
  
  
  
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!! Kernels
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  interface qrm_front_qt
     subroutine cqrm_front_qt(front, qrm_sdata_b, work)
       use cqrm_fdata_mod
       use cqrm_sdata_mod
       type(cqrm_front_type)                :: front
       type(cqrm_sdata_type)                :: qrm_sdata_b
       type(cqrm_ws_type)                   :: work
     end subroutine cqrm_front_qt
  end interface qrm_front_qt

  interface qrm_assemble_qt
     subroutine cqrm_assemble_qt(qrm_spfct, front, qrm_sdata_b, info)
       use cqrm_spfct_mod
       use cqrm_fdata_mod
       use cqrm_sdata_mod
       type(cqrm_spfct_type)                :: qrm_spfct
       type(cqrm_front_type)                :: front
       type(cqrm_sdata_type)                :: qrm_sdata_b
       integer, optional                    :: info
     end subroutine cqrm_assemble_qt
  end interface qrm_assemble_qt

  interface qrm_front_r
     subroutine cqrm_front_r(front, qrm_sdata_b, qrm_sdata_x)
       use cqrm_fdata_mod
       use cqrm_sdata_mod
       type(cqrm_front_type)                :: front
       type(cqrm_sdata_type)                :: qrm_sdata_b, qrm_sdata_x
     end subroutine cqrm_front_r
  end interface qrm_front_r

  interface qrm_assemble_r
     subroutine cqrm_assemble_r(qrm_spfct, front, qrm_sdata_b, qrm_sdata_x, info)
       use cqrm_spfct_mod
       use cqrm_sdata_mod
       type(cqrm_spfct_type), target        :: qrm_spfct
       type(cqrm_front_type)                :: front
       type(cqrm_sdata_type)                :: qrm_sdata_b, qrm_sdata_x
       integer, optional                    :: info
     end subroutine cqrm_assemble_r
  end interface qrm_assemble_r

  interface qrm_front_q
     subroutine cqrm_front_q(front, qrm_sdata_b, work)
       use cqrm_fdata_mod
       use cqrm_sdata_mod
       type(cqrm_front_type)                :: front
       type(cqrm_sdata_type)                :: qrm_sdata_b
       type(cqrm_ws_type)                   :: work
     end subroutine cqrm_front_q
  end interface qrm_front_q

  interface qrm_assemble_q
     subroutine cqrm_assemble_q(qrm_spfct, front, qrm_sdata_b, info)
       use cqrm_spfct_mod
       use cqrm_sdata_mod
       type(cqrm_spfct_type), target        :: qrm_spfct
       type(cqrm_front_type)                :: front
       type(cqrm_sdata_type)                :: qrm_sdata_b
       integer, optional                    :: info
     end subroutine cqrm_assemble_q
  end interface qrm_assemble_q

  interface qrm_front_rt
     subroutine cqrm_front_rt(front, qrm_sdata_b, qrm_sdata_x)
       use cqrm_fdata_mod
       use cqrm_sdata_mod
       type(cqrm_front_type)                :: front
       type(cqrm_sdata_type)                :: qrm_sdata_b, qrm_sdata_x
     end subroutine cqrm_front_rt
  end interface qrm_front_rt

  interface qrm_assemble_rt
     subroutine cqrm_assemble_rt(qrm_spfct, front, qrm_sdata_b, qrm_sdata_x, info)
       use cqrm_spfct_mod
       use cqrm_sdata_mod
       type(cqrm_spfct_type), target        :: qrm_spfct
       type(cqrm_front_type)                :: front
       type(cqrm_sdata_type)                :: qrm_sdata_b, qrm_sdata_x
       integer, optional                    :: info
     end subroutine cqrm_assemble_rt
  end interface qrm_assemble_rt

  interface qrm_spfct_unmqr_subtree
     subroutine cqrm_spfct_unmqr_subtree(transp, qrm_spfct, iroot, qrm_sdata_b, work, info)
       use cqrm_spfct_mod
       use cqrm_sdata_mod
       type(cqrm_spfct_type), intent(in)    :: qrm_spfct
       type(cqrm_sdata_type)                :: qrm_sdata_b
       type(cqrm_ws_type), intent(inout)    :: work
       character(len=*), intent(in)         :: transp
       integer                              :: iroot
       integer, optional                    :: info
     end subroutine cqrm_spfct_unmqr_subtree
  end interface qrm_spfct_unmqr_subtree

  interface qrm_spfct_trsm_subtree
     subroutine cqrm_spfct_trsm_subtree(transp, qrm_spfct, iroot, qrm_sdata_b, qrm_sdata_x, info)
       use cqrm_spfct_mod
       use cqrm_sdata_mod
       type(cqrm_spfct_type), intent(in)    :: qrm_spfct
       type(cqrm_sdata_type)                :: qrm_sdata_b, qrm_sdata_x
       character(len=*), intent(in)         :: transp
       integer                              :: iroot
       integer, optional                    :: info
     end subroutine cqrm_spfct_trsm_subtree
  end interface qrm_spfct_trsm_subtree
  
end module cqrm_solve_mod
