!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!!
!! This file contains a module with all the generic interfaces for the typed analysis routines.
!!
!! ##############################################################################################


!> This module contains the generic interfaces for all the analysis routines.
module dqrm_analysis_mod

  !> Generic interface for the dqrm_analyse routine
  interface qrm_analyse
     subroutine dqrm_analyse(qrm_mat, qrm_spfct, transp, info)
       use dqrm_spmat_mod
       use dqrm_spfct_mod
       type(dqrm_spmat_type)                :: qrm_mat
       type(dqrm_spfct_type)                :: qrm_spfct
       character, optional, intent(in)      :: transp
       integer, optional                    :: info
     end subroutine dqrm_analyse
  end interface qrm_analyse

  !> Generic interface for the dqrm_analyse routine
  interface qrm_analyse_async
     subroutine dqrm_analyse_async(qrm_dscr, qrm_mat, qrm_spfct, transp)
       use qrm_dscr_mod
       use dqrm_spmat_mod
       use dqrm_spfct_mod
       type(qrm_dscr_type)                  :: qrm_dscr
       type(dqrm_spmat_type)                :: qrm_mat
       type(dqrm_spfct_type)                :: qrm_spfct
       character, optional, intent(in)      :: transp
     end subroutine dqrm_analyse_async
  end interface qrm_analyse_async

  !> Generic interface for the dqrm_analyse routine
  interface qrm_analysis_core
     subroutine dqrm_analysis_core(qrm_mat, qrm_spfct, qrm_dscr, transp)
       use qrm_dscr_mod
       use dqrm_spmat_mod
       use dqrm_spfct_mod
       type(dqrm_spmat_type)                :: qrm_mat
       type(dqrm_spfct_type)                :: qrm_spfct
       type(qrm_dscr_type)                  :: qrm_dscr
       character, optional, intent(in)      :: transp
     end subroutine dqrm_analysis_core
  end interface qrm_analysis_core

  !> Generic interface for the dqrm_compute_graph  routine
  interface qrm_compute_graph
     subroutine dqrm_compute_graph(qrm_mat, graph, transp, info)
       use dqrm_spmat_mod
       type(dqrm_spmat_type)                :: qrm_mat
       type(dqrm_spmat_type), intent(out)   :: graph
       character                            :: transp
       integer, optional                    :: info
     end subroutine dqrm_compute_graph
  end interface

  !>  Generic interface for the dqrm_do_colamd routine
  interface qrm_do_colamd
     subroutine dqrm_do_colamd(graph, cperm, info)
       use dqrm_spmat_mod
       type(dqrm_spmat_type)                :: graph
       integer                              :: cperm(:)
       integer, optional                    :: info
     end subroutine dqrm_do_colamd
  end interface

  !> Generic interface for the dqrm_do_metis routine
  interface qrm_do_metis
     subroutine dqrm_do_metis(graph, cperm, info)
       use dqrm_spmat_mod
       type(dqrm_spmat_type)                :: graph
       integer                              :: cperm(:)
       integer, optional                    :: info
     end subroutine dqrm_do_metis
  end interface

  !> Generic interface for the dqrm_do_scotch routine
  interface qrm_do_scotch
     subroutine dqrm_do_scotch(graph, cperm, info)
       use dqrm_spmat_mod
       type(dqrm_spmat_type)                :: graph
       integer                              :: cperm(:)
       integer, optional                    :: info
     end subroutine dqrm_do_scotch
  end interface

  !> Generic interface for the dqrm_ata_graph routine
  interface qrm_ata_graph
     subroutine dqrm_ata_graph(g_csc, ata_graph, info)
       use dqrm_spmat_mod
       type(dqrm_spmat_type), intent(in)    :: g_csc
       type(dqrm_spmat_type), intent(out)   :: ata_graph
       integer, optional                    :: info
     end subroutine dqrm_ata_graph
  end interface

  !> Generic interface for the dqrm_do_ordering routine
  interface qrm_do_ordering
     subroutine dqrm_do_ordering(graph, qrm_spfct, cperm, cperm_in, info)
       use dqrm_spmat_mod
       use dqrm_spfct_mod
       type(dqrm_spmat_type)                :: graph
       type(dqrm_spfct_type)                :: qrm_spfct
       integer                              :: cperm(:)
       integer, pointer                     :: cperm_in(:)
       integer, optional                    :: info
     end subroutine dqrm_do_ordering
  end interface

  !> Generic interface for the dqrm_elim_tree routine
  interface qrm_elim_tree
     subroutine dqrm_elim_tree(graph, cperm, parent, info)
       use dqrm_spmat_mod
       type(dqrm_spmat_type), intent(in)    :: graph
       integer, intent(in)                  :: cperm(:)
       integer, allocatable                 :: parent(:)
       integer, optional                    :: info
     end subroutine dqrm_elim_tree
  end interface

  !> Generic interface for the dqrm_rowcount routine
  interface qrm_rowcount
     subroutine dqrm_rowcount(graph, parent, porder, rc, info)
       use dqrm_spmat_mod
       type(dqrm_spmat_type)                :: graph
       integer                              :: parent(:), porder(:), rc(:)
       integer, optional                    :: info
     end subroutine dqrm_rowcount
  end interface

  !> Generic interface for the dqrm_symbolic routine
  interface qrm_symbolic
     subroutine dqrm_symbolic(graph, qrm_spfct, info)
       use dqrm_spmat_mod
       use dqrm_spfct_mod
       type(dqrm_spmat_type), intent(inout) :: graph
       type(dqrm_spfct_type)                :: qrm_spfct
       integer, optional                    :: info
     end subroutine dqrm_symbolic
  end interface
  
  !> Generic interface for the dqrm_rowperm routine
  interface qrm_rowperm
     subroutine dqrm_rowperm(qrm_mat, cperm, rperm, nvar, stair, info)
       use dqrm_spmat_mod
       type(dqrm_spmat_type)                :: qrm_mat
       integer                              :: cperm(:), rperm(:), nvar(:), stair(:)
       integer, optional                    :: info
     end subroutine dqrm_rowperm
  end interface

  interface qrm_compute_memory
     subroutine dqrm_compute_memory(qrm_spfct, transp, info)
       use dqrm_spfct_mod
       type(dqrm_spfct_type)                :: qrm_spfct
       character                            :: transp
       integer, optional                    :: info
     end subroutine dqrm_compute_memory
  end interface qrm_compute_memory

  
end module dqrm_analysis_mod
   
