!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!!
!! This file contains the C interface for qr_mumps.
!!
!! ##############################################################################################


#include "qrm_common.h"

!> This module contains the definition of the qr_mumps C interface common
!! to all precisions/types.
module qrm_c_comm_interface
  use iso_c_binding
  use qrm_error_mod
  use qrm_common_mod
  use qrm_string_mod
  use qrm_parameters_mod
  
contains

  !> C equivalent of the qrm_init
  !> routine
  function qrm_init_c(ncpu, ngpu) result(info) bind(c)
    use qrm_dscr_mod
    implicit none
    integer(c_int), value  :: ncpu, ngpu
    integer(c_int)         :: info

    call qrm_init(ncpu=ncpu, ngpu=ngpu, info=info)

    return
  end function qrm_init_c
    
  !> C equivalent of the qrm_finalize
  !> routine
  subroutine qrm_finalize_c() bind(c)
    use qrm_dscr_mod
    implicit none

    call qrm_finalize()

    return
  end subroutine qrm_finalize_c
    
  !> C equivalent of the qrm_glob_set_i4
  !> routine (only for global, integer type)
  function qrm_glob_set_i4_c(string, val) result(info) bind(c)
    implicit none
    
    character(kind=c_char) :: string(*)
    integer(c_int), value  :: val
    integer(c_int)         :: info
    
    character(len=:), allocatable :: a

    a = qrm_c2f_string(string)
    
    call qrm_glob_set_i4(a, val, info)

    return
    
  end function qrm_glob_set_i4_c

  !> C equivalent of the qrm_glob_set_r4
  !> routine (only for global, integer type)
  function qrm_glob_set_r4_c(string, val) result(info) bind(c)
    implicit none

    character(kind=c_char) :: string(*)
    real(c_float), value   :: val
    integer(c_int)         :: info
    
    character(len=:), allocatable :: a

    a = qrm_c2f_string(string)
    
    call qrm_glob_set_r4(a, val, info)

    return
    
  end function qrm_glob_set_r4_c

  !> C equivalent of the qrm_glob_get_i4
  !> routine (only for global, integer type)
  function qrm_glob_get_i4_c(string, val) result(info) bind(c)
    implicit none

    character(kind=c_char) :: string(*)
    integer(c_int)         :: val
    integer(c_int)         :: info
    
    character(len=:), allocatable :: a

    a = qrm_c2f_string(string)
    
    call qrm_glob_get_i4(a, val, info)

    return
    
  end function qrm_glob_get_i4_c

  !> C equivalent of the qrm_glob_get_r4
  !> routine (only for global, real type)
  function qrm_glob_get_r4_c(string, val) result(info) bind(c)
    implicit none

    character(kind=c_char) :: string(*)
    real(c_float)          :: val
    integer(c_int)         :: info
    
    character(len=:), allocatable :: a

    a = qrm_c2f_string(string)
    
    call qrm_glob_get_r4(a, val, info)

    return
    
  end function qrm_glob_get_r4_c

  !> C equivalent of the qrm_glob_get_i8
  !> routine (only for global, integer64 type)
  function qrm_glob_get_i8_c(string, val) result(info) bind(c)
    implicit none

    character(kind=c_char)        :: string(*)
    integer(c_long_long)          :: val
    integer(c_int)                :: info
    
    character(len=:), allocatable :: a

    a = qrm_c2f_string(string)

    call qrm_glob_get_i8(a, val, info)

    return
    
  end function qrm_glob_get_i8_c



  
end module qrm_c_comm_interface

