!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


module qrm_parameters_mod

  use iso_fortran_env

  integer,      parameter :: r32=real32, r64=real64, i32=int32, i64=int64
  real(r32),    parameter :: qrm_sone        = real(1.0,r32)
  real(r64),    parameter :: qrm_done        = real(1.0,r64)
  complex(r32), parameter :: qrm_cone        = cmplx(1.0, 0.0, r32)
  complex(r64), parameter :: qrm_zone        = cmplx(1.0, 0.0, r64)
  real(r32),    parameter :: qrm_smone       = real(-1.0,r32)
  real(r64),    parameter :: qrm_dmone       = real(-1.0,r64)
  complex(r32), parameter :: qrm_cmone       = cmplx(-1.0, 0.0, r32)
  complex(r64), parameter :: qrm_zmone       = cmplx(-1.0, 0.0, r64)
  real(r32),    parameter :: qrm_szero       = real(0.0,r32)
  real(r64),    parameter :: qrm_dzero       = real(0.0,r64)
  complex(r32), parameter :: qrm_czero       = cmplx(0.0, 0.0, r32)
  complex(r64), parameter :: qrm_zzero       = cmplx(0.0, 0.0, r64)
  integer(i32), parameter :: qrm_ione        = int(1, i32)
  integer(i64), parameter :: qrm_i8one       = int(1, i64)
  integer(i32), parameter :: qrm_izero       = int(0, i32)
  character,    parameter :: qrm_conj_transp = 'c'
  character,    parameter :: qrm_transp      = 't'
  character,    parameter :: qrm_no_transp   = 'n'
  character,    parameter :: qrm_unit        = 'u'
  character,    parameter :: qrm_no_unit     = 'n'
  character,    parameter :: qrm_upper       = 'u'
  character,    parameter :: qrm_lower       = 'l'
  character,    parameter :: qrm_both        = 'b'
  character,    parameter :: qrm_left        = 'l'
  character,    parameter :: qrm_right       = 'r'
  character,    parameter :: qrm_colwise     = 'c'
  character,    parameter :: qrm_forward     = 'f'
  character,    parameter :: qrm_no          = 'n'
  character,    parameter :: qrm_square      = 's'
  character,    parameter :: qrm_triang      = 't'

  
  integer,      parameter :: qrm_sizeof_s  = storage_size(qrm_sone) /8
  integer,      parameter :: qrm_sizeof_d  = storage_size(qrm_done) /8
  integer,      parameter :: qrm_sizeof_c  = storage_size(qrm_cone) /8
  integer,      parameter :: qrm_sizeof_z  = storage_size(qrm_zone) /8
  integer,      parameter :: qrm_sizeof_i  = storage_size(qrm_ione) /8
  integer,      parameter :: qrm_sizeof_i8 = storage_size(qrm_i8one)/8
  
  integer, parameter :: qrm_noinit_=-999

  !> The unit for output messages
  integer, save      :: qrm_ounit         = qrm_noinit_
  !> The unit for error messages
  integer, save      :: qrm_eunit         = qrm_noinit_
  !> The unit for debug messages
  integer, save      :: qrm_dunit         = qrm_noinit_

  integer, save      :: qrm_ncpu          = 0 
  integer, save      :: qrm_ngpu          = 0 
  integer, save      :: qrm_print_etree   = qrm_noinit_
  
  !> The number of entries in icntl
  integer, parameter :: nicntl=20

  !> The number of entries in rcntl
  integer, parameter :: nrcntl=10

  !> constants to set control parameters
  integer, parameter :: qrm_ordering_       = 1
  integer, parameter :: qrm_sing_           = 2
  integer, parameter :: qrm_minamalg_       = 3
  integer, parameter :: qrm_mb_             = 4
  integer, parameter :: qrm_nb_             = 5
  integer, parameter :: qrm_ib_             = 6
  integer, parameter :: qrm_bh_             = 7
  integer, parameter :: qrm_keeph_          = 8
  integer, parameter :: qrm_rhsnb_          = 9
  integer, parameter :: qrm_nlz_            = nicntl
  integer, parameter :: qrm_cnode_          = nicntl-1
  integer, parameter :: qrm_pinth_          = nicntl-2

  integer, parameter :: qrm_amalgth_        = 1
  integer, parameter :: qrm_mem_relax_      = 2
  integer, parameter :: qrm_rd_eps_         = 3
  integer, parameter :: qrm_rweight_        = nrcntl

  integer, parameter :: qrm_auto_           = 0
  integer, parameter :: qrm_natural_        = 1
  integer, parameter :: qrm_given_          = 2
  integer, parameter :: qrm_colamd_         = 3
  integer, parameter :: qrm_metis_          = 4
  integer, parameter :: qrm_scotch_         = 5

  integer, parameter :: qrm_e_facto_flops_  = 1
  integer, parameter :: qrm_e_nnz_r_        = 2
  integer, parameter :: qrm_e_nnz_h_        = 3
  integer, parameter :: qrm_facto_flops_    = 4
  integer, parameter :: qrm_nnz_r_          = 5
  integer, parameter :: qrm_nnz_h_          = 6
  integer, parameter :: qrm_e_facto_mempeak_= 7
  integer, parameter :: qrm_rd_num_         = 8

  integer, parameter :: qrm_no_             = 0
  integer, parameter :: qrm_yes_            = 1

  integer, parameter :: qrm_allop_          = 0
  integer, parameter :: qrm_analyse_        = 1
  integer, parameter :: qrm_factorize_      = 2
  integer, parameter :: qrm_solve_          = 3
  integer, parameter :: qrm_apply_          = 4

  integer, save      :: qrm_default_ordering_ = qrm_noinit_ 
  integer, save      :: qrm_default_minamalg_ = qrm_noinit_ 
  integer, save      :: qrm_default_mb_       = qrm_noinit_ 
  integer, save      :: qrm_default_nb_       = qrm_noinit_ 
  integer, save      :: qrm_default_ib_       = qrm_noinit_ 
  integer, save      :: qrm_default_bh_       = qrm_noinit_ 
  integer, save      :: qrm_default_keeph_    = qrm_noinit_ 
  integer, save      :: qrm_default_rhsnb_    = qrm_noinit_ 
  integer, save      :: qrm_default_nlz_      = qrm_noinit_ 
  integer, save      :: qrm_default_cnode_    = qrm_noinit_ 
  integer, save      :: qrm_default_sing_     = qrm_noinit_ 
  integer, save      :: qrm_default_pinth_    = qrm_noinit_ 
  integer, save      :: qrm_default_ncpu_     = qrm_noinit_ 
  integer, save      :: qrm_default_ngpu_     = qrm_noinit_ 

  real(r32), save    :: qrm_default_amalgth_
  real(r32), save    :: qrm_default_rweight_
  real(r32), save    :: qrm_default_mem_relax_
  real(r32), save    :: qrm_default_rd_eps_


  interface qrm_parameters_reset
     subroutine qrm_parameters_reset()
     end subroutine qrm_parameters_reset
  end interface qrm_parameters_reset
  
  interface qrm_set
     subroutine qrm_glob_set_i4(string, ival, info)
       character(len=*)                        :: string
       integer                                 :: ival
       integer, optional                       :: info
     end subroutine qrm_glob_set_i4
     subroutine qrm_glob_set_r4(string, rval, info)
       import r32
       character(len=*)                        :: string
       real(r32)                               :: rval
       integer, optional                       :: info
     end subroutine qrm_glob_set_r4
  end interface qrm_set

  interface qrm_get
     subroutine qrm_glob_get_i4(string, ival, info)
       character(len=*)                        :: string
       integer                                 :: ival
       integer, optional                       :: info
     end subroutine qrm_glob_get_i4
     subroutine qrm_glob_get_i8(string, iival, info)
       import i64
       character(len=*)                        :: string
       integer(i64)                            :: iival
       integer, optional                       :: info
     end subroutine qrm_glob_get_i8
     subroutine qrm_glob_get_r4(string, rval, info)
       import r32
       character(len=*)                        :: string
       real(r32)                               :: rval
       integer, optional                       :: info
     end subroutine qrm_glob_get_r4
  end interface qrm_get

end module qrm_parameters_mod
