!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


! subroutine qrm_parameters_reset()
!   use qrm_parameters_mod, savesym => qrm_parameters_reset
!   implicit none


!   qrm_ncpu_              = 1
!   qrm_ngpu_              = 0
!   qrm_default_nb_        = 256
!   qrm_default_ib_        = 32
!   qrm_default_bh_        = -1
!   qrm_default_keeph_     = qrm_yes_
!   qrm_default_rhsnb_     = -1
!   qrm_default_nlz_       = 8
!   qrm_default_cnode_     = 1
!   qrm_default_sing_      = qrm_no_
!   qrm_default_pinth_     = 10000000

!   qrm_default_amalgth_   = 0.05
!   qrm_default_rweight_   = 0.001
!   qrm_default_mem_relax_ = -100.0
!   qrm_default_rd_eps_    = 0.e0

!   return
! end subroutine qrm_parameters_reset

subroutine qrm_glob_set_i4(string, ival, info)
  use qrm_string_mod
  use qrm_error_mod
  use qrm_memhandling_mod
  use qrm_parameters_mod, savesym=> qrm_glob_set_i4
  implicit none

  character(len=*)                        :: string
  integer                                 :: ival
  integer, optional                       :: info

  character(len=len(string))              :: istring

  ! error management
  integer                                 :: err
  character(len=*), parameter             :: name='qrm_glob_set_i4'

  err = 0

  istring = qrm_str_tolower(string)

  if(index(istring,'qrm_eunit') .eq. 1) then
     qrm_eunit = ival
  else if(index(istring,'qrm_print_etree') .eq. 1) then
     qrm_print_etree = ival
  else if(index(istring,'qrm_ounit') .eq. 1) then
     qrm_ounit = ival
  else if(index(istring,'qrm_dunit') .eq. 1) then
     qrm_dunit = ival
  else if(index(istring,'qrm_ncpu') .eq. 1) then
     qrm_default_ncpu_ = ival
  else if(index(istring,'qrm_ngpu') .eq. 1) then
     qrm_default_ngpu_ = ival
  else if(index(istring,'qrm_max_mem') .eq. 1) then
     qrm_max_mem = ival
  else if(index(istring,'qrm_tot_mem') .eq. 1) then
     qrm_tot_mem = ival
  else if(index(istring,'qrm_ordering') .eq. 1) then
     qrm_default_ordering_ = ival
  else if(index(istring,'qrm_minamalg') .eq. 1) then
     qrm_default_minamalg_ = ival
  else if(index(istring,'qrm_mb') .eq. 1) then
     qrm_default_mb_ = ival
  else if(index(istring,'qrm_nb') .eq. 1) then
     qrm_default_nb_ = ival
  else if(index(istring,'qrm_ib') .eq. 1) then
     qrm_default_ib_ = ival
  else if(index(istring,'qrm_bh') .eq. 1) then
     qrm_default_bh_ = ival
  else if(index(istring,'qrm_keeph') .eq. 1) then
     qrm_default_keeph_ = ival
  else if(index(istring,'qrm_rhsnb') .eq. 1) then
     qrm_default_rhsnb_ = ival
  else if(index(istring,'qrm_nlz') .eq. 1) then
     qrm_default_nlz_ = ival
  else if(index(istring,'qrm_pinth') .eq. 1) then
     qrm_default_pinth_ = ival
  else
     err = 23
     call qrm_error_print(err, name, aed=string)
     goto 9999
  end if

9999 continue

  if(present(info)) info = err
  return
end subroutine qrm_glob_set_i4

subroutine qrm_glob_get_i4(string, ival, info)
  use qrm_string_mod
  use qrm_error_mod
  use qrm_parameters_mod, savesym=> qrm_glob_get_i4
  implicit none

  character(len=*)                        :: string
  integer                                 :: ival
  integer, optional                       :: info

  integer(kind=8)                         :: iival

  ! error management
  integer                                 :: err
  character(len=*), parameter             :: name='qrm_glob_get_i4'

  err = 0

  call qrm_glob_get_i8(string, iival, err)
  ival = iival

  if(present(info)) info = err
  return

end subroutine qrm_glob_get_i4

subroutine qrm_glob_get_i8(string, iival, info)
  use qrm_string_mod
  use qrm_error_mod
  use qrm_memhandling_mod
  use qrm_parameters_mod, savesym=> qrm_glob_get_i8
  implicit none

  character(len=*)                        :: string
  integer(kind=8)                         :: iival
  integer, optional                       :: info

  character(len=len(string))              :: istring

  ! error management
  integer                                 :: err
  character(len=*), parameter             :: name='qrm_glob_get_i4'

  err = 0

  istring = qrm_str_tolower(string)

  if(index(istring,'qrm_max_mem') .eq. 1) then
     iival = qrm_max_mem
  else if(index(istring,'qrm_tot_mem') .eq. 1) then
     iival = qrm_tot_mem
  else if(index(istring,'qrm_print_etree') .eq. 1) then
     iival = qrm_print_etree
  else if(index(istring,'qrm_ounit') .eq. 1) then
     iival = qrm_ounit
  else if(index(istring,'qrm_eunit') .eq. 1) then
     iival = qrm_eunit
  else if(index(istring,'qrm_dunit') .eq. 1) then
     iival = qrm_dunit
  else if(index(istring,'qrm_ncpu') .eq. 1) then
     iival = qrm_ncpu
  else if(index(istring,'qrm_ngpu') .eq. 1) then
     iival = qrm_ngpu
  else if(index(istring,'qrm_ordering') .eq. 1) then
     iival = qrm_default_ordering_
  else if(index(istring,'qrm_minamalg') .eq. 1) then
     iival = qrm_default_minamalg_
  else if(index(istring,'qrm_mb') .eq. 1) then
     iival = qrm_default_mb_
  else if(index(istring,'qrm_nb') .eq. 1) then
     iival = qrm_default_nb_
  else if(index(istring,'qrm_ib') .eq. 1) then
     iival = qrm_default_ib_
  else if(index(istring,'qrm_bh') .eq. 1) then
     iival = qrm_default_bh_
  else if(index(istring,'qrm_keeph') .eq. 1) then
     iival = qrm_default_keeph_
  else if(index(istring,'qrm_rhsnb') .eq. 1) then
     iival = qrm_default_rhsnb_
  else if(index(istring,'qrm_nlz') .eq. 1) then
     iival = qrm_default_nlz_
  else if(index(istring,'qrm_pinth') .eq. 1) then
     iival = qrm_default_pinth_
  else
     err = 23
     call qrm_error_print(err, name, aed=string)
     goto 9999
  end if

9999 continue
  if(present(info)) info = err
  return

end subroutine qrm_glob_get_i8


subroutine qrm_glob_set_r4(string, rval, info)
  use qrm_string_mod
  use qrm_error_mod
  use qrm_mem_mod
  use qrm_parameters_mod, savesym=> qrm_glob_set_r4
  implicit none

  character(len=*)                        :: string
  real(r32)                               :: rval
  integer, optional                       :: info

  character(len=len(string))              :: istring

  ! error management
  integer                                 :: err
  character(len=*), parameter             :: name='qrm_glob_set_r4'

  err = 0

  istring = qrm_str_tolower(string)

  if(index(istring,'qrm_amalgth') .eq. 1) then
     qrm_default_amalgth_ = rval
  else if(index(istring,'qrm_rweight') .eq. 1) then
     qrm_default_rweight_ = rval
  else if(index(istring,'qrm_mem_relax') .eq. 1) then
     qrm_default_mem_relax_ = rval
  else if(index(istring,'qrm_rd_eps') .eq. 1) then
     qrm_default_rd_eps_ = rval
  else
     err = 23
     call qrm_error_print(err, name, aed=string)
     goto 9999
  end if

9999 continue
  if(present(info)) info = err
  return

end subroutine qrm_glob_set_r4

subroutine qrm_glob_get_r4(string, rval, info)
  use qrm_string_mod
  use qrm_error_mod
  use qrm_mem_mod
  use qrm_parameters_mod, savesym=> qrm_glob_get_r4
  implicit none

  character(len=*)                        :: string
  real(r32)                               :: rval
  integer, optional                       :: info

  character(len=len(string))              :: istring

  ! error management
  integer                                 :: err
  character(len=*), parameter             :: name='qrm_glob_get_r4'

  err = 0

  istring = qrm_str_tolower(string)

  if(index(istring,'qrm_amalgth') .eq. 1) then
     rval = qrm_default_amalgth_
  else if(index(istring,'qrm_rweight') .eq. 1) then
     rval = qrm_default_rweight_
  else if(index(istring,'qrm_mem_relax') .eq. 1) then
     rval = qrm_default_mem_relax_
  else if(index(istring,'qrm_rd_eps') .eq. 1) then
     rval = qrm_default_rd_eps_
  else
     err = 23
     call qrm_error_print(err, name, aed=string)
     goto 9999
  end if

9999 continue
  if(present(info)) info = err
  return

end subroutine qrm_glob_get_r4


subroutine qrm_parameters_reset()
  use qrm_parameters_mod, savesym => qrm_parameters_reset
  implicit none

  character(LEN=20) :: str
  
  integer           :: ierr, len
  
  call get_environment_variable(name="QRM_PRINT_ETREE", value=str, length=len, status=ierr)
  if(qrm_print_etree.eq.qrm_noinit_) then
     if(ierr .eq. 0) then
        read(str(1:len),*)qrm_print_etree
        qrm_print_etree = min(qrm_print_etree,2)
     end if
  end if
  
  call get_environment_variable(name="QRM_OUNIT", value=str, length=len, status=ierr)
  if(qrm_ounit.eq.qrm_noinit_) then
     if(ierr .ge. 1) then
        qrm_ounit  = 6
     else
        read(str(1:len),*)qrm_ounit
     end if
  end if
  
  call get_environment_variable(name="QRM_EUNIT", value=str, length=len, status=ierr)
  if(qrm_eunit.eq.qrm_noinit_) then
     if(ierr .ge. 1) then
        qrm_eunit  = 1
     else
        read(str(1:len),*)qrm_eunit
     end if
  end if

  call get_environment_variable(name="QRM_DUNIT", value=str, length=len, status=ierr)
  if(qrm_dunit.eq.qrm_noinit_) then
     if(ierr .ge. 1) then
        qrm_dunit  = 0
     else
        read(str(1:len),*)qrm_dunit
     end if
  end if

#if defined (have_starpu)
  if(qrm_default_ncpu_ .eq. qrm_noinit_) then
     call get_environment_variable(name="QRM_NCPU", value=str, length=len, status=ierr)
     if(ierr .ge. 1) then
        qrm_default_ncpu_  = 1
     else
        read(str(1:len),*)qrm_default_ncpu_
     end if
  end if
#else
  qrm_default_ncpu_  = 1
#endif
  
#if defined (have_starpu) && defined (have_cuda)
  if(qrm_default_ngpu_ .eq. qrm_noinit_) then
     call get_environment_variable(name="QRM_NGPU", value=str, length=len, status=ierr)
     if(ierr .ge. 1) then
        qrm_default_ngpu_  = 0
     else
        read(str(1:len),*)qrm_default_ngpu_
     end if
  end if
#else
  qrm_default_ngpu_  = 0
#endif
  
  call get_environment_variable(name="QRM_ORDERING", value=str, length=len, status=ierr)
  if(qrm_default_ordering_.eq.qrm_noinit_) then
     if(ierr .ge. 1) then
        qrm_default_ordering_  = qrm_auto_
     else
        read(str(1:len),*)qrm_default_ordering_
     end if
  end if
  
  call get_environment_variable(name="QRM_MINAMALG", value=str, length=len, status=ierr)
  if(qrm_default_minamalg_.eq.qrm_noinit_) then
     if(ierr .ge. 1) then
        qrm_default_minamalg_  = 4
     else
        read(str(1:len),*)qrm_default_minamalg_
     end if
  end if
  
  call get_environment_variable(name="QRM_MB", value=str, length=len, status=ierr)
  if(qrm_default_mb_.eq.qrm_noinit_) then
     if(ierr .ge. 1) then
        qrm_default_mb_  = 256
     else
        read(str(1:len),*)qrm_default_mb_
     end if
  end if
  
  call get_environment_variable(name="QRM_NB", value=str, length=len, status=ierr)
  if(qrm_default_nb_.eq.qrm_noinit_) then
     if(ierr .eq. 0) then
        read(str(1:len),*)qrm_default_nb_
     end if
  end if

  call get_environment_variable(name="QRM_IB", value=str, length=len, status=ierr)
  if(qrm_default_ib_.eq.qrm_noinit_) then
     if(ierr .ge. 1) then
        qrm_default_ib_  = 32
     else
        read(str(1:len),*)qrm_default_ib_
     end if
  end if

  call get_environment_variable(name="QRM_BH", value=str, length=len, status=ierr)
  if(qrm_default_bh_.eq.qrm_noinit_) then
     if(ierr .ge. 1) then
        qrm_default_bh_  = -1
     else
        read(str(1:len),*)qrm_default_bh_
     end if
  end if

  call get_environment_variable(name="QRM_KEEPH", value=str, length=len, status=ierr)
  if(qrm_default_keeph_.eq.qrm_noinit_) then
     if(ierr .ge. 1) then
        qrm_default_keeph_  = qrm_yes_
     else
        read(str(1:len),*)qrm_default_keeph_
     end if
  end if

  call get_environment_variable(name="QRM_RHSNB", value=str, length=len, status=ierr)
  if(qrm_default_rhsnb_.eq.qrm_noinit_) then
     if(ierr .ge. 1) then
        qrm_default_rhsnb_  = -1
     else
        read(str(1:len),*)qrm_default_rhsnb_
     end if
  end if

  call get_environment_variable(name="QRM_NLZ", value=str, length=len, status=ierr)
  if(qrm_default_nlz_.eq.qrm_noinit_) then
     if(ierr .ge. 1) then
        qrm_default_nlz_  = 8
     else
        read(str(1:len),*)qrm_default_nlz_
     end if
  end if

  call get_environment_variable(name="QRM_CNODE", value=str, length=len, status=ierr)
  if(qrm_default_cnode_.eq.qrm_noinit_) then
     if(ierr .ge. 1) then
        qrm_default_cnode_  = 1
     else
        read(str(1:len),*)qrm_default_cnode_
     end if
  end if

  call get_environment_variable(name="QRM_SING", value=str, length=len, status=ierr)
  if(qrm_default_sing_.eq.qrm_noinit_) then
     if(ierr .ge. 1) then
        qrm_default_sing_  = qrm_no_
     else
        read(str(1:len),*)qrm_default_sing_
     end if
  end if
  call get_environment_variable(name="QRM_PINTH", value=str, length=len, status=ierr)
  if(qrm_default_pinth_.eq.qrm_noinit_) then
     if(ierr .ge. 1) then
        qrm_default_pinth_  = 10000000
     else
        read(str(1:len),*)qrm_default_pinth_
     end if
  end if

  call get_environment_variable(name="QRM_AMALGTH", value=str, length=len, status=ierr)
  if(ierr .ge. 1) then
       qrm_default_amalgth_  = 0.05_r32
    else
     read(str(1:len),*)qrm_default_amalgth_
  end if

  call get_environment_variable(name="QRM_RWEIGHT", value=str, length=len, status=ierr)
  if(ierr .ge. 1) then
       qrm_default_rweight_  = 0.001_r32
    else
     read(str(1:len),*)qrm_default_rweight_
  end if

  call get_environment_variable(name="QRM_MEM_RELAX", value=str, length=len, status=ierr)
  if(ierr .ge. 1) then
       qrm_default_mem_relax_  = -100.0_r32
    else
     read(str(1:len),*)qrm_default_mem_relax_
  end if

  call get_environment_variable(name="QRM_RD_EPS", value=str, length=len, status=ierr)
  if(ierr .ge. 1) then
       qrm_default_rd_eps_  = 0.0_r32
    else
     read(str(1:len),*)qrm_default_rd_eps_
  end if

  return
end subroutine qrm_parameters_reset
