!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!!
!! This file contains a module with all the generic interfaces for the typed methods routines.
!!
!! ##############################################################################################


!> This module contains generic methods
module sqrm_methods_mod

  interface qrm_backslash
     procedure               ::  sqrm_spmat_backslash2d, sqrm_spmat_backslash1d
     procedure               ::  sqrm_spfct_backslash2d, sqrm_spfct_backslash1d
  end interface qrm_backslash
  
  interface qrm_spbackslash
     procedure               ::  sqrm_spmat_backslash2d, sqrm_spmat_backslash1d
  end interface qrm_spbackslash

  interface qrm_spmat_backslash
     procedure               ::  sqrm_spmat_backslash2d, sqrm_spmat_backslash1d
  end interface qrm_spmat_backslash

  interface sqrm_spmat_backslash
     subroutine sqrm_spmat_backslash2d(qrm_mat, b, x, transp, cperm, info)
       use sqrm_spmat_mod
       type(sqrm_spmat_type) :: qrm_mat
       real(r32)          :: b(:,:), x(:,:)
       integer, optional     :: cperm(:)
       character, optional   :: transp
       integer, optional     :: info
     end subroutine sqrm_spmat_backslash2d
     subroutine sqrm_spmat_backslash1d(qrm_mat, b, x, transp, cperm, info)
       use sqrm_spmat_mod
       type(sqrm_spmat_type) :: qrm_mat
       real(r32)          :: b(:), x(:)
       character, optional   :: transp
       integer, optional     :: cperm(:)
       integer, optional     :: info
     end subroutine sqrm_spmat_backslash1d
  end interface sqrm_spmat_backslash

  interface qrm_spfct_backslash
     procedure               ::  sqrm_spfct_backslash2d, sqrm_spfct_backslash1d
  end interface qrm_spfct_backslash

  interface sqrm_spfct_backslash
     subroutine sqrm_spfct_backslash2d(qrm_mat, b, x, transp, info)
       use sqrm_spfct_mod
       type(sqrm_spfct_type) :: qrm_mat
       real(r32)          :: b(:,:), x(:,:)
       character, optional   :: transp
       integer, optional     :: info
     end subroutine sqrm_spfct_backslash2d
     subroutine sqrm_spfct_backslash1d(qrm_mat, b, x, transp, info)
       use sqrm_spfct_mod
       type(sqrm_spfct_type) :: qrm_mat
       real(r32)          :: b(:), x(:)
       character, optional   :: transp
       integer, optional     :: info
     end subroutine sqrm_spfct_backslash1d
  end interface sqrm_spfct_backslash
  
  
  interface qrm_spposv
     procedure               ::  sqrm_spmat_posv2d, sqrm_spmat_posv1d
  end interface qrm_spposv

  interface qrm_spmat_posv
     procedure               ::  sqrm_spmat_posv2d, sqrm_spmat_posv1d
  end interface qrm_spmat_posv

  interface sqrm_spmat_posv
     subroutine sqrm_spmat_posv2d(qrm_mat, b, x, cperm, info)
       use sqrm_spmat_mod
       type(sqrm_spmat_type) :: qrm_mat
       real(r32)          :: b(:,:), x(:,:)
       integer, optional     :: cperm(:)
       integer, optional     :: info
     end subroutine sqrm_spmat_posv2d
     subroutine sqrm_spmat_posv1d(qrm_mat, b, x, cperm, info)
       use sqrm_spmat_mod
       type(sqrm_spmat_type) :: qrm_mat
       real(r32)          :: b(:), x(:)
       integer, optional     :: cperm(:)
       integer, optional     :: info
     end subroutine sqrm_spmat_posv1d
  end interface sqrm_spmat_posv

  interface qrm_spfct_potrs
     procedure               ::  sqrm_spfct_potrs2d, sqrm_spfct_potrs1d
  end interface qrm_spfct_potrs

  interface sqrm_spfct_potrs
     subroutine sqrm_spfct_potrs2d(qrm_mat, b, x, info)
       use sqrm_spfct_mod
       type(sqrm_spfct_type) :: qrm_mat
       real(r32)          :: b(:,:), x(:,:)
       integer, optional     :: info
     end subroutine sqrm_spfct_potrs2d
     subroutine sqrm_spfct_potrs1d(qrm_mat, b, x, info)
       use sqrm_spfct_mod
       type(sqrm_spfct_type) :: qrm_mat
       real(r32)          :: b(:), x(:)
       integer, optional     :: info
     end subroutine sqrm_spfct_potrs1d
  end interface sqrm_spfct_potrs

  interface qrm_least_squares
     procedure               ::  sqrm_spmat_gels2d, sqrm_spmat_gels1d
  end interface qrm_least_squares

  interface qrm_min_norm
     procedure               :: sqrm_spmat_gels2d, sqrm_spmat_gels1d
  end interface qrm_min_norm

  interface qrm_spmat_gels
     procedure               :: sqrm_spmat_gels2d, sqrm_spmat_gels1d
  end interface qrm_spmat_gels

  interface sqrm_spmat_gels
     subroutine sqrm_spmat_gels2d(qrm_mat, b, x, transp, cperm, info)
       use sqrm_spmat_mod
       type(sqrm_spmat_type) :: qrm_mat
       real(r32)          :: b(:,:), x(:,:)
       character, optional   :: transp
       integer, optional     :: cperm(:)
       integer, optional     :: info
     end subroutine sqrm_spmat_gels2d
     subroutine sqrm_spmat_gels1d(qrm_mat, b, x, transp, cperm, info)
       use sqrm_spmat_mod
       type(sqrm_spmat_type) :: qrm_mat
       real(r32)          :: b(:), x(:)
       character, optional   :: transp
       integer, optional     :: cperm(:)
       integer, optional     :: info
     end subroutine sqrm_spmat_gels1d
  end interface sqrm_spmat_gels
  
  interface qrm_spfct_geqrs
     procedure               :: sqrm_spfct_geqrs2d, sqrm_spfct_geqrs1d
  end interface qrm_spfct_geqrs

  interface sqrm_spfct_geqrs
     subroutine sqrm_spfct_geqrs2d(qrm_mat, b, x, transp, info)
       use sqrm_spfct_mod
       type(sqrm_spfct_type) :: qrm_mat
       real(r32)          :: b(:,:), x(:,:)
       character, optional   :: transp
       integer, optional     :: info
     end subroutine sqrm_spfct_geqrs2d
     subroutine sqrm_spfct_geqrs1d(qrm_mat, b, x, transp, info)
       use sqrm_spfct_mod
       type(sqrm_spfct_type) :: qrm_mat
       real(r32)          :: b(:), x(:)
       character, optional   :: transp
       integer, optional     :: info
     end subroutine sqrm_spfct_geqrs1d
  end interface sqrm_spfct_geqrs
  

  
  interface qrm_residual_norm
     procedure               :: sqrm_residual_norm2d, sqrm_residual_norm1d
  end interface qrm_residual_norm

  interface sqrm_residual_norm
     subroutine sqrm_residual_norm2d(qrm_mat, b, x, nrm, transp, info)
       use sqrm_spmat_mod
       real(r32)             :: nrm(:)
       type(sqrm_spmat_type) :: qrm_mat
       real(r32)          :: b(:,:), x(:,:)
       character, optional   :: transp
       integer, optional     :: info
     end subroutine sqrm_residual_norm2d
     subroutine sqrm_residual_norm1d(qrm_mat, b, x, nrm, transp, info)
       use sqrm_spmat_mod
       real(r32)             :: nrm
       type(sqrm_spmat_type) :: qrm_mat
       real(r32)          :: b(:), x(:)
       character, optional   :: transp
       integer, optional     :: info
     end subroutine sqrm_residual_norm1d
  end interface sqrm_residual_norm



  interface qrm_residual_orth
     procedure               :: sqrm_residual_orth2d, sqrm_residual_orth1d
     procedure               :: sqrm_residual_and_orth2d, sqrm_residual_and_orth1d
  end interface qrm_residual_orth

  interface sqrm_residual_orth
     subroutine sqrm_residual_orth2d(qrm_mat, r, nrm, transp, info)
       use sqrm_spmat_mod
       real(r32)             :: nrm(:)
       type(sqrm_spmat_type) :: qrm_mat
       real(r32)          :: r(:,:)
       character, optional   :: transp
       integer, optional     :: info
     end subroutine sqrm_residual_orth2d
     subroutine sqrm_residual_orth1d(qrm_mat, r, nrm, transp, info)
       use sqrm_spmat_mod
       real(r32)             :: nrm
       type(sqrm_spmat_type) :: qrm_mat
       real(r32)          :: r(:)
       character, optional   :: transp
       integer, optional     :: info
     end subroutine sqrm_residual_orth1d
     subroutine sqrm_residual_and_orth2d(qrm_mat, b, x, nrm, transp, info)
       use sqrm_spmat_mod
       real(r32)             :: nrm(:)
       type(sqrm_spmat_type) :: qrm_mat
       real(r32)          :: b(:,:), x(:,:)
       character, optional   :: transp
       integer, optional     :: info
     end subroutine sqrm_residual_and_orth2d
     subroutine sqrm_residual_and_orth1d(qrm_mat, b, x, nrm, transp, info)
       use sqrm_spmat_mod
       real(r32)             :: nrm
       type(sqrm_spmat_type) :: qrm_mat
       real(r32)          :: b(:), x(:)
       character, optional   :: transp
       integer, optional     :: info
     end subroutine sqrm_residual_and_orth1d
  end interface sqrm_residual_orth


  interface qrm_get_r
     subroutine sqrm_get_r(qrm_spfct, r, info)
       use sqrm_spmat_mod
       use sqrm_spfct_mod
       type(sqrm_spfct_type) :: qrm_spfct
       type(sqrm_spmat_type) :: r
       integer, optional     :: info
     end subroutine sqrm_get_r
  end interface qrm_get_r
  
end module sqrm_methods_mod
