!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_solve_tasks.F90
!! This file contains all the kernels for the solve phase
!!
!!
!! ##############################################################################################

#include "qrm_common.h"

!==========================================================================================
!==========================================================================================
subroutine zqrm_front_qt(front, qrm_sdata_b, work)
  use zqrm_fdata_mod
  use zqrm_sdata_mod
  implicit none

  type(zqrm_front_type)           :: front
  type(zqrm_ws_type)              :: work
  type(zqrm_sdata_type)           :: qrm_sdata_b

  integer                         :: lt, ft, lct, nsteps, s, bi, bk, bl, bh
  integer                         :: m, n, k, ib, l, node, info, ii, jj, i, j, mb
  integer                         :: nbr, nbc, nbe, first
  
  ! error management
  character(len=*), parameter     :: name='front_qt'
  ! shortcut
  if (min(front%m, front%n) .le. 0) goto 9999

  node = front%num
  
  n  = size(qrm_sdata_b%p,2)
  ! write(*,'(" =--> Apply Q''  : ",i4)')front%num

  call zqrm_dsmat_gemqr_async(qrm_seq_dscr_, qrm_conj_transp, &
       front%f, front%t, &
       qrm_sdata_b%front_rhs(node)%rhs, &
       front%ib, front%bh, work)
  
  ! scatter front%b into qrm_sdata_b
  mb = qrm_sdata_b%front_rhs(node)%rhs%mb
  do i=1, front%npiv
     bi = (i-1)/front%mb+1
     ii = i - (bi-1)*front%mb
     do j=1, size(qrm_sdata_b%front_rhs(node)%rhs%blocks,2)
        qrm_sdata_b%p(front%rows(i),(j-1)*mb+1:min(n,j*mb)) =  &
             qrm_sdata_b%front_rhs(node)%rhs%blocks(bi,j)%c(ii,:)
     end do
  end do

  do i=front%ne+1, front%m
     bi = (i-1)/front%mb+1
     ii = i - (bi-1)*front%mb
     do j=1, size(qrm_sdata_b%front_rhs(node)%rhs%blocks,2)
        qrm_sdata_b%p(front%rows(i),(j-1)*mb+1:min(n,j*mb)) =  &
             qrm_sdata_b%front_rhs(node)%rhs%blocks(bi,j)%c(ii,:)
     end do
  end do

9999 continue

  return 

end subroutine zqrm_front_qt


subroutine zqrm_assemble_qt(qrm_fct, front, qrm_sdata_b, info)
  use zqrm_spfct_mod
  use qrm_mem_mod
  use qrm_error_mod
  use zqrm_sdata_mod
  use zqrm_fdata_mod
  use zqrm_dsmat_mod
  implicit none

  type(zqrm_spfct_type), target  :: qrm_fct
  type(zqrm_front_type)          :: front
  type(zqrm_sdata_type)          :: qrm_sdata_b
  integer, optional              :: info

  type(qrm_adata_type), pointer  :: adata
  type(zqrm_front_type), pointer :: child
  integer                        :: n, c, node, i, ii, bi, j, mb

  ! error management
  integer                        :: err
  character(len=*), parameter    :: name='qrm_assemble_qt'

  err = 0

  node = front%num
  n = size(qrm_sdata_b%p,2)
  
  adata => qrm_fct%adata
  if(front%ne .gt. 0) then
     ! note that n cannot e greater than front%mb otherwise we have
     ! more than one block-column in the rhs
     call zqrm_dsmat_init(qrm_sdata_b%front_rhs(node)%rhs, &
          front%m, n, front%mb, front%mb, &
          info=err, seq=.true.)
     __QRM_INFO_CHECK(err, name,'qrm_dsmat_init',9999)

     mb = qrm_sdata_b%front_rhs(node)%rhs%mb
     do i=1, front%anrows
        bi = (front%arowmap(i)-1)/front%mb+1
        ii = front%arowmap(i) - (bi-1)*front%mb
        do j=1, size(qrm_sdata_b%front_rhs(node)%rhs%blocks,2)
           qrm_sdata_b%front_rhs(node)%rhs%blocks(bi,j)%c(ii,:) = &
                qrm_sdata_b%p(front%rows(front%arowmap(i)),(j-1)*mb+1:min(n,j*mb))
        end do
     end do
  end if

  do c = adata%childptr(node), adata%childptr(node+1)-1
     child => qrm_fct%fdata%front_list(adata%child(c))

     ! the double condition below should be redundant but you never know...
     if ((child%ne.gt.child%npiv) .and. (front%ne.gt.0)) &
          call zqrm_dsmat_extadd_async(qrm_seq_dscr_,    &
          qrm_sdata_b%front_rhs(child%num)%rhs,          &
          qrm_sdata_b%front_rhs(node)%rhs,               &
          i=child%npiv+1, j=1,                           &
          m=child%ne-child%npiv, n=n, l=0,               &
          inout='o', op='c',                             &
          rowmap = child%rowmap)
          ! call zqrm_dsmat_copy_async(qrm_seq_dscr_,      &
          ! qrm_sdata_b%front_rhs(child%num)%rhs,          &
          ! qrm_sdata_b%front_rhs(node)%rhs,               &
          ! i=child%npiv+1, j=1,                           &
          ! m=child%ne-child%npiv, n=n,                    &
          ! rowmap = child%rowmap)

     call zqrm_dsmat_destroy(qrm_sdata_b%front_rhs(child%num)%rhs, seq=.true.)
  end do


9999 continue

  if(present(info)) info = err
  return

end subroutine zqrm_assemble_qt




!==========================================================================================
!==========================================================================================
subroutine zqrm_front_r(front, qrm_sdata_b, qrm_sdata_x)
 
  use zqrm_fdata_mod
  use zqrm_sdata_mod
  use qrm_parameters_mod
  implicit none
 
  type(zqrm_front_type)           :: front
  type(zqrm_sdata_type)           :: qrm_sdata_b, qrm_sdata_x
 
  integer                         :: bi, bj, bk, npi, npj
  integer                         :: m, n, k, i, j, node, info
 
  ! shortcut
  if (min(front%m, front%n) .le.0) goto 9999
  if (front%npiv .le. 0) goto 9999
 
  node = front%num
  
  ! write(*,'(" =--> Solve R  : ",i4)') front%num
 

  call zqrm_dsmat_trsm_async(qrm_seq_dscr_, &
       qrm_left, qrm_upper,                 &
       qrm_no_transp, qrm_no_unit,          &
       qrm_zone, front%f,                   &
       qrm_sdata_x%front_rhs(node)%rhs,     &
       m=front%npiv)

9999 continue 
  return
 
end subroutine zqrm_front_r
 
 
subroutine zqrm_assemble_r(qrm_fct, front, qrm_sdata_b, qrm_sdata_x, info)
  use zqrm_spfct_mod
  use qrm_mem_mod
  use qrm_error_mod
  use zqrm_fdata_mod
  use zqrm_sdata_mod
  implicit none
 
  type(zqrm_spfct_type), target  :: qrm_fct
  type(zqrm_front_type)          :: front
  type(zqrm_sdata_type)          :: qrm_sdata_b, qrm_sdata_x
  integer, optional              :: info
 
  type(qrm_adata_type), pointer  :: adata
  type(zqrm_front_type), pointer :: child
  integer                        :: n, c, node, i, j, mb, ii, bi
 
  ! error management
  integer                        :: err
  character(len=*), parameter    :: name='qrm_assemble_r'
 
  err = 0
 
  adata => qrm_fct%adata
 
  node = front%num
  n = size(qrm_sdata_b%p,2)

  
  do c = adata%childptr(node), adata%childptr(node+1)-1
     child => qrm_fct%fdata%front_list(adata%child(c))

     call zqrm_dsmat_init(qrm_sdata_x%front_rhs(child%num)%rhs, &
          child%n, n, child%mb, child%mb,                       &
          info=err, seq=.true.)
     __QRM_INFO_CHECK(err, name,'qrm_dsmat_init',9999)

     mb = qrm_sdata_x%front_rhs(child%num)%rhs%mb

     do i=1, child%npiv
        bi = (i-1)/child%mb+1
        ii = i - (bi-1)*child%mb
        do j=1, size(qrm_sdata_x%front_rhs(child%num)%rhs%blocks,2)
           qrm_sdata_x%front_rhs(child%num)%rhs%blocks(bi,j)%c(ii,:) = &
                qrm_sdata_b%p(child%rows(i),(j-1)*mb+1:min(n,j*mb))
        end do
     end do
        
     if(child%npiv.lt.child%n) then
        call zqrm_dsmat_extadd_async(qrm_seq_dscr_,  &
             qrm_sdata_x%front_rhs(child%num)%rhs,   &
             qrm_sdata_x%front_rhs(node)%rhs,        &
             i=child%npiv+1, j=1,                    &
             m=child%n-child%npiv, n=n, l=0,         &
             inout='i', op='c',                      &
             rowmap = child%colmap)
        ! call zqrm_dsmat_copy_async(qrm_seq_dscr_,  &
             ! qrm_sdata_x%front_rhs(child%num)%rhs, &
             ! qrm_sdata_x%front_rhs(node)%rhs,      &
             ! i=child%npiv+1, j=1,                  &
             ! m=child%n-child%npiv, n=n,            &
             ! rowmap = child%colmap, inout='i')
        
     end if
 
  end do
 
  if(front%ne .le. 0) goto 9999
 
  ! scatter front%rhs into x
  mb = qrm_sdata_x%front_rhs(node)%rhs%mb
  do i=1, front%npiv
     bi = (i-1)/front%mb+1
     ii = i - (bi-1)*front%mb
     do j=1, size(qrm_sdata_x%front_rhs(node)%rhs%blocks,2)
        qrm_sdata_x%p(front%cols(i),(j-1)*mb+1:min(n,j*mb)) = &
             qrm_sdata_x%front_rhs(node)%rhs%blocks(bi,j)%c(ii,:)
     end do
  end do
     
  call zqrm_dsmat_destroy(qrm_sdata_x%front_rhs(node)%rhs, seq=.true.)

9999 continue
 
  if(present(info)) info = err
  return
 
end subroutine zqrm_assemble_r
 
 
 
!==========================================================================================
!==========================================================================================
subroutine zqrm_front_q(front, qrm_sdata_b, work)
  use zqrm_utils_mod
  use zqrm_fdata_mod
  use qrm_mem_mod
  use qrm_common_mod
  use zqrm_sdata_mod
  implicit none
 
  type(zqrm_front_type)           :: front
  type(zqrm_sdata_type)           :: qrm_sdata_b
  type(zqrm_ws_type)              :: work
 
  integer                         :: nsteps, s, bi, bk, bl, nbr, nbc, nbe, first
  integer                         :: m, n, k, f, ib, l, node, info, i
  
  ! write(*,'(" =--> Apply Q  : ",i4)')front%num
 
  ! shortcut
  if (min(front%m, front%n) .le. 0) goto 9999
 
  node = front%num
  n = size(qrm_sdata_b%p,2)

  call zqrm_dsmat_gemqr_async(qrm_seq_dscr_, qrm_no_transp, &
       front%f, front%t, qrm_sdata_b%front_rhs(node)%rhs, front%ib, front%bh, &
       work)
  
9999 continue
  return 
 
end subroutine zqrm_front_q
 
subroutine zqrm_assemble_q(qrm_fct, front, qrm_sdata_b, info)
  use zqrm_spfct_mod
  use qrm_mem_mod
  use qrm_error_mod
  use qrm_common_mod
  use zqrm_sdata_mod
  use zqrm_dsmat_mod
  implicit none
 
  type(zqrm_spfct_type), target  :: qrm_fct
  type(zqrm_front_type)          :: front
  type(zqrm_sdata_type)          :: qrm_sdata_b
  integer, optional              :: info
 
  type(qrm_adata_type), pointer  :: adata
  type(zqrm_front_type), pointer :: child
  integer                        :: n, c, node, i, ii, bi, mb, j
 
  ! error management
  integer                        :: err
  character(len=*), parameter    :: name='qrm_assemble_q'
 
  err = 0
 
  adata => qrm_fct%adata
 
  n = size(qrm_sdata_b%p,2)
  node = front%num
  
  ! write(*,'(" =--> Assemble Q  : ",i4)') front%num

  do c = adata%childptr(front%num), adata%childptr(front%num+1)-1
     child => qrm_fct%fdata%front_list(adata%child(c))

     if(child%ne.le.0) cycle
     
     call zqrm_dsmat_init(qrm_sdata_b%front_rhs(child%num)%rhs, &
          m=child%m, n=n, mb=child%mb, nb=child%mb,             &
          info=err, seq=.true.)
     __QRM_INFO_CHECK(err, name,'qrm_dsmat_init',9999)
     
     mb = qrm_sdata_b%front_rhs(child%num)%rhs%mb
     do i=1, child%npiv
        bi = (i-1)/child%mb+1
        ii = i - (bi-1)*child%mb
        do j=1, size(qrm_sdata_b%front_rhs(child%num)%rhs%blocks,2)
           qrm_sdata_b%front_rhs(child%num)%rhs%blocks(bi,j)%c(ii,:) = &
                qrm_sdata_b%p(child%rows(i),(j-1)*mb+1:min(n,j*mb))
        end do
     end do

     do i=child%ne+1, child%m
        bi = (i-1)/child%mb+1
        ii = i - (bi-1)*child%mb
        do j=1, size(qrm_sdata_b%front_rhs(child%num)%rhs%blocks,2)
           qrm_sdata_b%front_rhs(child%num)%rhs%blocks(bi,j)%c(ii,:) = &
                qrm_sdata_b%p(child%rows(i),(j-1)*mb+1:min(n,j*mb))
        end do
     end do

     if(child%ne.gt.child%npiv)                                                 &
          call zqrm_dsmat_extadd_async(qrm_seq_dscr_,                           &
          qrm_sdata_b%front_rhs(child%num)%rhs,                                 &
          qrm_sdata_b%front_rhs(node)%rhs,                                      &
          i=child%npiv+1, j=1,                                                  &
          m=child%ne-child%npiv, n=qrm_sdata_b%front_rhs(child%num)%rhs%n, l=0, &
          inout='i', op='c',                                                    &
          rowmap=child%rowmap)
          ! call zqrm_dsmat_copy_async(qrm_seq_dscr_, &
          ! qrm_sdata_b%front_rhs(child%num)%rhs,     &
          ! qrm_sdata_b%front_rhs(node)%rhs,          &
          ! i=child%npiv+1, m=child%ne-child%npiv,    &
          ! rowmap=child%rowmap,                      &
          ! inout='i')
     
  end do
 
  if(front%ne .le. 0) goto 9999
  
  ! scatter front%b into b
  ! FIXME: check this: maybe only front%arows should be scattered
  mb = qrm_sdata_b%front_rhs(node)%rhs%mb
  do i=1, front%m
     bi = (i-1)/front%mb+1
     ii = i - (bi-1)*front%mb
     do j=1, size(qrm_sdata_b%front_rhs(node)%rhs%blocks,2)
        qrm_sdata_b%p(front%rows(i),(j-1)*mb+1:min(n,j*mb)) = &
             qrm_sdata_b%front_rhs(node)%rhs%blocks(bi,j)%c(ii,:)
     end do
  end do
  
  call zqrm_dsmat_destroy(qrm_sdata_b%front_rhs(node)%rhs, seq=.true.)
  
9999 continue
  
  if(present(info)) info = err
  return
 
end subroutine zqrm_assemble_q
 
 
 
!==========================================================================================
!==========================================================================================
subroutine zqrm_front_rt(front, qrm_sdata_b, qrm_sdata_x)
  use qrm_parameters_mod
  use zqrm_fdata_mod
  use zqrm_sdata_mod
  implicit none
 
  type(zqrm_front_type)           :: front
  type(zqrm_sdata_type)           :: qrm_sdata_b, qrm_sdata_x
 
! #if defined (sprec) || defined (dprec)
  character, parameter            :: tr=qrm_conj_transp, notr=qrm_no_transp
! #elif defined (cprec) || defined (zprec)
  ! character, parameter            :: tr='c', notr=qrm_no_transp
! #endif
 
  integer                         :: bi, bj, bk, m, n, k, i, ii, j, mb, node, info
 
  ! shortcut
  if (min(front%m, front%n) .le. 0) goto 10
  if (front%npiv .le. 0) goto 10
 
  node = front%num
  ! write(*,'(" =--> Solve RT  : ",i4)') front%num
 
  n = size(qrm_sdata_x%p,2)

  call zqrm_dsmat_trsm_async(qrm_seq_dscr_,      &
       qrm_left, qrm_upper,                      &
       qrm_conj_transp, qrm_no_unit,             &
       qrm_zone,                                 &
       front%f, qrm_sdata_x%front_rhs(node)%rhs, &
       m=front%npiv, n=n, k=front%n)

  mb = qrm_sdata_x%front_rhs(node)%rhs%mb

  do i=1, front%npiv
     bi = (i-1)/front%mb+1
     ii = i - (bi-1)*front%mb
     do j=1, size(qrm_sdata_x%front_rhs(node)%rhs%blocks,2)
        qrm_sdata_x%p(front%rows(i),(j-1)*mb+1:min(n,j*mb)) = &
             qrm_sdata_x%front_rhs(node)%rhs%blocks(bi,j)%c(ii,:)
     end do
  end do

  if(front%m.gt.front%ne) qrm_sdata_x%p(front%rows(front%ne+1:front%m),:) = qrm_zzero
 
10 continue
 
9999 continue ! error
  return
end subroutine zqrm_front_rt
 
 
subroutine zqrm_assemble_rt(qrm_fct, front, qrm_sdata_b, qrm_sdata_x, info)
  use qrm_parameters_mod
  use zqrm_spfct_mod
  use qrm_mem_mod
  use qrm_error_mod
  use qrm_common_mod
  use zqrm_sdata_mod
  implicit none
 
  type(zqrm_spfct_type), target  :: qrm_fct
  type(zqrm_front_type)          :: front
  type(zqrm_sdata_type)          :: qrm_sdata_b, qrm_sdata_x
  integer, optional              :: info
 
  type(qrm_adata_type), pointer  :: adata
  type(zqrm_front_type), pointer :: child
  integer                        :: n, c, node, i, ii, bi, mb, j
 
  ! error management
  integer                        :: err
  character(len=*), parameter    :: name='qrm_assemble_rt'
 
  err = 0
 
  adata => qrm_fct%adata
 
  node = front%num
  n  = size(qrm_sdata_b%p,2)
  
  if(front%ne .gt. 0) then
     call zqrm_dsmat_init(qrm_sdata_x%front_rhs(node)%rhs, &
          front%n, n, front%mb, front%mb,                  &
          info=err, seq=.true.,                            &
          init='v', ivalue=qrm_zzero)
     __QRM_INFO_CHECK(err, name,'qrm_dsmat_init',9999)

     mb = qrm_sdata_x%front_rhs(node)%rhs%mb
     do i=1, front%npiv
        bi = (i-1)/front%mb+1
        ii = i - (bi-1)*front%mb
        do j=1, size(qrm_sdata_x%front_rhs(node)%rhs%blocks,2)
           qrm_sdata_x%front_rhs(node)%rhs%blocks(bi,j)%c(ii,:) = &
                qrm_sdata_b%p(front%cols(i),(j-1)*mb+1:min(n,j*mb))
        end do
     end do
     
  end if

  do c = adata%childptr(node), adata%childptr(node+1)-1
     child => qrm_fct%fdata%front_list(adata%child(c))
 
     ! the double condition below should be redundant but you never know...
     if((child%n.gt.child%npiv) .and. (front%ne.gt.0))                          &
          call zqrm_dsmat_extadd_async(qrm_seq_dscr_,                           &
          qrm_sdata_x%front_rhs(child%num)%rhs,                                 &
          qrm_sdata_x%front_rhs(node)%rhs,                                      &
          i=child%npiv+1, j=1,                                                  &
          m=child%n-child%npiv, n=qrm_sdata_x%front_rhs(child%num)%rhs%n, l=0,  &
          inout='o', op='a',                                                    &
          rowmap=child%colmap)
          ! call zqrm_dsmat_axpy_async(qrm_seq_dscr_,   &
          ! qrm_sdata_x%front_rhs(child%num)%rhs,       &
          ! qrm_sdata_x%front_rhs(node)%rhs,            &
          ! i=child%npiv+1, m=child%n-child%npiv,       &
          ! rowmap=child%colmap)

     call zqrm_dsmat_destroy(qrm_sdata_x%front_rhs(child%num)%rhs, seq=.true.)
  end do

  
9999 continue
 
  if(present(info)) info = err
  return
 
end subroutine zqrm_assemble_rt
 
 
 
 
 
subroutine zqrm_spfct_unmqr_subtree(transp, qrm_fct, iroot, qrm_sdata_b, work, info)
  use zqrm_spfct_mod
  use qrm_common_mod
  use qrm_string_mod
  use zqrm_solve_mod, protect => zqrm_spfct_unmqr_subtree
  use zqrm_sdata_mod
  implicit none
  
  type(zqrm_spfct_type), intent(in), target :: qrm_fct
  type(zqrm_sdata_type)                     :: qrm_sdata_b
  type(zqrm_ws_type)                        :: work
  character(len=*), intent(in)              :: transp
  integer                                   :: iroot
  integer, optional                         :: info
 
  integer                                   :: inode, node, f, root, ileaf
  type(qrm_adata_type), pointer             :: adata
  type(zqrm_fdata_type), pointer            :: fdata
  type(zqrm_front_type), pointer            :: front
 
  integer                           :: err
  character(len=*), parameter       :: name='qrm_spfct_unmqr_subtree'
 
  err = 0
  ! write(*,'(" =--> Apply subtree : ",i4)')iroot

  ! simplify
  adata => qrm_fct%adata
  fdata => qrm_fct%fdata
  root  = adata%torder(iroot) 
 
  if(qrm_str_tolower(transp(1:1)) .eq. qrm_conj_transp) then
     inode = adata%small(root)
     subtreeqt: do
 
        node = adata%torder(inode)
        front => fdata%front_list(node)
        call zqrm_assemble_qt(qrm_fct, front, qrm_sdata_b, err)
        __QRM_INFO_CHECK(err, name,'qrm_assemble_qt',9999)
        call zqrm_front_qt(front, qrm_sdata_b, work)
        
        if(front%num.eq.root) exit subtreeqt
        inode = inode+1
        
     end do subtreeqt
  else
     inode = iroot
     ileaf = adata%small(root)
     
     subtreeq: do
 
        node = adata%torder(inode)
        front => fdata%front_list(node)
        call zqrm_front_q(front, qrm_sdata_b, work)
        call zqrm_assemble_q(qrm_fct, front, qrm_sdata_b, err)
        __QRM_INFO_CHECK(err, name,'qrm_assemble_q',9999)
        
        if(inode.eq.ileaf) exit subtreeq
        inode = inode-1
        
     end do subtreeq
     
  end if
 
9999 continue
 
  if(present(info)) info = err
  return
 
end subroutine zqrm_spfct_unmqr_subtree
 
 
subroutine zqrm_spfct_trsm_subtree(transp, qrm_fct, iroot, qrm_sdata_b, qrm_sdata_x, info)
  use zqrm_spfct_mod
  use qrm_common_mod
  use qrm_string_mod
  use zqrm_solve_mod, protect => zqrm_spfct_trsm_subtree
  use zqrm_sdata_mod
  implicit none
  
  type(zqrm_spfct_type), intent(in), target :: qrm_fct
  type(zqrm_sdata_type)                     :: qrm_sdata_b, qrm_sdata_x
  character(len=*), intent(in)              :: transp
  integer                                   :: iroot
  integer, optional                         :: info
 
  integer                                   :: inode, node, f, root, ileaf
  type(qrm_adata_type), pointer             :: adata
  type(zqrm_fdata_type), pointer            :: fdata
  type(zqrm_front_type), pointer            :: front
 
  integer                           :: err
  character(len=*), parameter       :: name='qrm_spfct_trsm_subtree'

  err = 0
 
  ! write(*,'(" =--> Solve subtree : ",i4)')iroot
  ! simplify
  adata => qrm_fct%adata
  fdata => qrm_fct%fdata
  root  = adata%torder(iroot) 
 
  if(qrm_str_tolower(transp(1:1)) .eq. qrm_conj_transp) then
     inode = adata%small(root)
     subtreert: do
 
        node = adata%torder(inode)
        front => fdata%front_list(node)
        call zqrm_assemble_rt(qrm_fct, front, qrm_sdata_b, qrm_sdata_x, err)
        __QRM_INFO_CHECK(err, name,'qrm_assemble_rt',9999)
        call zqrm_front_rt(front, qrm_sdata_b, qrm_sdata_x)
        
        if(front%num.eq.root) exit subtreert
        inode = inode+1
        
     end do subtreert
  else
     inode = iroot
     ileaf = adata%small(root)
     
     subtreer: do
 
        node = adata%torder(inode)
        front => fdata%front_list(node)
        call zqrm_front_r(front, qrm_sdata_b, qrm_sdata_x)
        call zqrm_assemble_r(qrm_fct, front, qrm_sdata_b, qrm_sdata_x, err)
        __QRM_INFO_CHECK(err, name,'qrm_assemble_r',9999)
        
        if(inode.eq.ileaf) exit subtreer
        inode = inode-1
        
     end do subtreer
     
  end if
 
9999 continue
 
  if(present(info)) info = err
  return
 
end subroutine zqrm_spfct_trsm_subtree


