!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> qrm_solve_tasks.F90
!! This file contains all the tasks for the solve phase
!!
!!
!! ##############################################################################################

#include "qrm_common.h"




subroutine zqrm_node_unmqr_task(qrm_dscr, transp, qrm_spfct, node, qrm_sdata_b)
  ! the Intel compler bugs here when used with vectorization options
  !DIR$ NOOPTIMIZE  
  use zqrm_spfct_mod
  use qrm_common_mod
  use qrm_string_mod
  use zqrm_sdata_mod
  use qrm_dscr_mod
  use zqrm_solve_mod, protect => zqrm_node_unmqr_task
#if defined (have_starpu)
  use zqrm_starpu_mod
  use fstarpu_mod
  use iso_c_binding
#endif
  implicit none
  
  type(qrm_dscr_type), target       :: qrm_dscr
  type(zqrm_spfct_type), target     :: qrm_spfct
  type(zqrm_sdata_type), target     :: qrm_sdata_b
  integer, target                   :: node
  character(len=*), target          :: transp

  logical                           :: issmall 
  type(zqrm_front_type), pointer    :: front
#if defined (have_starpu)
  type(c_ptr), target               :: hdls
  integer, target                   :: nc
  integer                           :: i, c
  character(kind=c_char)            :: transp_c
  type(c_ptr), target               :: qrm_spfct_c, dscrpt, qrm_sdata_b_c
#endif

  integer                           :: info
  character(len=*), parameter       :: name='qrm_node_unmqr_task'

  if(qrm_dscr%info.ne.0) return
  info = 0
  
#if defined (have_starpu)
  
  front => qrm_spfct%fdata%front_list(node)

  qrm_sdata_b_c = c_loc(qrm_sdata_b)
  dscrpt        = c_loc(qrm_dscr)
  qrm_spfct_c     = c_loc(qrm_spfct)
  
  if(qrm_spfct%adata%small(node) .eq. 0) then
     nc = qrm_spfct%adata%childptr(node+1)-qrm_spfct%adata%childptr(node)
  else
     nc = 0
  end if

  hdls = fstarpu_data_descr_array_alloc(nc+2)
  call fstarpu_data_descr_array_set(hdls, 0, &
       front%sym_handle, FSTARPU_R)
  
  call fstarpu_data_descr_array_set(hdls, 1, &
       qrm_sdata_b%front_rhs(node)%hdl, FSTARPU_RW)
  
  do i=1, nc
     c = qrm_spfct%adata%child(i-1 + qrm_spfct%adata%childptr(node))
     if(transp.eq.qrm_conj_transp) then
        call fstarpu_data_descr_array_set(hdls, 1+i, &
             qrm_sdata_b%front_rhs(c)%hdl, FSTARPU_R)
     else
        call fstarpu_data_descr_array_set(hdls, 1+i, &
             qrm_sdata_b%front_rhs(c)%hdl, FSTARPU_RW)
     end if
  end do

  nc = nc+2

  call fstarpu_insert_task((/ zqrm_node_unmqr_cl,                        &
       FSTARPU_VALUE    , c_loc(dscrpt)          , FSTARPU_SZ_C_PTR,     &
       FSTARPU_VALUE    , c_loc(qrm_spfct_c)     , FSTARPU_SZ_C_PTR,     &
       FSTARPU_VALUE    , c_loc(qrm_sdata_b_c)   , FSTARPU_SZ_C_PTR,     &
       FSTARPU_VALUE    , c_loc(transp)          , FSTARPU_SZ_CHARACTER, &
       FSTARPU_VALUE    , c_loc(node)            , FSTARPU_SZ_C_INT,     &
       FSTARPU_SCRATCH  , qrm_sdata_b%work%hdl   ,                       &
       FSTARPU_DATA_MODE_ARRAY, hdls, c_loc(nc)  ,                       &
       FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx)    ,                       &
       C_NULL_PTR /))

  call fstarpu_data_descr_array_free(hdls)
    
#else

  issmall = qrm_spfct%adata%small(node) .gt. 0
  if(issmall) then
     call zqrm_spfct_unmqr_subtree(transp, qrm_spfct, node, qrm_sdata_b, qrm_sdata_b%work, info)
     __QRM_INFO_CHECK(info, name, 'qrm_spfct_unmqr_subtree', 9999)
  else
     front => qrm_spfct%fdata%front_list(node)

     if(qrm_str_tolower(transp(1:1)) .eq. qrm_conj_transp) then
        call zqrm_assemble_qt(qrm_spfct, front, qrm_sdata_b, info)
        __QRM_INFO_CHECK(info, name, 'qrm_assemble_qt', 9999)
        call zqrm_front_qt(front, qrm_sdata_b, qrm_sdata_b%work)
     else
        call zqrm_front_q(front, qrm_sdata_b, qrm_sdata_b%work)
        call zqrm_assemble_q(qrm_spfct, front, qrm_sdata_b, info)
        __QRM_INFO_CHECK(info, name, 'qrm_assemble_q', 9999)
     end if
  end if

#endif

9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return

end subroutine zqrm_node_unmqr_task

#if defined (have_starpu)

recursive subroutine zqrm_node_unmqr_cpu_func(buffers, cl_arg) bind(C)
  use zqrm_spfct_mod
  use zqrm_fdata_mod
  use zqrm_sdata_mod
  use zqrm_factorization_mod
  use zqrm_solve_mod
  use fstarpu_mod
  use iso_c_binding
  use qrm_dscr_mod
  implicit none

  type(c_ptr), value             :: cl_arg
  type(c_ptr), value             :: buffers

  type(c_ptr), target            :: qrm_spfct_c, qrm_sdata_b_c, qrm_dscr_c
  integer, target                :: node
  type(zqrm_front_type), pointer :: front
  type(zqrm_sdata_type), pointer :: qrm_sdata_b

  type(qrm_dscr_type), pointer   :: qrm_dscr
  type(zqrm_spfct_type), pointer :: qrm_spfct
  character(kind=c_char), target :: transp
  integer                        :: info
  integer, target                :: nw, ldw
  type(zqrm_ws_type)             :: work
  logical                        :: issmall

  call fstarpu_unpack_arg(cl_arg,(/ &
       c_loc(qrm_dscr_c),           & 
       c_loc(qrm_spfct_c),            &
       c_loc(qrm_sdata_b_c),        &
       c_loc(transp),               &
       c_loc(node)/))
  
  ! until StarPU has a proper error handling method we have to check
  ! manually
  call c_f_pointer(qrm_dscr_c, qrm_dscr)
  if(qrm_dscr%info.ne.0) return

  call c_f_pointer(qrm_spfct_c, qrm_spfct)
  call c_f_pointer(qrm_sdata_b_c, qrm_sdata_b)

  nw   = fstarpu_matrix_get_ny(buffers, 0)
  ldw  = fstarpu_matrix_get_ld(buffers, 0)
  call c_f_pointer(fstarpu_matrix_get_ptr(buffers, 0), work%c, shape=(/ldw,nw/))

  issmall = qrm_spfct%adata%small(node) .gt. 0

  if(issmall) then
     call zqrm_spfct_unmqr_subtree(transp, qrm_spfct, node, qrm_sdata_b, work, info)
  else
     front => qrm_spfct%fdata%front_list(node)
     if(transp .eq. qrm_conj_transp) then
        call zqrm_assemble_qt(qrm_spfct, front, qrm_sdata_b, info)
        if(info.eq.0) call zqrm_front_qt(front, qrm_sdata_b, work)
     else
        call zqrm_front_q(front, qrm_sdata_b, work)
        call zqrm_assemble_q(qrm_spfct, front, qrm_sdata_b, info)
     end if
  end if

  call qrm_error_set(qrm_dscr%info, info)

#if defined(prnt_tsk)
  write(*,'("Apply --  node: ",i4,"   size:",i4,2x,i4)')node,size(b%p,1),size(b%p,2)
#endif
  return

end subroutine zqrm_node_unmqr_cpu_func

#endif





subroutine zqrm_node_trsm_task(qrm_dscr, transp, qrm_spfct, node, qrm_sdata_b, qrm_sdata_x)
  ! the Intel compler bugs here when used with vectorization options
  !DIR$ NOOPTIMIZE  
  use zqrm_spfct_mod
  use qrm_common_mod
  use qrm_string_mod
  use zqrm_sdata_mod
  use qrm_dscr_mod
  use zqrm_solve_mod, protect => zqrm_node_trsm_task
#if defined (have_starpu)
  use zqrm_starpu_mod
#endif
  implicit none
  
  type(qrm_dscr_type), target       :: qrm_dscr
  type(zqrm_spfct_type), target     :: qrm_spfct
  type(zqrm_sdata_type), target     :: qrm_sdata_b, qrm_sdata_x
  character(len=*), target          :: transp
  integer, target                   :: node
 
  logical                           :: issmall
  type(zqrm_front_type), pointer    :: front
 
#if defined (have_starpu)
  type(c_ptr)                       :: hdls
  integer, target                   :: nc      
  integer                           :: i, c
  character(kind=c_char)            :: transp_c
  type(c_ptr), target               :: qrm_spfct_c, dscrpt, qrm_sdata_b_c, qrm_sdata_x_c
#endif
 
  integer                           :: info
  character(len=*), parameter       :: name='qrm_node_trsm_task'

  if (qrm_dscr%info.ne.0) return
  info = 0
  
#if defined (have_starpu)

  front => qrm_spfct%fdata%front_list(node)

  qrm_sdata_b_c = c_loc(qrm_sdata_b)
  qrm_sdata_x_c = c_loc(qrm_sdata_x)
  dscrpt        = c_loc(qrm_dscr)
  qrm_spfct_c     = c_loc(qrm_spfct)
  
  if(qrm_spfct%adata%small(node) .eq. 0) then
     nc = qrm_spfct%adata%childptr(node+1)-qrm_spfct%adata%childptr(node)
  else
     nc = 0
  end if

  hdls = fstarpu_data_descr_array_alloc(2*nc+3)
  call fstarpu_data_descr_array_set(hdls, 0, &
       front%sym_handle, FSTARPU_R)
  
  call fstarpu_data_descr_array_set(hdls, 1, &
       qrm_sdata_x%front_rhs(node)%hdl, FSTARPU_RW)
  
  call fstarpu_data_descr_array_set(hdls, 2, &
       qrm_sdata_b%front_rhs(node)%hdl, FSTARPU_R)
  
  do i=1, nc
     c = qrm_spfct%adata%child(i-1 + qrm_spfct%adata%childptr(node))
     if(transp.eq.qrm_conj_transp) then
        call fstarpu_data_descr_array_set(hdls, 2+(i-1)*2+1, &
             qrm_sdata_x%front_rhs(c)%hdl, FSTARPU_R)
     else
        call fstarpu_data_descr_array_set(hdls, 2+(i-1)*2+1, &
             qrm_sdata_x%front_rhs(c)%hdl, FSTARPU_RW)
     end if
     call fstarpu_data_descr_array_set(hdls, 2+i*2, &
          qrm_sdata_b%front_rhs(c)%hdl, FSTARPU_R)
  end do

  nc = 2*nc+3

  call fstarpu_insert_task((/ zqrm_node_trsm_cl,                        &
       FSTARPU_VALUE    , c_loc(dscrpt)          , FSTARPU_SZ_C_PTR,     &
       FSTARPU_VALUE    , c_loc(qrm_spfct_c)     , FSTARPU_SZ_C_PTR,     &
       FSTARPU_VALUE    , c_loc(qrm_sdata_b_c)   , FSTARPU_SZ_C_PTR,     &
       FSTARPU_VALUE    , c_loc(qrm_sdata_x_c)   , FSTARPU_SZ_C_PTR,     &
       FSTARPU_VALUE    , c_loc(transp)          , FSTARPU_SZ_CHARACTER, &
       FSTARPU_VALUE    , c_loc(node)            , FSTARPU_SZ_C_INT,     &
       FSTARPU_DATA_MODE_ARRAY, hdls, c_loc(nc)  ,                       &
       FSTARPU_SCHED_CTX, c_loc(qrm_dscr%ctx)    ,                       &
       C_NULL_PTR /))

  call fstarpu_data_descr_array_free(hdls)
    
#else
 
  issmall = qrm_spfct%adata%small(node) .gt. 0
  
  if(issmall) then
     call zqrm_spfct_trsm_subtree(transp, qrm_spfct, node, qrm_sdata_b, qrm_sdata_x, info)
     __QRM_INFO_CHECK(info, name, 'qrm_spfct_trsm_subtree', 9999)
  else
     front => qrm_spfct%fdata%front_list(node)
     if(qrm_str_tolower(transp(1:1)) .eq. qrm_conj_transp) then
        call zqrm_assemble_rt(qrm_spfct, front, qrm_sdata_b, qrm_sdata_x, info)
        __QRM_INFO_CHECK(info, name, 'qrm_assemble_rt', 9999)
        call zqrm_front_rt(front, qrm_sdata_b, qrm_sdata_x)
     else
        call zqrm_front_r(front, qrm_sdata_b, qrm_sdata_x)
        call zqrm_assemble_r(qrm_spfct, front, qrm_sdata_b, qrm_sdata_x, info)
        __QRM_INFO_CHECK(info, name, 'qrm_assemble_r', 9999)
     end if
  end if
 
#endif
 
9999 continue
  call qrm_error_set(qrm_dscr%info, info)
  return
 
end subroutine zqrm_node_trsm_task




#if defined (have_starpu)

recursive subroutine zqrm_node_trsm_cpu_func(buffers, cl_arg) bind(C)
  use zqrm_spfct_mod
  use zqrm_fdata_mod
  use zqrm_sdata_mod
  use zqrm_factorization_mod
  use zqrm_solve_mod
  use fstarpu_mod
  use qrm_string_mod
  use iso_c_binding
  use qrm_dscr_mod
  implicit none

  type(c_ptr), value             :: cl_arg
  type(c_ptr), value             :: buffers

  type(c_ptr), target            :: qrm_spfct_c, qrm_sdata_b_c, qrm_sdata_x_c, qrm_dscr_c
  integer, target                :: node
  type(zqrm_front_type), pointer :: front
  type(zqrm_sdata_type), pointer :: qrm_sdata_b, qrm_sdata_x

  type(qrm_dscr_type), pointer   :: qrm_dscr
  type(zqrm_spfct_type), pointer :: qrm_spfct
  character(kind=c_char), target :: transp
  integer                        :: info
  logical                        :: issmall

  call fstarpu_unpack_arg(cl_arg,(/ &
       c_loc(qrm_dscr_c),           &
       c_loc(qrm_spfct_c),            &
       c_loc(qrm_sdata_b_c),        &
       c_loc(qrm_sdata_x_c),        &
       c_loc(transp),               &
       c_loc(node)/))
  
  ! until StarPU has a proper error handling method we have to check
  ! manually
  call c_f_pointer(qrm_dscr_c, qrm_dscr)
  if(qrm_dscr%info.ne.0) return

  call c_f_pointer(qrm_spfct_c, qrm_spfct)
  call c_f_pointer(qrm_sdata_b_c, qrm_sdata_b)
  call c_f_pointer(qrm_sdata_x_c, qrm_sdata_x)

  issmall = qrm_spfct%adata%small(node) .gt. 0

  if(issmall) then
     call zqrm_spfct_trsm_subtree(transp, qrm_spfct, node, qrm_sdata_b, qrm_sdata_x, info)
  else
     front => qrm_spfct%fdata%front_list(node)
     if(qrm_str_tolower(transp(1:1)) .eq. qrm_conj_transp) then
        call zqrm_assemble_rt(qrm_spfct, front, qrm_sdata_b, qrm_sdata_x, info)
        if(info.eq.0) call zqrm_front_rt(front, qrm_sdata_b, qrm_sdata_x)
     else
        call zqrm_front_r(front, qrm_sdata_b, qrm_sdata_x)
        call zqrm_assemble_r(qrm_spfct, front, qrm_sdata_b, qrm_sdata_x, info)
     end if
  end if

  call qrm_error_set(qrm_dscr%info, info)

#if defined(prnt_tsk)
  write(*,'("Solve --  node: ",i4,"   size:",i4,2x,i4)')node,size(b%p,1),size(b%p,2)
#endif
  return

end subroutine zqrm_node_trsm_cpu_func

#endif
