!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!
!! This file is part of qr_mumps.
!!
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as
!! published by the Free Software Foundation, either version 3 of
!! the License, or (at your option) any later version.
!!
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_spmat_transpose.F90
!! This file contains routines to transpose qrm_spmat_type matrices
!!
!!
!! ##############################################################################################

#include "qrm_common.h"


!> This subroutine transposes a qrm_spmat_type matrix in place
!!
!! @param[in,out] qrm_spmat      the matrix to be transposed
!!
subroutine zqrm_spmat_transpose_inplace(qrm_spmat, info)

  use qrm_error_mod
  use zqrm_spmat_mod, savesym => zqrm_spmat_transpose_inplace
  implicit none

  type(zqrm_spmat_type) :: qrm_spmat
  integer, optional                  :: info

  integer                            :: t
  integer, pointer                   :: tmp(:)
  
  integer                            :: err
  character(len=*), parameter        :: name='qrm_spmat_transpose_inplace'

  err = 0

  select case(qrm_spmat%fmt)
  case('coo')
     tmp => qrm_spmat%irn
     qrm_spmat%irn => qrm_spmat%jcn
     qrm_spmat%jcn => tmp
  case('csr')
     qrm_spmat%jptr => qrm_spmat%iptr
     qrm_spmat%irn  => qrm_spmat%jcn
     nullify(qrm_spmat%iptr)
     nullify(qrm_spmat%jcn)
  case('csc')
     qrm_spmat%iptr => qrm_spmat%jptr
     qrm_spmat%jcn  => qrm_spmat%irn
     nullify(qrm_spmat%jptr)
     nullify(qrm_spmat%irn)
  case default
     ! format not known. set info and jump to 9999
     info = 1
     call qrm_error_print(info, name, aed=qrm_spmat%fmt)
     goto 9999
  end select

  t = qrm_spmat%m
  qrm_spmat%m = qrm_spmat%n
  qrm_spmat%n = t

9999 continue
  if(present(info)) info = err
  return

end subroutine zqrm_spmat_transpose_inplace
