!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!!
!! This file contains a routine that computes the norm of a vector.
!!
!! ##############################################################################################

#include "qrm_common.h"

subroutine zqrm_matgen_7pt(qrm_spmat, nx, ny, nz, sym)
  use zqrm_spmat_mod
  implicit none

  ! Nx, Ny, Nz   INTEGER variable taht need be set to the 
  ! dimension of the grid.
  ! Sym contains the symmetry of the matrix 
  integer               :: nx, ny, nz
  type(zqrm_spmat_type) :: qrm_spmat
  integer, optional     :: sym

  integer               :: x, y, z, r
  integer               :: nnz, i, nnn, isym, info

  __QRM_IFPRESENT3(isym, sym, 1)
  
  if((isym.eq.1) .or. (isym.eq.2)) then
     nnz = (nx*ny*3 - nx - ny )*nz + nx*ny*(nz-1)
  elseif (isym.eq.0) then
     nnz = 5*nx*ny*nz -2*nx*nz -2*ny*nz + nx*ny*(2*nz-2)
  end if
  nnn  = nx*ny*nz
  
  ! write(*,*) 'N = ', nnn
  ! write(*,*) 'NNZ = ',nnz
  ! write(*,*) 'SYM = ',isym

  call zqrm_spmat_alloc(qrm_spmat, nnz, nnn, nnn, 'coo', info)

  qrm_spmat%sym = isym
  
  i = 0

  do z = 1, nz
     do y = 1, ny
        do x = 1, nx
           r = nx*ny*(z-1) + nx*(y-1) + x

           ! me
           i = i+1
#if defined(dprec) || defined(sprec)
           qrm_spmat%val(i)   = 6.0E0
#elif defined(zprec) || defined(cprec)
           qrm_spmat%val(i)   =(6.0E0,0.0E0)
#endif

           qrm_spmat%irn(i) = r
           qrm_spmat%jcn(i) = r

           ! east
           if(x.lt.nx) then
              i = i+1
#if defined(dprec) || defined(sprec)
              qrm_spmat%val(i)   = -1.0E0
#elif defined(zprec) || defined(cprec)
              qrm_spmat%val(i)   =(-1.0E0,0.0E0)
#endif
              qrm_spmat%irn(i) = r
              qrm_spmat%jcn(i) = r+1
           end if

           ! north
           if(y.lt.ny) then
              i = i+1
#if defined(dprec) || defined(sprec)
              qrm_spmat%val(i)   = -1.0E0
#elif defined(zprec) || defined(cprec)
              qrm_spmat%val(i)   =(-1.0E0,0.0E0)
#endif
              qrm_spmat%irn(i) = r
              qrm_spmat%jcn(i) = r+nx
           end if

           ! top
           if(z.lt.nz) then
              i = i+1
#if defined(dprec) || defined(sprec)
              qrm_spmat%val(i)   = -1.0E0
#elif defined(zprec) || defined(cprec)
              qrm_spmat%val(i)   =(-1.0E0,0.0E0)
#endif
              qrm_spmat%irn(i) = r
              qrm_spmat%jcn(i) = r+nx*ny
           end if

           if(isym.eq.0) then
              ! west
              if(x.gt.1) then
                 i = i+1
#if defined(dprec) || defined(sprec)
                 qrm_spmat%val(i)   = -1.0E0
#elif defined(zprec) || defined(cprec)
                 qrm_spmat%val(i)   =(-1.0E0,0.0E0)
#endif
                 qrm_spmat%irn(i) = r
                 qrm_spmat%jcn(i) = r-1
              end if

              ! south
              if(y.gt.1) then
                 i = i+1
#if defined(dprec) || defined(sprec)
                 qrm_spmat%val(i)   = -1.0E0
#elif defined(zprec) || defined(cprec)
                 qrm_spmat%val(i)   =(-1.0E0,0.0E0)
#endif
                 qrm_spmat%irn(i) = r
                 qrm_spmat%jcn(i) = r-nx
              end if

              ! bottom
              if(z.gt.1) then
                 i = i+1
#if defined(dprec) || defined(sprec)
                 qrm_spmat%val(i)   = -1.0E0
#elif defined(zprec) || defined(cprec)
                 qrm_spmat%val(i)   =(-1.0E0,0.0E0)
#endif
                 qrm_spmat%irn(i) = r
                 qrm_spmat%jcn(i) = r-nx*ny
              end if
           end if
        end do
     end do
  end do
  if(i.ne.nnz) write(*,'("Error, nnz not corrrect")')

  return
end subroutine zqrm_matgen_7pt


subroutine zqrm_matgen_mitt3d(qrm_spmat, n)
  use qrm_mem_mod
  use qrm_parameters_mod
  use zqrm_spmat_mod
  implicit none

  type(zqrm_spmat_type)  :: qrm_spmat
  integer                :: n

  integer :: i, j, k, ig, idx, ounit
  integer, pointer :: irn(:), jcn(:)
  complex(r64), pointer :: val(:)

  ! call qrm_get('qrm_ounit', ounit)

  qrm_spmat%n   = (n+2)*(n+2)*(n+2);
  qrm_spmat%m   = n*n*n;
  qrm_spmat%nz  = 27*n*n*n;
  qrm_spmat%sym = 0

  ! if(ounit.gt.0)write(ounit,'("Generating Mittelmann Jacobian matrix for constraints of a")')
  ! if(ounit.gt.0)write(ounit,'("3D control problem with Dirichlet boundary conditions.")')
  ! if(ounit.gt.0)Write(ounit,'("N=",i3,"   ===>  m=",i7,"  n=",i7,"  nz=",i10)')n,qrm_spmat%m, qrm_spmat%n, qrm_spmat%nz

  call qrm_alloc(qrm_spmat%irn, qrm_spmat%nz)
  call qrm_alloc(qrm_spmat%jcn, qrm_spmat%nz)
  call qrm_alloc(qrm_spmat%val, qrm_spmat%nz)
  irn => qrm_spmat%irn
  jcn => qrm_spmat%jcn
  val => qrm_spmat%val

  idx = 1
  ig = 1
  do i=1, n
     do j=1, n
        do k=1, n

           irn(idx) = ig; jcn(idx) = y_index(n ,i-1 ,j   ,k  )  ; val(idx) = -16; idx=idx+1; ! y(i-1 ,j   ,k  )     
           irn(idx) = ig; jcn(idx) = y_index(n ,i   ,j   ,k  )  ; val(idx) = 200; idx=idx+1; ! y(i   ,j   ,k  )       
           irn(idx) = ig; jcn(idx) = y_index(n ,i+1 ,j   ,k  )  ; val(idx) = -16; idx=idx+1; ! y(i+1 ,j   ,k  )     
           irn(idx) = ig; jcn(idx) = y_index(n ,i-1 ,j-1 ,k  )  ; val(idx) = -8 ; idx=idx+1; ! y(i-1 ,j-1 ,k  )   
           irn(idx) = ig; jcn(idx) = y_index(n ,i   ,j-1 ,k  )  ; val(idx) = -16; idx=idx+1; ! y(i   ,j-1 ,k  )     
           irn(idx) = ig; jcn(idx) = y_index(n ,i+1 ,j-1 ,k  )  ; val(idx) = -8 ; idx=idx+1; ! y(i+1 ,j-1 ,k  )   
           irn(idx) = ig; jcn(idx) = y_index(n ,i-1 ,j+1 ,k  )  ; val(idx) = -8 ; idx=idx+1; ! y(i-1 ,j+1 ,k  )   
           irn(idx) = ig; jcn(idx) = y_index(n ,i   ,j+1 ,k  )  ; val(idx) = -16; idx=idx+1; ! y(i   ,j+1 ,k  )     
           irn(idx) = ig; jcn(idx) = y_index(n ,i+1 ,j+1 ,k  )  ; val(idx) = -8 ; idx=idx+1; ! y(i+1 ,j+1 ,k  )   
           irn(idx) = ig; jcn(idx) = y_index(n ,i-1 ,j   ,k-1)  ; val(idx) = -8 ; idx=idx+1; ! y(i-1 ,j   ,k-1)   
           irn(idx) = ig; jcn(idx) = y_index(n ,i   ,j   ,k-1)  ; val(idx) = -16; idx=idx+1; ! y(i   ,j   ,k-1)     
           irn(idx) = ig; jcn(idx) = y_index(n ,i+1 ,j   ,k-1)  ; val(idx) = -8 ; idx=idx+1; ! y(i+1 ,j   ,k-1)   
           irn(idx) = ig; jcn(idx) = y_index(n ,i-1 ,j-1 ,k-1)  ; val(idx) = -1 ; idx=idx+1; ! y(i-1 ,j-1 ,k-1) 
           irn(idx) = ig; jcn(idx) = y_index(n ,i   ,j-1 ,k-1)  ; val(idx) = -8 ; idx=idx+1; ! y(i   ,j-1 ,k-1)   
           irn(idx) = ig; jcn(idx) = y_index(n ,i+1 ,j-1 ,k-1)  ; val(idx) = -1 ; idx=idx+1; ! y(i+1 ,j-1 ,k-1) 
           irn(idx) = ig; jcn(idx) = y_index(n ,i-1 ,j+1 ,k-1)  ; val(idx) = -1 ; idx=idx+1; ! y(i-1 ,j+1 ,k-1) 
           irn(idx) = ig; jcn(idx) = y_index(n ,i   ,j+1 ,k-1)  ; val(idx) = -8 ; idx=idx+1; ! y(i   ,j+1 ,k-1)   
           irn(idx) = ig; jcn(idx) = y_index(n ,i+1 ,j+1 ,k-1)  ; val(idx) = -1 ; idx=idx+1; ! y(i+1 ,j+1 ,k-1) 
           irn(idx) = ig; jcn(idx) = y_index(n ,i-1 ,j   ,k+1)  ; val(idx) = -8 ; idx=idx+1; ! y(i-1 ,j   ,k+1)   
           irn(idx) = ig; jcn(idx) = y_index(n ,i   ,j   ,k+1)  ; val(idx) = -16; idx=idx+1; ! y(i   ,j   ,k+1)     
           irn(idx) = ig; jcn(idx) = y_index(n ,i+1 ,j   ,k+1)  ; val(idx) = -8 ; idx=idx+1; ! y(i+1 ,j   ,k+1)   
           irn(idx) = ig; jcn(idx) = y_index(n ,i-1 ,j-1 ,k+1)  ; val(idx) = -1 ; idx=idx+1; ! y(i-1 ,j-1 ,k+1) 
           irn(idx) = ig; jcn(idx) = y_index(n ,i   ,j-1 ,k+1)  ; val(idx) = -8 ; idx=idx+1; ! y(i   ,j-1 ,k+1)   
           irn(idx) = ig; jcn(idx) = y_index(n ,i+1 ,j-1 ,k+1)  ; val(idx) = -1 ; idx=idx+1; ! y(i+1 ,j-1 ,k+1) 
           irn(idx) = ig; jcn(idx) = y_index(n ,i-1 ,j+1 ,k+1)  ; val(idx) = -1 ; idx=idx+1; ! y(i-1 ,j+1 ,k+1) 
           irn(idx) = ig; jcn(idx) = y_index(n ,i   ,j+1 ,k+1)  ; val(idx) = -8 ; idx=idx+1; ! y(i   ,j+1 ,k+1)   
           irn(idx) = ig; jcn(idx) = y_index(n ,i+1 ,j+1 ,k+1)  ; val(idx) = -1 ; idx=idx+1; ! y(i+1 ,j+1 ,k+1) 

           ig=ig+1

        end do
     end do
  end do

  ! do i=1, qrm_spmat%nz
  ! write(*,*)qrm_spmat%irn(i),qrm_spmat%jcn(i),qrm_spmat%val(i)
  ! end do

  return

contains

  function y_index(n, i, j, k)
    integer :: n, i, j, k, y_index
    
    y_index = k + (n+2)*j + (n+2)*(n+2)*i + 1
    return
  end function y_index
  
end subroutine zqrm_matgen_mitt3d


subroutine zqrm_tikhonov(qrm_spmat, gamma)
  use qrm_mem_mod
  use zqrm_spmat_mod
  implicit none
  
  type(zqrm_spmat_type)  :: qrm_spmat
  real(r64)              :: gamma
  
  integer                :: i
  integer, parameter     :: ione=1
  real(r64)              :: fnorma, dznrm2
  
  fnorma = dznrm2(qrm_spmat%nz, qrm_spmat%val(1), ione)
  
  call qrm_realloc(qrm_spmat%irn, qrm_spmat%nz+min(qrm_spmat%m,qrm_spmat%n), copy=.true.)
  call qrm_realloc(qrm_spmat%jcn, qrm_spmat%nz+min(qrm_spmat%m,qrm_spmat%n), copy=.true.)
  call qrm_realloc(qrm_spmat%val, qrm_spmat%nz+min(qrm_spmat%m,qrm_spmat%n), copy=.true.)
  
  do i=1, min(qrm_spmat%m,qrm_spmat%n)
     qrm_spmat%val(qrm_spmat%nz+i) = gamma*fnorma
     if(qrm_spmat%m .ge. qrm_spmat%n) then
        qrm_spmat%irn(qrm_spmat%nz+i) = qrm_spmat%m+i
        qrm_spmat%jcn(qrm_spmat%nz+i) = i
     else
        qrm_spmat%jcn(qrm_spmat%nz+i) = qrm_spmat%n+i
        qrm_spmat%irn(qrm_spmat%nz+i) = i
     end if
  end do
  
  qrm_spmat%nz = qrm_spmat%nz+min(qrm_spmat%m,qrm_spmat%n)
  if(qrm_spmat%m .ge. qrm_spmat%n) then
     qrm_spmat%m = qrm_spmat%m + qrm_spmat%n
  else
     qrm_spmat%n = qrm_spmat%n + qrm_spmat%m
  end if
  return
end subroutine zqrm_tikhonov
