!! ##############################################################################################
!!
!! Copyright 2012-2020 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################


!! ##############################################################################################
!> @file qrm_test_c.F90
!! This file contains tests for the C interface
!!
!!
!! ##############################################################################################

function cqrm_test_c(c, m) result(ok)
  use cqrm_testing_mod, protect => cqrm_test_c
  use cqrm_c_interface
  implicit none
  logical :: ok

  integer :: c, m

  integer, parameter :: ncases=15
  logical :: cases(ncases)

  if(c .eq. -1) then
     cases = .true.
  else if (c .le. ncases) then
     cases = .false.
     cases(c) = .true.
  end if

  ok = .true.
  
  if(cases(1 )) ok = case1 (m) .and. ok 
  if(cases(2 )) ok = case2 (m) .and. ok 


  return


contains

  function case1(m) result(ok)
    use cqrm_mod
    use cqrm_c_interface
    implicit none

    logical :: ok
    integer :: m

    interface
        function cqrm_test_solve_c(m, n, nz, &
            irn, jcn, val, &
            b_c, x_c, r_c, &
            eps, err) result(info) bind(c)
         use iso_c_binding
         use cqrm_c_interface
         real(r32)               :: err
         real(r32), value        :: eps
         integer(c_int), value   :: m, n, nz
         ! type(cqrm_spmat_type_c) :: qrm_spmat_c
         type(c_ptr), value      :: b_c, x_c, r_c, irn, jcn, val
         integer(c_int)          :: info
       end function cqrm_test_solve_c
    end interface
    

    type(cqrm_spmat_type), pointer :: qrm_spmat
    integer                        :: info
    complex(r32), pointer          :: x(:), b(:), r(:)
    real(r32)                      :: err
    integer                        :: i
    ! type(cqrm_spmat_type_c)        :: qrm_spmat_c
    type(c_ptr)                    :: b_c, x_c, r_c, irn_c, jcn_c, val_c

    nullify(b) 
    nullify(x) 
    nullify(r) 
    
    if((m.ne.-1) .and. ((m.lt.11).and.(m.gt.size(matrices)))) then
       write(*,'("Matrix ",i2," is not available for this test")')m
       return
    end if

    ok = .true.
    
    ! loop over all the file-matrices
    do i=11, size(matrices)
       if((m.ne.-1) .and. (i.ne.m)) cycle
       info = 0
       
       ! get the data
       qrm_spmat => cqrm_get_test_mat(i)
       if(info.eq.0) call qrm_alloc(b, qrm_spmat%m, info)
       if(info.eq.0) call qrm_alloc(r, qrm_spmat%m, info)
       if(info.eq.0) call qrm_alloc(x, qrm_spmat%n, info)
       if(info.ne.0) goto 9999
       call clarnv(2, iseed, size(b), b(1))
       r = b
          
       irn_c = c_loc(qrm_spmat%irn(1))
       jcn_c = c_loc(qrm_spmat%jcn(1))
       val_c = c_loc(qrm_spmat%val(1))
       b_c   = c_loc(b(1))
       x_c   = c_loc(x(1))
       r_c   = c_loc(r(1))
          
       ! solve and get back the error
       info = cqrm_test_solve_c(qrm_spmat%m, qrm_spmat%n, qrm_spmat%nz, &
            irn_c, jcn_c, val_c,                                        &
            b_c, x_c, r_c,                                              &
            eps, err)
       
       ! print message
       call cqrm_prnt_testmesg(3, "c_int", 1, 1, i, (info .eq. 0) .and. (err .lt. eps))
       ok = ok .and. (info .eq. 0) .and. (err .lt. eps)
       
9999   continue
       call qrm_dealloc(b)
       call qrm_dealloc(r)
       call qrm_dealloc(x)

    end do

    return

  end function case1

  
  !> @brief Test C problem solve
  function case2(m) result(ok)
    use cqrm_mod
    use cqrm_c_interface
    implicit none
    logical :: ok
    integer :: m

    interface
        function cqrm_test_methods_c(m, n, nz, &
            irn, jcn, val, &
            b_c, x_c, r_c, &
            eps, err) result(info) bind(c)
         use iso_c_binding
         use cqrm_c_interface
         real(r32)               :: err
         real(r32), value        :: eps
         integer(c_int), value   :: m, n, nz
         ! type(cqrm_spmat_type_c) :: qrm_spmat_c
         type(c_ptr), value      :: b_c, x_c, r_c, irn, jcn, val
         integer(c_int)          :: info
       end function cqrm_test_methods_c
    end interface
    
    type(cqrm_spmat_type), pointer :: qrm_spmat
    integer                        :: info
    complex(r32), pointer          :: x(:), b(:), r(:)
    real(r32)                      :: err
    integer                        :: i
    type(c_ptr)                    :: b_c, x_c, r_c, irn_c, jcn_c, val_c

    nullify(b) 
    nullify(x) 
    nullify(r) 
    
    if((m.ne.-1) .and. ((m.lt.11).and.(m.gt.size(matrices)))) then
       write(*,'("Matrix ",i2," is not available for this test")')m
       return
    end if

    ok = .true.
    
    ! loop over all the file-matrices
    do i=11, size(matrices)
       if((m.ne.-1) .and. (i.ne.m)) cycle
       info = 0
       
       ! get the data
       qrm_spmat => cqrm_get_test_mat(i)
       if(info.eq.0) call qrm_alloc(b, qrm_spmat%m, info)
       if(info.eq.0) call qrm_alloc(r, qrm_spmat%m, info)
       if(info.eq.0) call qrm_alloc(x, qrm_spmat%n, info)
       if(info.ne.0) goto 9999
       call clarnv(2, iseed, size(b), b(1))

       irn_c = c_loc(qrm_spmat%irn(1))
       jcn_c = c_loc(qrm_spmat%jcn(1))
       val_c = c_loc(qrm_spmat%val(1))
       b_c   = c_loc(b(1))
       x_c   = c_loc(x(1))
       r_c   = c_loc(r(1))
          
       ! solve and get back the error
       info = cqrm_test_methods_c(qrm_spmat%m, qrm_spmat%n, qrm_spmat%nz, &
            irn_c, jcn_c, val_c,                                        &
            b_c, x_c, r_c,                                              &
            eps, err)
       
       ! print message
       call cqrm_prnt_testmesg(3, "c_int", 2, 1, i, (info .eq. 0) .and. (err .lt. eps))
       ok = ok .and. (info .eq. 0) .and. (err .lt. eps)          

9999   continue
       call qrm_dealloc(b)
       call qrm_dealloc(r)
       call qrm_dealloc(x)

    end do

    return

  end function case2
  



end function cqrm_test_c
