!! ##############################################################################################
!!
!! Copyright 2012-2018 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################

program sqrm_dspotr

  use sqrm_mod
  implicit none

  type(sqrm_dsmat_type), target  :: a, rr, b, x, r
  type(qrm_dscr_type)            :: qrm_dscr

  integer                        :: info, i, j, k, ounit, it, nth 
  integer                        :: m, n, mb, ts, te, tr, verb
  real(kind(1.d0))               :: tfct, tslv, gflops
  real(r32), allocatable      :: af(:,:)
  integer                        :: iseed(4)
  logical                        :: pin=.true.
  real(r32)                      :: rnrm, bnrm, xnrm, err, anrm
  integer                        :: ncpu, ngpu
  
  call qrm_init()

  call get_cla(m, mb, verb)

  call qrm_dscr_init(qrm_dscr)

  call qrm_get('qrm_ncpu', ncpu)
  call qrm_get('qrm_ngpu', ngpu)

  ounit = 0
  if(verb.ge.3) call qrm_get('qrm_ounit', ounit)
  
  if(ounit.gt.0) write(ounit,'(" ")')
  if(ounit.gt.0) write(ounit,'(60("="))')
  if(ounit.gt.0) write(ounit,'("   Generating problem ")')
  if(ounit.gt.0) write(ounit,'(" ")')
  if(ounit.gt.0) write(ounit,'("   - Matrix size (MxM) : ",i5,"x",i5)')m,m
  if(ounit.gt.0) write(ounit,'("   - mb                : ",i5)')mb
  if(ounit.gt.0) write(ounit,'("   - ncpu              : ",i5)')ncpu
  if(ounit.gt.0) write(ounit,'("   - ngpu              : ",i5)')ngpu

  gflops = (real(m)**3)/3.d0/1e9

  call qrm_alloc(af, m, m)
  iseed = (/1, 1, 1, 1/)
  call slarnv(2, iseed, m*m, af)

  do i=1, m
     do j=1, i-1
        af(i,j) = (af(j,i))
     end do
     af(i,i) = real(af(i,i))+real(m)
  end do

  call qrm_dsmat_init(a , m, m, mb=mb, nb=mb, init='m', imat=af, info=info)
  call qrm_dealloc(af)

  call qrm_dsmat_init(rr, m, m, mb=mb, nb=mb, init='v', ivalue=qrm_szero, pin=pin, info=info)
  call qrm_dsmat_copy(a, rr, info=info)
  
  call qrm_dsmat_nrm(a, anrm, info=info)

  ! call qrm_dsmat_print(rr, 'f15.10',ir=.true.)
  
  ! ########       ###        ######     ########     #######  
  ! ##            ## ##      ##    ##       ##       ##     ## 
  ! ##           ##   ##     ##             ##       ##     ## 
  ! ######      ##     ##    ##             ##       ##     ## 
  ! ##          #########    ##             ##       ##     ## 
  ! ##          ##     ##    ##    ##       ##       ##     ## 
  ! ##          ##     ##     ######        ##        #######  
  
  if(ounit.gt.0) write(ounit,'(" ")')
  if(ounit.gt.0) write(ounit,'(60("="))')
  if(ounit.gt.0) write(ounit,'("   Factorization ")')
  if(ounit.gt.0) write(ounit,'(" ")')
  
  call qrm_dscr_init(qrm_dscr)
  call qrm_barrier()

  call qrm_start_profiling()
  call system_clock(ts)
  call qrm_dsmat_potrf_async(qrm_dscr, 'u', rr)
  call qrm_barrier()
  call system_clock(te,tr)
  call qrm_stop_profiling()
  
  call qrm_dscr_destroy(qrm_dscr)
  tfct = max(real(te-ts)/real(tr),1e-20) ! the max is to avoid signaling divide by zero
  if(ounit.gt.0) write(ounit,'("         -- time             : ",f8.3)')tfct
  if(ounit.gt.0) write(ounit,'("         -- Gflop/s          : ",f7.2)')gflops/tfct



  !  ######      #######     ##          ##     ##    ######## 
  ! ##    ##    ##     ##    ##          ##     ##    ##       
  ! ##          ##     ##    ##          ##     ##    ##       
  !  ######     ##     ##    ##          ##     ##    ######   
  !       ##    ##     ##    ##           ##   ##     ##       
  ! ##    ##    ##     ##    ##            ## ##      ##       
  !  ######      #######     ########       ###       ########   
  
  if(ounit.gt.0) write(ounit,'(" ")')
  if(ounit.gt.0) write(ounit,'(" ")')
  if(ounit.gt.0) write(ounit,'(60("="))')

  call qrm_dsmat_init(b , m, 1, mb, mb, init='v', ivalue=qrm_sone , info=info)
  call qrm_dsmat_init(r , m, 1, mb, mb, init='v', ivalue=qrm_szero, info=info)
  call qrm_dsmat_init(x , m, 1, mb, mb, init='v', ivalue=qrm_szero, info=info)
  call qrm_dsmat_copy(b, x, info=info)

  if(ounit.gt.0) write(ounit,'("   Solving Ax=b")')
  if(ounit.gt.0) write(ounit,'(" ")')
  
  call system_clock(ts)
  call qrm_dsmat_trsm(qrm_left, qrm_upper, qrm_transp, qrm_no_unit, qrm_sone, rr, x, info=info)
  call qrm_dsmat_trsm(qrm_left, qrm_upper, qrm_no_transp, qrm_no_unit, qrm_sone, rr, x, info=info)
  call system_clock(te,tr)

  tslv = max(real(te-ts)/real(tr),1e-20) ! the max is to avoid signaling divide by zero
  if(ounit.gt.0) write(ounit,'("         -- time             : ",f8.3)')tslv
  
  call qrm_dsmat_nrm(x, xnrm, info=info)
  call qrm_dsmat_copy(b, r, info=info)
  call qrm_dsmat_gemm(qrm_no_transp, qrm_no_transp, qrm_sone, a, x, qrm_smone, r, info=info)
  call qrm_dsmat_nrm(r, rnrm, info=info)

  if(ounit.gt.0) write(ounit,'("         -- ||x||            : ",e12.5)')xnrm     
  if(ounit.gt.0) write(ounit,'("         -- ||r||/||A||      : ",e12.5)')rnrm/anrm
  if(ounit.gt.0) write(ounit,'(" ")')

  ounit = 0
  if(verb.eq.2) call qrm_get('qrm_ounit', ounit)
  if(ounit.gt.0) write(ounit,'("|   m   |  mb  | ncpu | ngpu |   tf    | Gflop/s |   ts   |   ||x||     | ||r||/||A|| |")')

  if(verb.eq.2 .or. verb.eq.1) call qrm_get('qrm_ounit', ounit)
  if(ounit.gt.0) write(ounit,'("|",i6," |",i5," |",i5," |",i5," |",f8.3," |",f8.2," |",f7.3," |",e12.5," |",e12.5," |")')&
       m, mb, ncpu, ngpu, tfct, gflops/tfct, tslv, xnrm, rnrm/anrm
  
10 continue 

  ! cleanup
  
  call qrm_dsmat_destroy(a)
  call qrm_dsmat_destroy(rr, pin)
  call qrm_dsmat_destroy(b)
  call qrm_dsmat_destroy(x)
  call qrm_dsmat_destroy(r)
     
  call qrm_finalize()
     
  stop


contains

  subroutine get_cla(m, mb, verb)
    use qrm_cli_mod
    implicit none

    integer            :: m, mb, verb, info
    type(qrm_cli_type) :: cli
    
    call cli%init(title='sqrm_dspotr', description='A test program for the Cholesky factorization of dense matrices')
    call cli%add('-m'  , 'Number of rows in the matrix.'                              , def=1024            )
    call cli%add('-mb' , 'Blocks rowsize.'                                            , def=qrm_default_mb_ )
    call cli%add('-v'  , 'Verbosity: 3=full, 2=batch w/ header, 1=batch w/o header.'  , def=3 )

    call cli%parse(info)

    call cli%get('-m'  , m)
    call cli%get('-mb' , mb)
    call cli%get('-v'  , verb)

    return
    
  end subroutine get_cla
  
end program sqrm_dspotr
