!! ##############################################################################################
!!
!! Copyright 2012-2018 CNRS, INPT
!! Copyright 2013-2015 UPS
!!  
!! This file is part of qr_mumps.
!!  
!! qr_mumps is free software: you can redistribute it and/or modify
!! it under the terms of the GNU Lesser General Public License as 
!! published by the Free Software Foundation, either version 3 of 
!! the License, or (at your option) any later version.
!!  
!! qr_mumps is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU Lesser General Public License for more details.
!!  
!! You can find a copy of the GNU Lesser General Public License
!! in the qr_mumps/doc directory.
!!
!! ##############################################################################################

program zqrm_dsgeqr

  use zqrm_dsmat_mod
  use zqrm_mod
  use qrm_string_mod
#if defined(have_starpu) 
  use fstarpu_mod
#endif
  implicit none


  type(zqrm_dsmat_type), target  :: afull, qrfull, bfull, xfull, rfull, tfull
  type(zqrm_dsmat_type), target  :: acheck, qrcheck, bcheck, xcheck, rcheck
  type(zqrm_dsmat_type), target  :: a, qr, b, x, r, t
  type(zqrm_ws_type)             :: w
  type(qrm_dscr_type)            :: qrm_dscr

  integer                        :: info, i, j, k, ounit, it, nth, pid, parts, verb
  integer                        :: qm, qn, m, n, mb, nb, ib, bh, warm, ts, te, tr, pp
  integer, allocatable           :: stair(:)
  real(r32)                      :: tfct, tslv, gflops, mx, sp
  character(LEN=20)              :: str
  integer                        :: iseed(4)
  logical                        :: do_stair=.false., pin=.true.
  real(r64)                      :: rnrm, onrm, bnrm, xnrm, err, anrm
  complex(r64)                   :: alpha
  character                      :: transp
  integer                        :: ncpu, ngpu

  
  call qrm_init()

  call get_cla(m, n, mb, nb, ib, bh, pp, transp, do_stair, verb)

  call qrm_dscr_init(qrm_dscr)

  call qrm_get('qrm_ncpu', ncpu)
  call qrm_get('qrm_ngpu', ngpu)
  ounit = 0
  if(verb.ge.3) call qrm_get('qrm_ounit', ounit)
  
  if(ounit.gt.0) write(ounit,'(" ")')
  if(ounit.gt.0) write(ounit,'(60("="))')
  if(ounit.gt.0) write(ounit,'("   Generating problem ")')
  if(ounit.gt.0) write(ounit,'(" ")')
  if(ounit.gt.0) write(ounit,'("   - Matrix size (MxN) : ",i5,"x",i5)')m,n
  if(ounit.gt.0) write(ounit,'("   - mb                : ",i5)')mb
  if(ounit.gt.0) write(ounit,'("   - nb                : ",i5)')nb
  if(ounit.gt.0) write(ounit,'("   - ib                : ",i5)')ib
  if(ounit.gt.0) write(ounit,'("   - bh                : ",i5)')bh
  if(ounit.gt.0) write(ounit,'("   - ncpu              : ",i5)')ncpu
  if(ounit.gt.0) write(ounit,'("   - ngpu              : ",i5)')ngpu
  if(ounit.gt.0) write(ounit,'("   - prepartition      : ",i5)')pp
  if(ounit.gt.0) write(ounit,'("   - do_stair          : ",l5)')do_stair
  if(ounit.gt.0) write(ounit,'("   - transp            : ",a5)')transp

  ! initialize original data
  if(transp.eq.qrm_conj_transp) then
     qm = n
     qn = m
  else
     qm = m
     qn = n
  end if

  if(qm.ge.qn) then
     gflops = (2.d0*(real(qm)-real(qn)/3.d0)*real(qn)**2)/1e9
  else
     gflops = (2.d0*(real(qn)-real(qm)/3.d0)*real(qm)**2)/1e9
  end if

  if(do_stair) then
     call qrm_alloc(stair, qn)
     sp = 0.2e0
     call fill_stair(qm, qn, floor(qm*0.15), sp, stair, info)
     call qrm_dsmat_init(qr, qm, qn, mb, nb, pin=pin, stair=stair, init='r', info=info)
     call qrm_dealloc(stair)
  else
     call qrm_dsmat_init(qr, qm, qn, mb, nb, pin=pin, init='r', info=info)
  end if

  call qrm_dsmat_init(a, m, n, mb, nb, pin=pin, init='v', ivalue=qrm_zzero, info=info)
  call qrm_dsmat_copy(qr, a, transp=transp, info=info)
  
  call qrm_dsmat_nrm(a, anrm, info=info)

  if(ounit.gt.0) write(ounit,'(" ")')
  if(ounit.gt.0) write(ounit,'(" ")')
  

  ! ########       ###        ######     ########     #######  
  ! ##            ## ##      ##    ##       ##       ##     ## 
  ! ##           ##   ##     ##             ##       ##     ## 
  ! ######      ##     ##    ##             ##       ##     ## 
  ! ##          #########    ##             ##       ##     ## 
  ! ##          ##     ##    ##    ##       ##       ##     ## 
  ! ##          ##     ##     ######        ##        #######  
  
  if(ounit.gt.0) write(ounit,'(" ")')
  if(ounit.gt.0) write(ounit,'(60("="))')
  if(ounit.gt.0) write(ounit,'("   Factorization ")')
  if(ounit.gt.0) write(ounit,'(" ")')
  
  ! call qrm_dsmat_print(qr, 'f20.15', unit=20)
  
  call qrm_dsmat_init_t(qr, t, ib, bh, pin, info=info)
  call qrm_ws_init(w, 2*max(mb,nb), ib)
  call qrm_dscr_init(qrm_dscr)
  call qrm_barrier()

#if defined(have_starpu) 
  call fstarpu_fxt_start_profiling()
#endif
  call system_clock(ts)
  call qrm_dsmat_geqr_async(qrm_dscr, qr, t, ib, bh, w)
  call qrm_barrier()
  call system_clock(te,tr)
#if defined(have_starpu) 
  call fstarpu_fxt_stop_profiling()
#endif
  
  ! call qrm_dsmat_print(qr, 'f20.15', unit=21)

  ! call qrm_dsmat_print(t, 'f20.15', unit=22)
  
  call qrm_dscr_destroy(qrm_dscr)
  tfct = max(real(te-ts)/real(tr),1e-20) ! the max is to avoid signaling divide by zero
  if(ounit.gt.0) write(ounit,'("         -- time             : ",f8.3)')tfct
  if(ounit.gt.0) write(ounit,'("         -- Gflop/s          : ",f7.2)')gflops/tfct



  !  ######      #######     ##          ##     ##    ######## 
  ! ##    ##    ##     ##    ##          ##     ##    ##       
  ! ##          ##     ##    ##          ##     ##    ##       
  !  ######     ##     ##    ##          ##     ##    ######   
  !       ##    ##     ##    ##           ##   ##     ##       
  ! ##    ##    ##     ##    ##            ## ##      ##       
  !  ######      #######     ########       ###       ########   
  
  if(ounit.gt.0) write(ounit,'(" ")')
  if(ounit.gt.0) write(ounit,'(" ")')
  if(ounit.gt.0) write(ounit,'(60("="))')

  call qrm_dsmat_init(b , m, 1, mb, mb, init='v', ivalue=qrm_zone , info=info)
  call qrm_dsmat_init(r , m, 1, mb, mb, init='v', ivalue=qrm_zzero, info=info)
  call qrm_dsmat_init(x , n, 1, mb, mb, init='v', ivalue=qrm_zzero, info=info)
  call qrm_dsmat_copy(b, r, info=info)

  if(m.gt.n) then
     if(ounit.gt.0) write(ounit,'("   Solving argmin(x, ||Ax-b||)")')
  else
     if(transp.eq.qrm_conj_transp) then
        if(ounit.gt.0) write(ounit,'("   Solving min(||x|| | Ax=b)")')
     else
        if(ounit.gt.0) write(ounit,'("   Solving Ax=b")')
     end if
  end if
  if(ounit.gt.0) write(ounit,'(" ")')
  
  call system_clock(ts)
  if((m.gt.n) .or. (transp.ne.qrm_conj_transp)) then
     call qrm_dsmat_gemqr(qrm_conj_transp, qr, t, b, ib, bh, w, info=info)
     call qrm_dsmat_copy(b, x, m=min(m,n), info=info)
     call qrm_dsmat_trsm(qrm_left, qrm_upper, qrm_no_transp, qrm_no_unit, qrm_zone, qr, x, info=info)
  else
     call qrm_dsmat_trsm(qrm_left, qrm_upper, qrm_conj_transp, qrm_no_unit, qrm_zone, qr, b, info=info)
     call qrm_dsmat_copy(b, x, m=min(m,n), info=info)
     call qrm_dsmat_gemqr(qrm_no_transp, qr, t, x, ib, bh, w, info=info)
  end if
  call system_clock(te,tr)

  tslv = max(real(te-ts)/real(tr),1e-20) ! the max is to avoid signaling divide by zero
  if(ounit.gt.0) write(ounit,'("         -- time             : ",f8.3)')tslv
  
  ! one more copy to make sure bfull is not modified
  call qrm_dsmat_nrm(x, xnrm, info=info)
  call qrm_dsmat_gemm(qrm_no_transp, qrm_no_transp, qrm_zmone, a, x, qrm_zone, r, info=info)
  ! compute A'*r and store the result in x
  call qrm_dsmat_gemm(qrm_conj_transp, qrm_no_transp, qrm_zone, a, r, qrm_zzero, x, info=info)
  ! call qrm_dsmat_copy(r, rcheck, info=info)
  call qrm_dsmat_nrm(x, onrm, info=info)
  call qrm_dsmat_nrm(r, rnrm, info=info)

  if(ounit.gt.0) write(ounit,'("         -- ||x||            : ",e12.5)')xnrm     
  if(ounit.gt.0) write(ounit,'("         -- ||r||/||A||      : ",e12.5)')rnrm/anrm
  if(ounit.gt.0) write(ounit,'("         -- ||A^Tr||/||r||   : ",e12.5)')onrm/rnrm
  if(ounit.gt.0) write(ounit,'(" ")')

  ounit = 0
  if(verb.eq.2) call qrm_get('qrm_ounit', ounit)
  if(ounit.gt.0) write(ounit,'("|   m   |   n   |  mb  |  nb  |  ib |  ib | ncpu | ngpu |   tf    | Gflop/s'// &
       ' |   ts   |   ||x||     | ||r||/||A|| | ||A^Tr||/||r|| |")')
  
  if(verb.eq.2 .or. verb.eq.1) call qrm_get('qrm_ounit', ounit)
  if(ounit.gt.0) write(ounit,'("|",i6," |",i6," |",i5," |",i5," |",i4," |",i4," |",i5," |",i5," |",f8.3," |",f8.2,'// &
       '" |",f7.3," |",e12.5," |",e12.5," |   ",e12.5," |")')&
       m, n, mb, nb, ib, bh, ncpu, ngpu, tfct, gflops/tfct, tslv, xnrm, rnrm/anrm,onrm/rnrm
  
  
10 continue 

  ! cleanup
  
  call qrm_dsmat_destroy(a, pin)
  call qrm_dsmat_destroy(qr, pin)
  call qrm_dsmat_destroy(b)
  call qrm_dsmat_destroy(x)
  call qrm_dsmat_destroy(r)
  call qrm_dsmat_destroy(t, pin)
  call qrm_ws_destroy(w)
     
  call qrm_finalize()
     
  stop


contains

subroutine fill_stair(m, n, npiv, sp, stair, info)
  implicit none

  integer              :: m, n, npiv, k
  integer              :: stair(:)
  real(r32)            :: sp
  integer, optional    :: info
                        
                        
  real(r32)            :: rnd, zcnt
  integer              :: ln, pnt, set, div, izcnt, cnt, maxz, s
  integer, allocatable :: zcol(:)
  integer, allocatable :: seed(:)
  
10 continue
  maxz = (min(m,n)*(min(m,n)+1))/2 - (npiv*(npiv+1)/2)
  
  zcnt = real(m)*real(n)*sp
  izcnt = ceiling(zcnt)

  if(izcnt .gt. maxz) then
     if(.not. present(info)) then
        if(ounit.gt.0) write(ounit,'("too sparse!")')
        stop
     else
        sp = sp/2.0
        if(ounit.gt.0) write(ounit,'("Too sparse. Reducing sp to ",f6.3)')sp
        goto 10
     end if
  end if

  allocate(zcol(n))
  zcol = 0

  call random_seed(size = s)
  allocate(seed(s))
  seed = 1
  call random_seed(put = seed)
  deallocate(seed)
  
  loop1: do
     cnt = 1
     loop2: do
        if(izcnt .eq. 0) exit loop1
        if(cnt .gt. n) exit loop2
        
        call random_number(rnd)
        k = ceiling(rnd*real(min(izcnt,m-npiv)))
!         k = min(k, m-npiv)
        zcol(cnt) = zcol(cnt)+k
        izcnt = izcnt -k
        cnt = cnt+1

     end do loop2

  end do loop1

  call qsortc(zcol)
  stair = m
  do k=1, n
     stair(k) = m-zcol(k)
  end do

  if(present(info)) info = 1

  return
end subroutine fill_stair

  recursive subroutine qsortc(a)
    integer, intent(in out), dimension(:) :: a
    integer :: iq

    if(size(a) > 1) then
       call partition(a, iq)
       call qsortc(a(:iq-1))
       call qsortc(a(iq:))
    endif
  end subroutine qsortc

  subroutine partition(a, marker)
    integer, intent(in out), dimension(:) :: a
    integer, intent(out) :: marker
    integer :: i, j
    integer :: temp
    integer :: x      ! pivot point
    x = a(1)
    i= 0
    j= size(a) + 1

    do
       j = j-1
       do
          if (a(j) >= x) exit
          j = j-1
       end do
       i = i+1
       do
          if (a(i) <= x) exit
          i = i+1
       end do
       if (i < j) then
          ! exchange a(i) and a(j)
          temp = a(i)
          a(i) = a(j)
          a(j) = temp
       elseif (i == j) then
          marker = i+1
          return
       else
          marker = i
          return
       endif
    end do

  end subroutine partition


  subroutine get_cla(m, n, mb, nb, ib, bh, pp, transp, do_stair, verb)
    use qrm_cli_mod
    implicit none

    integer            :: m, n, mb, nb, ib, bh, pp, verb
    logical            :: do_stair
    type(qrm_cli_type) :: cli
    character          :: transp
    logical            :: itransp
    integer            :: eunit

    call qrm_get('qrm_eunit', eunit)

    call cli%init(title='zqrm_dsgeqr', description='A test program for te QR factorization of dense matrices')
    call cli%add('-m'  , 'Number of rows in the matrix.'                              , def=1024             )
    call cli%add('-n'  , 'Number of cols in the matrix. Equal to m if not specified.' , def=1024             )
    call cli%add('-mb' , 'Blocks rowsize.'                                            , def=qrm_default_mb_  )
    call cli%add('-nb' , 'Blocks colsize. Equal to mb if not specified.'              , def=qrm_default_nb_  )
    call cli%add('-ib' , 'Internal blocksize.'                                        , def=qrm_default_ib_  )
    call cli%add('-bh' , 'Size of reduction trees subdomains.'                        , def=qrm_default_bh_  )
    call cli%add('-p'  , 'Number of pre-partitioned blocks.'                          , def=0                )
    call cli%add('-t'  , 'If set, factorize the transpose if m>=n.'                   , def=.false.          )
    call cli%add('-s'  , 'If set, generate a staircase structure.'                    , def=.false.          )
    call cli%add('-v'  , 'Verbosity: 3=full, 2=batch w/ header, 1=batch w/o header.'  , def=3 )

    call cli%parse(info)
    if(info.ne.0) stop
    
    call cli%get('-m', m)
    if(.not. cli%present('-n')) then
       n = m
    else
       call cli%get('-n', n)
    end if
    call cli%get('-mb', mb)
    if(.not. cli%present('-nb')) then
       nb = mb
    else
       call cli%get('-nb', nb)
       if(nb.gt.mb) then
          nb = mb
       else if (mod(mb,nb).ne.0) then
          if(eunit.gt.0) write(eunit,'("mb has to be a multiple of nb. Setting nb equal to mb (",i4,")")')mb
          nb = mb
       end if
    end if

    call cli%get('-ib', ib)
    if(ib.gt.nb) then
       ib = nb
    else if (mod(nb,ib).ne.0) then
       if(eunit.gt.0) write(eunit,'("nb has to be a multiple of ib. Setting ib equal to nb (",i4,")")')nb
       ib = nb
    end if

    call cli%get('-bh' , bh)
    call cli%get('-p'  , pp)
    call cli%get('-t'  , itransp)
    call cli%get('-s'  , do_stair)

    if(itransp) then
       if (m.gt.n) then
          if(eunit.gt.0) write(eunit,'("Transpose is only accepted if m<=n. Ignoring.")')
          transp = qrm_no_transp
       else
          transp = qrm_conj_transp
       end if
    else
       transp = qrm_no_transp
    end if

    call cli%get('-v'  , verb)
        
    return
    
  end subroutine get_cla
  
end program zqrm_dsgeqr
